Package: atena
Type: Package
Title: Analysis of Transposable Elements
Version: 1.19.0
Authors@R: c(
    person("Beatriz", "Calvo-Serra", email="beatriz.calvo@upf.edu",
           role=c("aut")),
    person("Robert", "Castelo", email="robert.castelo@upf.edu",
           role=c("aut", "cre")))
Description: Quantify expression of transposable elements (TEs) from
        RNA-seq data through different methods, including ERVmap,
        TEtranscripts and Telescope. A common interface is provided to
        use each of these methods, which consists of building a
        parameter object, calling the quantification function with this
        object and getting a SummarizedExperiment object as output
        container of the quantified expression profiles. The
        implementation allows one to quantify TEs and gene transcripts
        in an integrated manner.
License: Artistic-2.0
Encoding: UTF-8
Depends: R (>= 4.3.0), SummarizedExperiment
Imports: methods, stats, Matrix, BiocGenerics, MatrixGenerics,
        BiocParallel, S4Vectors, IRanges, Seqinfo, GenomicFeatures,
        GenomicRanges, GenomicAlignments, Rsamtools, GenomeInfoDb,
        SQUAREM, sparseMatrixStats, AnnotationHub, matrixStats, cli
Suggests: covr, BiocStyle, knitr, rmarkdown, RUnit,
        TxDb.Dmelanogaster.UCSC.dm6.ensGene, RColorBrewer
biocViews: Transcription, Transcriptomics, RNASeq, Sequencing,
        Preprocessing, Software, GeneExpression, Coverage,
        DifferentialExpression, FunctionalGenomics
VignetteBuilder: knitr
URL: https://github.com/rcastelo/atena
BugReports: https://github.com/rcastelo/atena/issues
RoxygenNote: 7.3.2
Collate: 'AllGenerics.R' 'AllClasses.R' 'ERVmap.R' 'TEtranscripts.R'
        'Telescope.R' 'annotations.R' 'atena.R' 'atenaMethod.R'
        'overlappingModes.R' 'qtex.R' 'utils.R' 'zzz.R'
Config/pak/sysreqs: make libbz2-dev libicu-dev liblzma-dev libpng-dev
        libxml2-dev libssl-dev xz-utils zlib1g-dev
Repository: https://bioc.r-universe.dev
Date/Publication: 2026-04-28 12:56:57 UTC
RemoteUrl: https://github.com/bioc/atena
RemoteRef: HEAD
RemoteSha: 4524201929394e9a44ff4ffe5d3da45d16e553d6
NeedsCompilation: no
Packaged: 2026-07-04 05:38:37 UTC; root
Author: Beatriz Calvo-Serra [aut],
  Robert Castelo [aut, cre]
Maintainer: Robert Castelo <robert.castelo@upf.edu>
Built: R 4.6.1; ; 2026-07-04 06:14:20 UTC; unix

# Generated by roxygen2: do not edit by hand

export(ERVmapParam)
export(OneCodeToFindThemAll)
export(TEtranscriptsParam)
export(TelescopeParam)
export(annotaTEs)
export(atenaParam)
export(getDNAtransposons)
export(getLINEs)
export(getLTRs)
export(getSINEs)
export(ovIntersectionStrict)
export(ovUnion)
export(rmskatenaparser)
export(rmskbasicparser)
export(rmskidentity)
exportClasses(ERVmapParam)
exportClasses(QuantifyParam)
exportClasses(TEtranscriptsParam)
exportClasses(TelescopeParam)
exportClasses(atenaParam)
exportMethods(features)
exportMethods(path)
exportMethods(qtex)
exportMethods(show)
importClassesFrom(GenomicAlignments,GAlignments)
importClassesFrom(GenomicRanges,GenomicRanges_OR_GenomicRangesList)
importClassesFrom(Matrix,lgCMatrix)
importClassesFrom(Rsamtools,BamFileList)
importFrom(AnnotationHub,AnnotationHub)
importFrom(AnnotationHub,query)
importFrom(BiocGenerics,basename)
importFrom(BiocGenerics,path)
importFrom(BiocParallel,SerialParam)
importFrom(BiocParallel,bplapply)
importFrom(GenomeInfoDb,"seqlevelsStyle<-")
importFrom(GenomeInfoDb,keepSeqlevels)
importFrom(GenomeInfoDb,seqlevelsStyle)
importFrom(GenomicAlignments,cigar)
importFrom(GenomicAlignments,explodeCigarOpLengths)
importFrom(GenomicAlignments,extractAlignmentRangesOnReference)
importFrom(GenomicAlignments,findOverlaps)
importFrom(GenomicAlignments,first)
importFrom(GenomicAlignments,qwidth)
importFrom(GenomicAlignments,readGAlignmentPairs)
importFrom(GenomicAlignments,readGAlignments)
importFrom(GenomicAlignments,readGAlignmentsList)
importFrom(GenomicAlignments,second)
importFrom(GenomicAlignments,seqnames)
importFrom(GenomicAlignments,start)
importFrom(GenomicFeatures,features)
importFrom(GenomicRanges,"mcols<-")
importFrom(GenomicRanges,GRanges)
importFrom(GenomicRanges,GRangesList)
importFrom(GenomicRanges,end)
importFrom(GenomicRanges,granges)
importFrom(GenomicRanges,mcols)
importFrom(GenomicRanges,pintersect)
importFrom(GenomicRanges,reduce)
importFrom(GenomicRanges,seqnames)
importFrom(GenomicRanges,start)
importFrom(GenomicRanges,strand)
importFrom(GenomicRanges,width)
importFrom(IRanges,CharacterList)
importFrom(IRanges,IRanges)
importFrom(IRanges,IRangesList)
importFrom(IRanges,findOverlapPairs)
importFrom(IRanges,mean)
importFrom(IRanges,ranges)
importFrom(IRanges,relist)
importFrom(IRanges,unique)
importFrom(Matrix,Matrix)
importFrom(Matrix,colSums)
importFrom(Matrix,rowSums)
importFrom(Matrix,sparseMatrix)
importFrom(Matrix,summary)
importFrom(Matrix,t)
importFrom(Matrix,which)
importFrom(MatrixGenerics,rowRanges)
importFrom(Rsamtools,"asMates<-")
importFrom(Rsamtools,"yieldSize<-")
importFrom(Rsamtools,BamFileList)
importFrom(Rsamtools,ScanBamParam)
importFrom(Rsamtools,asMates)
importFrom(Rsamtools,bamFlagTest)
importFrom(Rsamtools,scanBamFlag)
importFrom(Rsamtools,scanBamHeader)
importFrom(Rsamtools,yieldSize)
importFrom(S4Vectors,DataFrame)
importFrom(S4Vectors,Hits)
importFrom(S4Vectors,List)
importFrom(S4Vectors,Rle)
importFrom(S4Vectors,countQueryHits)
importFrom(S4Vectors,countSubjectHits)
importFrom(S4Vectors,decode)
importFrom(S4Vectors,first)
importFrom(S4Vectors,from)
importFrom(S4Vectors,isSorted)
importFrom(S4Vectors,mcols)
importFrom(S4Vectors,nLnode)
importFrom(S4Vectors,nRnode)
importFrom(S4Vectors,nrow)
importFrom(S4Vectors,pc)
importFrom(S4Vectors,queryHits)
importFrom(S4Vectors,rownames)
importFrom(S4Vectors,runValue)
importFrom(S4Vectors,second)
importFrom(S4Vectors,split)
importFrom(S4Vectors,subjectHits)
importFrom(S4Vectors,to)
importFrom(SQUAREM,squarem)
importFrom(Seqinfo,"genome<-")
importFrom(Seqinfo,"seqinfo<-")
importFrom(Seqinfo,"seqlevels<-")
importFrom(Seqinfo,genome)
importFrom(Seqinfo,seqinfo)
importFrom(Seqinfo,seqlengths)
importFrom(Seqinfo,seqlevels)
importFrom(Seqinfo,seqnames)
importFrom(SummarizedExperiment,SummarizedExperiment)
importFrom(cli,cli_alert_info)
importFrom(cli,cli_alert_success)
importFrom(matrixStats,logSumExp)
importFrom(methods,as)
importFrom(methods,formalArgs)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,show)
importFrom(sparseMatrixStats,colSums2)
importFrom(sparseMatrixStats,rowMaxs)
importFrom(sparseMatrixStats,rowSums2)
importFrom(stats,aggregate)
importFrom(stats,median)
importFrom(stats,runif)
importFrom(stats,setNames)

# atena: analysis of transposable elements in R and Bioconductor

[![Bioconductor Time](https://bioconductor.org/shields/years-in-bioc/atena.svg)](https://bioconductor.org/packages/release/bioc/html/atean.html "How long has been atena in a release of Bioconductor")
[![Bioconductor Downloads](https://bioconductor.org/shields/downloads/release/atena.svg)](https://bioconductor.org/packages/stats/bioc/atena/ "Ranking by number of downloads. A lower number means the package is downloaded more frequently. Determined within a package type (software, experiment, annotation, workflow) and uses the number of distinct IPs for the last 12 months.")
[![Support posts](https://bioconductor.org/shields/posts/atena.svg)](https://support.bioconductor.org/t/atena/ "Support site activity on atena, last 6 months: answered posts/total posts.")
[![R-CMD-check-bioc](https://github.com/rcastelo/atena/workflows/R-CMD-check-bioc/badge.svg)](https://github.com/rcastelo/atena/actions?query=workflow%3AR-CMD-check-bioc)
[![codecov.io](https://codecov.io/github/rcastelo/atena/coverage.svg?branch=devel)](https://app.codecov.io/github/rcastelo/atena?branch=devel)
<img align="right" src="https://raw.githubusercontent.com/Bioconductor/BiocStickers/master/atena/atena.png" height="200"/>

**Current Bioconductor build status**
- `release` [![Bioconductor Availability](https://bioconductor.org/shields/availability/release/atena.svg)](https://bioconductor.org/packages/release/bioc/html/atena.html#archives "Whether atena release is available on all platforms")
[![Bioconductor Dependencies](https://bioconductor.org/shields/dependencies/release/atena.svg)](https://bioconductor.org/packages/release/bioc/html/atena.html#since "Number of recursive dependencies needed to install package")
[![Bioconductor Commits](https://bioconductor.org/shields/lastcommit/release/bioc/atena.svg)](https://bioconductor.org/checkResults/release/bioc-LATEST/atena "Time since last commit, possible values: today, < 1 week, < 1 month, < 3 months, since release, before release")
[![Bioconductor Release Build](https://bioconductor.org/shields/build/release/bioc/atena.svg)](https://bioconductor.org/checkResults/release/bioc-LATEST/atena/ "Bioconductor release build")
- `development` [![Bioconductor Availability](https://bioconductor.org/shields/availability/devel/atena.svg)](https://bioconductor.org/packages/devel/bioc/html/atena.html#archives "Whether atena devel is available on all platforms")
[![Bioconductor Dependencies](https://bioconductor.org/shields/dependencies/devel/atena.svg)](https://bioconductor.org/packages/devel/bioc/html/atena.html#since "Number of recursive dependencies needed to install package")
[![Bioconductor Commits](https://bioconductor.org/shields/lastcommit/devel/bioc/atena.svg)](https://bioconductor.org/checkResults/devel/bioc-LATEST/atena "Time since last commit, possible values: today, < 1 week, < 1 month, < 3 months, since release, before release")
[![Bioconductor Devel Build](https://bioconductor.org/shields/build/devel/bioc/atena.svg)](https://bioconductor.org/checkResults/devel/bioc-LATEST/atena/ "Bioconductor devel build")

The `atena` package provides access from R/Bioconductor to the following
methods for the quantification of transposable element (TE) expression:

* Jin Y et al. TEtranscripts: a package for including transposable elements
  in differential expression analysis of RNA-seq datasets.
  Bioinformatics. 2015;31(22):3593-3599. DOI:
  [10.1093/bioinformatics/btv422](https://doi.org/10.1093/bioinformatics/btv422).

* Tokuyama M et al. ERVmap analysis reveals genome-wide transcription of human
  endogenous retroviruses. PNAS. 2018;115(50):12565-12572. DOI:
  [10.1073/pnas.1814589115](https://doi.org/10.1073/pnas.1814589115).

* Bendall et al. Telescope: characterization of the retrotranscriptome by
  accurate estimation of transposable element expression.
  PLOS Comp. Biol. 2019;15(9):e1006453. DOI:
  [10.1371/journal.pcbi.1006453](https://doi.org/10.1371/journal.pcbi.1006453).

## Installation

`atena` has been submitted to Bioconductor.

To install the __release__ version of this package please go to its package
release landing page at
[https://bioconductor.org/packages/atena](https://bioconductor.org/packages/atena)
and follow the instructions there to install it.

This github repository contains the __development__ version of the
R/Bioconductor package `atena`. This version is unstable and should be used
only to test new features. To install the development version you first need to
install the
[development version of Bioconductor](https://bioconductor.org/developers/how-to/useDevel)
and then type the following line from the R shell:

```r
BiocManager::install("atena", version="devel")
```

Alternatively, you can directly install it from this GitHub repo as follows:

```
BiocManager::install("rcastelo/atena")
```

## Questions, bug reports and issues

For questions and bug reports regarding the __release__ version of **atena**
please use the
[Bioconductor support site](https://support.bioconductor.org "Bioconductor support site").
For bug reports and issues regarding this __development__ version of **atena**
please use the GitHub issues [tab](https://github.com/rcastelo/atena/issues) at
the top-left of this page.

## Contributing

Contributions to the software codebase of atena are welcome as long as
contributors abide to the terms of the
[Bioconductor Contributor Code of Conduct](https://bioconductor.org/about/code-of-conduct).
If you want to contribute to the development of GenomicScores please open an
[issue](https://github.com/rcastelo/atena/issues) to start discussing your
suggestion or, in case of a bugfix or a straightforward feature, directly a
[pull request](https://github.com/rcastelo/atena/pulls).

## Funding

This software project is supported in part by the Spanish
[Ministry of Science, Innovation and Universities](https://www.ciencia.gob.es).

#' Build an atena parameter object
#'
#' Build an object of the class \code{atenaParam}.
#'
#' @param bfl A \code{BamFile} or \code{BamFileList} object, or a character
#' string vector of BAM filenames.
#'
#' @param teFeatures A \code{GRanges} or \code{GRangesList} object. Elements
#' in this object should have names, which are used as a grouping factor for
#' genomic ranges forming a common locus. This grouping is performed previous
#' to TE expression quantification, unlike the aggregation of quantifications 
#' performed when the \code{aggregateby} parameter is specified, which is 
#' performed after individual TE instances are quantified.
#'
#' @param aggregateby Character vector with column names from the annotation
#' to be used to aggregate quantifications. By default, this is an empty
#' vector, which means that the names of the input \code{GRanges} or
#' \code{GRangesList} object given in the \code{teFeatures} parameter are used
#' to aggregate quantifications.
#' 
#' @param ovMode Character vector indicating the overlapping mode. Available
#' options are: "ovUnion" (default) and "ovIntersectionStrict",
#' which implement the corresponding methods from HTSeq
#' (\url{https://htseq.readthedocs.io/en/release_0.11.1/count.html}).
#' Ambiguous alignments (alignments overlapping > 1 feature) are not counted.
#'
#' @param geneFeatures (Default NULL) A \code{GRanges} or 
#' \code{GRangesList} object with the
#' gene annotated features to be quantified. Unique reads are first tallied
#' with respect to these gene features whereas multi-mapping reads are
#' preferentially assigned to TEs. Elements should have names indicating the
#' gene name/id. In case that \code{geneFeatures} is a \code{GRanges} and
#' contains a metadata column
#' named \code{type}, only the elements with \code{type} = \code{exon} are 
#' considered for the analysis. Then, exon counts are summarized to the gene
#' level. If NULL, gene expression is not quantified.
#'
#' @param singleEnd (Default TRUE) Logical value indicating if reads are single
#' (\code{TRUE}) or paired-end (\code{FALSE}).
#'
#' @param strandMode (Default 1) Numeric vector which can take values 0, 1 or
#' 2.
#' The strand mode is a per-object switch on
#' \code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
#' objects that controls the behavior of the strand getter. See
#' \code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
#' class for further detail. If \code{singleEnd = TRUE}, then \code{strandMode}
#' is ignored.
#'
#' @param ignoreStrand (Default FALSE) A logical which defines if the strand
#' should be taken into consideration when computing the overlap between reads
#' and annotated features. When \code{ignoreStrand = FALSE}, an aligned read
#' is considered to be overlapping an annotated feature as long as they
#' have a non-empty intersecting genomic range on the same strand, while when
#' \code{ignoreStrand = TRUE} the strand is not considered.
#'
#' @param fragments (Default TRUE) A logical; applied to paired-end data only.
#' When \code{fragments=FALSE}, the read-counting method only counts
#' ‘mated pairs’ from opposite strands (non-ambiguous properly paired reads), 
#' while when \code{fragments=TRUE} same-strand pairs, singletons, reads with 
#' unmapped pairs and other ambiguous or not properly paired fragments
#' are also counted (see "Pairing criteria" in 
#' \code{\link[GenomicAlignments]{readGAlignments}()}). 
#' For further details see
#' \code{\link[GenomicAlignments]{summarizeOverlaps}()}.
#'
#' @param pi_prior (Default 0) A positive numeric object indicating the prior
#' on pi. The same prior can be specified for all features setting
#' \code{pi_prior} as a scalar, or each feature can have a specific prior by
#' setting \code{pi_prior} as a vector with \code{names()} corresponding to
#' all feature names. Setting a pi prior is equivalent to adding n unique
#' reads. 
#'
#' @param theta_prior (Default 0) A positive numeric object indicating the
#' prior on Q. The same prior can be specified for all features setting
#' \code{theta_prior} as a scalar, or each feature can have a specific prior by
#' setting \code{theta_prior} as a vector with \code{names()} corresponding to
#' all feature names. Equivalent to adding n non-unique reads.
#'
#' @param em_epsilon (Default 1e-7) A numeric scalar indicating the EM
#' Algorithm Epsilon cutoff.
#'
#' @param maxIter A positive integer scalar storing the maximum number of
#' iterations of the EM SQUAREM algorithm (Du and Varadhan, 2020). Default
#' is 100 and this value is passed to the \code{maxiter} parameter of the
#' \code{\link[SQUAREM]{squarem}()} function.
#' 
#' @param reassign_mode (Default 'exclude') Character vector indicating
#' reassignment mode after EM step. 
#' Available methods are 'exclude' (reads with more than one best
#' assignment are excluded from the final counts), 'choose' (when reads have
#' more than one best assignment, one of them is randomly chosen), 'average'
#' (the read count is divided evenly among the best assignments) and 'conf'
#' (only assignments that exceed a certain threshold -defined by 
#' \code{conf_prob} parameter- are accepted, then the read count is
#' proportionally divided among the assignments above \code{conf_prob}).
#' 
#' @param conf_prob (Default 0.9) Minimum probability for high confidence
#' assignment.
#'
#' @param verbose (Default \code{TRUE}) Logical value indicating whether to
#' report progress.
#'
#' @details
#' This is the constructor function for objects of the class
#' \code{atenaParam-class}. This type of object is the input to the
#' function \code{\link{qtex}()} for quantifying expression of transposable
#' elements, which will call the atena method with this type of object. The
#' atena method uses a multiple '__no_feature' approach in which as many
#' '__no_feature' features as different overlapping patterns of multimapping
#' reads in the overlapping matrix are used to represent alignments mapping
#' outside annotations.
#'
#' @return A [`atenaParam`] object.
#'
#' @examples
#' bamfiles <- list.files(system.file("extdata", package="atena"),
#'                        pattern="*.bam", full.names=TRUE)
#' \dontrun{
#' ## use the following two instructions to fetch annotations, they are here
#' ## commented out to enable running this example quickly when building and
#' ## checking the package
#' rmskat <- annotaTEs(genome="dm6", parsefun=rmskatenaparser,
#'                     strict=FALSE, insert=500)
#' rmskLTR <- getLTRs(rmskat, relLength=0.8,
#'                    fullLength=TRUE,
#'                    partial=TRUE,
#'                    otherLTR=TRUE)
#' }
#'
#' ## DO NOT TYPE THIS INSTRUCTION, WHICH JUST LOADS A PRE-COMPUTED ANNOTATION
#' ## YOU SHOULD USE THE INSTRUCTIONS ABOVE TO FETCH ANNOTATIONS
#' rmskLTR <- readRDS(system.file("extdata", "rmskatLTRrlen80flenpartoth.rds",
#'                                package="atena"))
#'
#' ## build a parameter object for the atena method
#' atpar <- atenaParam(bfl=bamfiles,
#'                     teFeatures=rmskLTR,
#'                     singleEnd=TRUE,
#'                     ignoreStrand=TRUE)
#' atpar
#'
#'
#' @importFrom methods is new
#' @importFrom Rsamtools BamFileList
#' @importFrom S4Vectors mcols
#' @importFrom cli cli_alert_info cli_alert_success
#' @export
#' @rdname atenaParam-class
atenaParam <- function(bfl, teFeatures, aggregateby=character(0),
                       ovMode="ovUnion", geneFeatures=NULL, singleEnd=TRUE,
                       strandMode=1L, ignoreStrand=FALSE, fragments=TRUE,
                       pi_prior=0L, theta_prior=0L, em_epsilon=1e-7,
                       maxIter=100L, reassign_mode="exclude", conf_prob=0.9,
                       verbose=TRUE) {

    if (verbose)
        cli_alert_info("Locating BAM files")
    bfl <- .checkBamFileListArgs(bfl, singleEnd, fragments)
    
    if (!reassign_mode %in% c("exclude", "choose", "average", "conf"))
      stop("'reassign_mode' should be one of 'exclude', 'choose', 'average' or 'conf'")
    
    if (!ovMode %in% c("ovUnion","ovIntersectionStrict"))
      stop("'ovMode' should be one of 'ovUnion', 'ovIntersectionStrict'")
    
    if (verbose)
        cli_alert_info("Processing features")
    features <- .processFeatures(teFeatures, deparse(substitute(teFeatures)),
                                 geneFeatures,deparse(substitute(geneFeatures)),
                                 aggregateby, aggregateexons=TRUE)

    .checkPriors(names(features), names(pi_prior), names(theta_prior))
    
    obj <- new("atenaParam", bfl=bfl, features=features,
               aggregateby=aggregateby, ovMode=ovMode,
               singleEnd=singleEnd,ignoreStrand=ignoreStrand,
               strandMode=as.integer(strandMode), fragments=fragments,
               pi_prior=pi_prior, theta_prior=theta_prior,
               em_epsilon=em_epsilon, maxIter=as.integer(maxIter),
               reassign_mode=reassign_mode, conf_prob=conf_prob)
    if (verbose)
        cli_alert_success("Parameter object successfully created")
    obj
}

#' @param object A \linkS4class{atenaParam} object.
#'
#' @importFrom Seqinfo seqlevels
#' @export
#' @aliases show,atenaParam-method
#' @rdname atenaParam-class
setMethod("show", "atenaParam",
        function(object) {
            cat(class(object), "object\n")
            cat(sprintf("# BAM files (%d): %s\n", length(object@bfl),
                        .pprintnames(names(object@bfl))))
            cat(sprintf("# features (%s length %d): %s\n",
                        class(object@features),
                        length(object@features),
                        ifelse(is.null(names(object@features)),
                                paste("on",
                                    .pprintnames(seqlevels(features(object)))),
                                .pprintnames(names(features(object))))))
            if (length(object@aggregateby) > 0)
                cat(sprintf("# aggregated by: %s\n",
                            paste(object@aggregateby, collapse=", ")))
            cat(sprintf("# %s; %s",
                        ifelse(object@singleEnd, "single-end", "paired-end"),
                        ifelse(object@ignoreStrand, "unstranded", "stranded")))
            if (!object@ignoreStrand)
                cat(sprintf(" (strandMode=%d)", object@strandMode))
            if (!object@singleEnd)
                cat(sprintf("; %s",
                        ifelse(object@fragments, "counting properly paired, same-strand pairs, singletons, reads with unmapped pairs and other fragments",
                                "counting properly paired reads")))
            cat("\n")
        })

#' @importFrom BiocParallel SerialParam bplapply
#' @importFrom S4Vectors DataFrame
#' @importFrom SummarizedExperiment SummarizedExperiment
#' @export
#' @aliases qtex
#' @aliases qtex,atenaParam-method
#' @rdname qtex
setMethod("qtex", "atenaParam",
        function(x, phenodata=NULL, mode=ovUnion, yieldSize=1e6L,
                 auxiliaryFeatures=FALSE,
                 BPPARAM=SerialParam(progressbar=TRUE)) {
            .checkPhenodata(phenodata, length(x@bfl))
            
            cnt <- bplapply(x@bfl, .qtex_atena, atpar=x, mode=x@ovMode,
                            yieldSize=yieldSize, BPPARAM=BPPARAM)
            cnt <- .processMultiNofeature(cnt, x)
            cnt <- do.call("cbind", cnt)
            colData <- .createColumnData(cnt, phenodata)
            colnames(cnt) <- rownames(colData)
            whnofeat <- grep(x=rownames(cnt), pattern = "no_feature")
            fnames <- rownames(cnt)[-whnofeat]
            if (!auxiliaryFeatures) {
                cnt <- cnt[-whnofeat, ]
                whnofeat <- integer(0)
            }
            features <- .consolidateFeatures(x, fnames, whnofeat)
            SummarizedExperiment(assays=list(counts=cnt),
                                 rowRanges=features,
                                 colData=colData)
        })


#' @importFrom stats setNames
#' @importFrom Rsamtools scanBamFlag ScanBamParam yieldSize yieldSize<-
#' @importFrom BiocGenerics path
#' @importFrom GenomicAlignments readGAlignments readGAlignmentsList
#' @importFrom GenomicAlignments readGAlignmentPairs
#' @importFrom S4Vectors Hits queryHits subjectHits
#' @importFrom Matrix Matrix t which
#' @importFrom sparseMatrixStats rowSums2 colSums2
#' @importFrom SQUAREM squarem
.qtex_atena <- function(bf, atpar, mode, yieldSize=1e6L) {
    mode <- match.fun(mode)
    readfun <- .getReadFunction(atpar@singleEnd, atpar@fragments)
    .checkreassignModes(atpar)
    sbflags <- .getScanBamFlag_ts(atpar@singleEnd, atpar@fragments)
    param <- ScanBamParam(flag=sbflags, what="flag", tag="AS")
    ## iste <- as.vector(attributes(atpar@features)$isTE[,1])
    iste <- mcols(features(atpar))$isTE
    if (any(duplicated(names(features(atpar)[iste]))))
        stop(".qtex_atena: duplicated names in TE annotations.")

    ov <- Hits(nLnode=0, nRnode=length(atpar@features), sort.by.query=TRUE)
    alnreadids <- character(0)
    asvalues <- integer()
    alen <- numeric(0)
    mreadlen <- numeric(0)
    salnmask <- logical(0)
    strand_arg <- "strandMode" %in% formalArgs(readfun)
    yieldSize(bf) <- yieldSize
    open(bf)
    while (length(alnreads <- do.call(readfun,
                                c(list(file = bf), list(param=param),
                                list(strandMode=atpar@strandMode)[strand_arg],
                                list(use.names=TRUE))))) {
        alnreads <- .matchSeqinfo(alnreads, atpar@features)
        alnreadids <- c(alnreadids, names(alnreads))
        asvalues <- c(asvalues, .getAlignmentASScoreTS(alnreads, tag = "AS"))
        readlen <- .getAlignmentLength(alnreads)
        alen <- c(alen, readlen)
        salnmask <- c(salnmask, any(.secondaryAlignmentMask(alnreads)))
        thisov <- mode(alnreads, atpar@features,
                       ignoreStrand=atpar@ignoreStrand, inter.feature=TRUE)
        ov <- .appendHits(ov, thisov)
    }
    # close(bf)
    on.exit(close(bf))
    .checkOvandsaln(ov, salnmask)
    maskuniqaln <- .getMaskUniqueAln(alnreadids)
    ## fetch all different read identifiers from the overlapping alignments
    readids <- unique(alnreadids[queryHits(ov)])
    ## Adding "no_feature" overlaps to 'ov'
    ov <- .getNoFeatureOv(maskuniqaln, ov, alnreadids)
    mt <- match(readids, alnreadids)
    readids <- unique(alnreadids[queryHits(ov)]) # updating 'readids'
    cntvec <- .atEMstep(atpar, alnreadids, readids, ov, asvalues,
                        iste, maskuniqaln, mt, alen)
    setNames(as.integer(cntvec), names(cntvec))
}


#' @importFrom S4Vectors Hits queryHits subjectHits
#' @importFrom Matrix Matrix t which
#' @importFrom SQUAREM squarem
#' @importClassesFrom Matrix lgCMatrix
.atEMstep <- function(atpar, alnreadids, readids, ov, asvalues,
                      iste, maskuniqaln, mt, alen) {
    ## initialize vector of counts derived from multi-mapping reads
    cntvec <- rep(0L, length(atpar@features) + 1)
    
    alnreadidx <- match(alnreadids, readids)
    rd_idx <- sort(unique(alnreadidx[queryHits(ov)]))
    
    ## fetch all different transcripts from the overlapping alignments
    tx_idx <- sort(unique(subjectHits(ov)))
    if (all(maskuniqaln)) {
        istex <- as.vector(iste[tx_idx])
    } else {
        istex <- as.vector(iste[tx_idx])[-length(tx_idx)] # removing no_feature
    }
    # asvalues <- (asvalues-min(asvalues)+1) / (max(asvalues)+1 - min(asvalues))
    asvalues <- .rescaleASat(asvalues, alen = alen)
    QmatAT <- .buildOvValuesMatrix(atpar, ov, asvalues, alnreadidx,
                                    rd_idx, tx_idx)
    
    if (!all(iste)) { ## Correcting for preference of reads to genes/TEs
        if (all(maskuniqaln)) {
          QmatAT <- .correctPreferenceTS(QmatAT, maskuniqaln, mt, istex)
          stopifnot(!any(rowSums2(QmatAT[,istex]) > 0 &
                           rowSums2(QmatAT[,!istex]) > 0))
        } else {
            QmatAT_nofeat <- QmatAT[,ncol(QmatAT), drop = FALSE]
            QmatAT <- .correctPreferenceTS(QmatAT[,-ncol(QmatAT)], maskuniqaln,
                                           mt, istex)
            stopifnot(!any(rowSums2(QmatAT[,istex]) > 0 &
                               rowSums2(QmatAT[,!istex]) > 0))
            QmatAT <- cbind(QmatAT, QmatAT_nofeat)
        }
    }
    
    nfeatures <- ncol(QmatAT)
    QmatAT <- .multiNofeature(QmatAT)
    nnofeat <- ncol(QmatAT) - nfeatures + 1
    cntvec <- c(cntvec, rep(0L, nnofeat - 1))
    tx_idx <- c(tx_idx, (max(tx_idx)+1):(max(tx_idx)+nnofeat -1))
    
    # --- EM-step --- 
    # QmatAT <- QmatAT / rowSums2(QmatAT)
    QmatAT@x <- QmatAT@x / rowSums2(QmatAT)[QmatAT@i +1]
    
    # Getting 'y' indicator of unique and multi-mapping status from QmatAT 
    # maskmulti <- ifelse(rowSums2(QmatAT > 0) == 1, 0, 1)
    maskmulti <- rowSums2(QmatAT > 0) > 1
    
    # π estimate proportional to relative abundances of features
    Pi <- colSums2(QmatAT / rowSums2(QmatAT))
    
    # θ proportional to relative abundances of features given by multimapping reads
    Theta <- colSums2(QmatAT[maskmulti,] / rowSums2(QmatAT[maskmulti,]))
    
    ## The SQUAREM algorithm to run the EM procedure
    a <- .processPriors(atpar@pi_prior, atpar, tx_idx, nnofeat) 
    b <- .processPriors(atpar@theta_prior, atpar, tx_idx, nnofeat) 

    Thetaenv <- new.env()
    assign("Theta", Theta, envir=Thetaenv)
    atres <- squarem(par=Pi, Thetaenv=Thetaenv, Q=QmatAT,maskmulti=maskmulti,
                    a=a, b=b, fixptfn=.atFixedPointFun,
                    control=list(tol=atpar@em_epsilon, maxiter=atpar@maxIter))
    Pi <- atres$par
    Pi[Pi < 0] <- 0 ## Pi estimates are sometimes negatively close to zero
    # --- end EM-step ---
    
    Theta <- get("Theta", envir=Thetaenv)
    X <- .tsEstep(QmatAT, Theta, maskmulti, Pi)
    cntvec <- .reassign(X, atpar@reassign_mode, atpar@conf_prob, cntvec, tx_idx)
    
    nofeat_names <- paste("no_feature", 1:nnofeat, sep = "")
    names(cntvec) <- c(names(atpar@features), nofeat_names)
    nofeat <- cntvec[nofeat_names]
    cntvec <- .tssummarizeCounts(cntvec[seq_along(atpar@features)], iste, atpar)
    cntvec <- c(cntvec, nofeat)
    cntvec
}


## private function .multiNofeature()
## Implements multiple '__no_feature' method
#' @importFrom stats aggregate
#' @importFrom Matrix Matrix
.multiNofeature <- function(QmatAT) { 
  ## Implementation of multiple no_feature method
  ovalnmat <- QmatAT > 0
  comb <-  aggregate(summary(ovalnmat)$j, by = list(ovalnmat@i + 1), 
                     FUN = function(x) paste0(x, collapse = "_"))
  combid <- match(comb$x, unique(comb$x)) # col id
  QmatAT[,ncol(QmatAT)] # values no_feature
  1:nrow(QmatAT) # row id
  ovmatnofeat <- Matrix(do.call("numeric", list(1)),
                        nrow=nrow(QmatAT), ncol=max(combid))
  ovmatnofeat[cbind(1:nrow(QmatAT), combid)] <- QmatAT[,ncol(QmatAT)]
  
  QmatAT_nofeatmulti <- cbind(QmatAT[,-ncol(QmatAT)], ovmatnofeat)
  return(QmatAT_nofeatmulti)
}

## private function .processMultiNofeature()
## Addresses different number of '__no_feature' features from the multiple
## '__no_feature' method
.processMultiNofeature <- function(cnt, x) {
  # adapted from cbindX() from gdata package
  if (length(cnt) > 1) {
    nrowsam <- lengths(cnt)
    maxi <- which.max(nrowsam)
    test <- nrowsam < nrowsam[maxi]
    for(i in seq_along(nrowsam)) {
      if(test[i]) {
        add <- rep(0, nrowsam[maxi] - nrowsam[i])
        cnt[[i]] <- c(cnt[[i]], add)
        names(cnt[[i]]) <- names(cnt[[maxi]])
      }
    }
  }
  return(cnt)
}



#' @importFrom matrixStats logSumExp
.rescaleASat <- function(asvalues, alen) {
  # asrescale <- rescale(asvalues, c(1, max(asvalues) - min(asvalues) + 1))
  # asrescale <- asrescale + alen
  # asrescale <- asrescale/max(asrescale)
  # asrescale <- expm1(asrescale*100)
  asrescale <- exp(asvalues - logSumExp(asvalues))
  asrescale
}


## private function .checkPriors()
## Checks that if length(prior) > 1, all features after .processFeatures()
## have a corresponding prior with the same name
.checkPriors <- function(namesfeatures, namespriora, namespriorb) {
  if (length(namespriora) > 1 & (!(all(namesfeatures %in% namespriora))))
    stop("after processing annotations (.processFeatures()), not all features have a pi_prior matching its name. Check that 'names()' of 'teFeatures' or 'geneFeatures' match those of 'pi_prior'")
  
  if (length(namespriorb) > 1 & (!(all(namesfeatures %in% namespriorb))))
    stop("after processing annotations (.processFeatures()), not all features have a theta_prior matching its name. Check that 'names()' of 'teFeatures' or 'geneFeatures' match those of 'theta_prior'")
}


## private function .processPriors()
## Assigns priors to "no_feature" and checks the length of the vector
## with priors
.processPriors <- function(prior, atpar, tx_idx, nnofeat) {
  if (length(prior) > 1) {
    featuresnames <- names(atpar@features)[tx_idx[1:(length(tx_idx)-nnofeat)]]
    prior <- prior[featuresnames]
    prior <- c(prior, rep(0, nnofeat)) # setting prior of 0 to __no_feature
  } else if (length(prior) == 1) {
    prior <- rep(prior, length(tx_idx))
  }
  
  if (any(is.na(prior))) {
    stop(sprintf("Prior not found for feature: %s", 
                 featuresnames[which(is.na(prior))]))
  }
  return(prior)
}


## private function .atEstep()
## E-step of the EM algorithm based on Telescope
.atEstep <- function(Q, Theta, maskmulti, Pi) {
  # X <- t(t(Q) * Pi)
  # # X[maskmulti, ] <- t(t(X[maskmulti, ]) * Theta)
  # # quicker computation of previous line
  # wh <- which(X*maskmulti > 0, arr.ind = TRUE)
  # X[wh] <- t(t(X[wh]) * Theta[wh[, "col"]])
  X <- Q
  j <- rep(seq_len(ncol(X)), diff(X@p))
  X@x <- X@x * Pi[j]
  #wh <- which(X@x * maskmulti[X@i + 1] > 0)
  whmulti <- as.integer(maskmulti[X@i + 1])
  wh <- which(X@x * whmulti > 0)
  j <- rep(seq_len(ncol(X)), diff(X@p))
  X@x[wh] <- X@x[wh] * Theta[j][wh]
  X <- X[rowSums2(X) > 0, , drop = FALSE]
  X <- X/rowSums2(X)
  X
}

## private function .atMstepPi()
## M-step of the EM algorithm based on Telescope
.atMstepPi <- function(X, a) {
  Pi <- (colSums2(X) + a)/(sum(X@x) + sum(a))
  Pi
}

## private function .atMstepTheta()
## Update the estimate of the MAP value of θ
.atMstepTheta <- function(X, maskmulti, b) {
  Theta <- ((colSums2(X[maskmulti, , drop=FALSE]) + b) / 
              (sum(maskmulti) + sum(b)))
}

## private function .atFixedPointFun()
## fixed point function of the EM algorithm based on Telescope
.atFixedPointFun <- function(Pi, Thetaenv, Q, maskmulti, a, b) {
  Theta <- get("Theta", envir=Thetaenv)
  X <- .atEstep(Q, Theta, maskmulti, Pi)
  Pi2 <- .atMstepPi(X, a)
  Theta2 <- .atMstepTheta (X, maskmulti, b)
  assign("Theta", Theta2, envir=Thetaenv)
  
  Pi2
}

## pretty-print names (private)
.pprintnames <- function(x) {
    y <- x
    if (length(x) > 2)
        y <- c(y[1], "...", y[length(y)])
    y <- paste(y, collapse=", ")
    y
}

## private function .checkPhenoData()

#' @importFrom S4Vectors nrow rownames
.checkPhenodata <- function(pdata, nr) {
    if (!is.null(pdata)) {
        if (nrow(pdata) != nr)
            stop("number of rows in 'phenodata' is different than the number of input BAM files in the input parameter object 'x'.")
        if (is.null(rownames(pdata)))
            stop("'phenodata' has no row names.")
    }
}

## private function .createColumnData()

#' @importFrom S4Vectors DataFrame
.createColumnData <- function(m, pdata) {
    colData <- DataFrame(row.names=gsub(".bam$", "", colnames(m)))
    if (!is.null(pdata))
        colData <- pdata
    
    colData
}

## private function .checkBamFileListArgs()
## adapted from GenomicAlignments/R/summarizeOverlaps-methods.R

#' @importFrom Rsamtools BamFileList asMates asMates<-
.checkBamFileListArgs <- function(bfl, singleEnd, fragments) {
    if (missing(bfl) || !class(bfl) %in% c("character", "BamFileList"))
        stop("argument 'bfl' should be either a string character vector of BAM file names or a 'BamFileList' object")
    
    if (length(bfl) == 0)
        stop("argument 'bfl' is empty")
  
    if (is.character(bfl)) {
        mask <- vapply(bfl, FUN=file.exists, FUN.VALUE=logical(1))
        if (any(!mask))
            stop(sprintf("The following input BAM files cannot be found:\n%s",
                         paste(paste("  ", bfl[!mask]), collapse="\n")))
    }
    
    if (!is(bfl, "BamFileList"))
        bfl <- BamFileList(bfl, asMates=!singleEnd)
    
    if (singleEnd) {
        if (all(isTRUE(asMates(bfl))))
            stop("cannot specify both 'singleEnd=TRUE' and 'asMates=TRUE'")
        # if (fragments)
        #     stop("when 'fragments=TRUE', 'singleEnd' should be FALSE")
    } else
        asMates(bfl) <- TRUE
    
    bfl
}

## private function .checkBamReadMapper()
## extracts the name of the read mapper software from one or more BAM files
## parameters: bamfiles - BAM file names

#' @importFrom Rsamtools scanBamHeader
.checkBamReadMapper <- function(bamfiles) {
    if (missing(bamfiles) || !"character" %in% class(bamfiles))
        stop("argument 'bamfiles' should be a string character vector of BAM file names")
    
    mask <- vapply(bamfiles, FUN = file.exists, FUN.VALUE = logical(1L))
    if (any(!mask))
        stop(sprintf("The following input BAM files cannot be found:\n%s",
                    paste(paste("  ", bamfiles[!mask]), collapse="\n")))
    
    hdr <- scanBamHeader(bamfiles)
    readaligner <- vapply(hdr, FUN = function(x) {
                            ra <- NA_character_
                            if (!is.null(x$text[["@PG"]])) {
                                pgstr <- x$text[["@PG"]]
                                mt <- gregexpr("^PN:", pgstr)
                                wh <- which(vapply(mt, FUN = function(x) x!=-1,
                                            FUN.VALUE = logical(1L)))
                                ra <- substr(pgstr[[wh]],
                                            attr(mt[[wh]], "match.length") + 1,
                                            100000L)
                            }
                            tolower(ra)
                    }, FUN.VALUE = character(1))
    readaligner <- readaligner[!duplicated(readaligner)]
    readaligner <- as.vector(readaligner[!is.na(readaligner)])
    if (length(readaligner) == 0)
        warning("no read aligner software information in BAM files.")
    if (any(readaligner[1] != readaligner))
        warning(sprintf("different read aligner information in BAM files. Assuming %s",
                        readaligner[1]))
    
    readaligner[1]
}

## private function .processFeatures()
## builds a single 'GRanges' or 'GRangesList' object from input TE and gene
## features.
## parameters: teFeatures - a 'GRanges' or 'GRangesList' object with
##                          TE annotations
##             teFeaturesobjname - the name of 'teFeatures'
##             geneFeatures - a 'GRanges' or 'GRangesList' object with
##                            gene annotations
##             geneFeaturesobjname - the name of 'geneFeatures'
##             aggregateby - names of metadata columns in 'teFeatures'
##                           to be used later for aggregating estimated
##                           counts.

#' @importFrom S4Vectors mcols Rle decode DataFrame
#' @importFrom Seqinfo seqlevels<- seqlevels
#' @importFrom GenomicRanges mcols<- mcols
.processFeatures <- function(teFeatures, teFeaturesobjname, geneFeatures,
                             geneFeaturesobjname, aggregateby,
                             aggregateexons) {
    
    if (missing(teFeatures))
        stop("missing 'teFeatures' argument.")
    
    if (!is(teFeatures, "GRanges") && !is(teFeatures, "GRangesList"))
        stop(sprintf("TE features object '%s' should be either a 'GRanges' or a 'GRangesList' object.",
                     teFeaturesobjname))
    
    if (is.null(names(teFeatures)) && length(aggregateby) == 0)
        stop(sprintf("the TE features object '%s' has no names and no aggregation metadata columns have been specified.",
                     teFeaturesobjname))
    
    mdteFeatures <- mcols(teFeatures)
    features <- NULL
    if (is(teFeatures, "GRangesList"))
        teFeatures <- unlist(teFeatures)
    
    if (length(aggregateby) > 0) {
        mask <- !aggregateby %in% colnames(mcols(teFeatures))
        if (any(mask)) {
            fstr <- sprintf(paste("%%s not in metadata columns of the TE",
                                  "features object %s."), teFeaturesobjname)
            stop(sprintf(fstr, paste(aggregateby[mask], collapse=", ")))
        }
        mask <- aggregateby %in% c("Status", "RelLength", "Class")
        if (any(mask))
            stop(sprintf("%s cannot be used to aggregate quantifications.",
                         paste(aggregateby[mask], collapse=", ")))

    }
    
    if (!is.null(geneFeatures)) {
        if (is(geneFeatures, "GRangesList"))
            geneFeatures <- unlist(geneFeatures)

        if (!"type" %in% colnames(mcols(geneFeatures)))
          mcols(geneFeatures)$type <- "exon"
        
        features <- .joinTEsGenes(teFeatures, geneFeatures, geneFeaturesobjname)
    } else {
        features <- teFeatures
        mcols(features)$isTE <- rep(TRUE, length(features))
    }
    
    iste <- decode(mcols(features)$isTE)
    
    if (!is.null(geneFeatures)) {
        if (!all(iste) && !is.null(mcols(geneFeatures)$type)) {
            if (is(features, "GRanges")) {
                iste <- aggregate(iste, by=list(names(features)), unique)
                features <- .groupGeneExons(features, aggregateexons)
                mtname <- match(names(features), iste$Group.1)
                iste <- iste[mtname,"x"]
                
            } else if (is(geneFeatures, "GRanges")) {
                ## when gene annotations were a GRanges but TE annotations
                ## a GRangesList, aggregate genes at the exon level only
                features_g <- unlist(features[which(!iste)])
                mcols(features_g)$isTE <- FALSE
                mcols(features_g)$type <- mcols(features)[which(!iste), "type"]
                features_g <- .groupGeneExons(features_g,
                                              aggregateexons)
                mcols(features_g)$isTE <- FALSE
                mcols(features_g)$type <- "exon"
                iste_g <- unlist(lapply(relist(iste[which(!iste)], features_g),
                                        function(x) x[1]),  use.names=FALSE)
                iste <- c(iste[which(iste)], iste_g)
                features <- c(features[which(iste)], features_g)
            }
        }
      
    } else if (aggregateexons && is(features, "GRanges") &&
               !is.null(names(features))) {
        iste <- aggregate(iste, by=list(names(features)), unique)
        features <- .groupGeneExons(features, aggregateexons)
        mtname <- match(names(features), iste$Group.1)
        iste <- iste[mtname, "x"]
    }
    if (!is.null(names(features))) {
      mdat <- DataFrame(ids=names(features), isTE=iste)
      mdteFeatures$ids <- rownames(mdteFeatures)
      mdat <- merge(mdteFeatures, mdat, all=TRUE)
      mt <- match(names(features), mdat$ids)
      stopifnot(all(!is.na(mt))) ## QC
      mcols(features) <- mdat[mt, -match("ids", colnames(mdat))]
    }

    features
}


## private function .groupGeneExons()
## groups exons from the same gene creating a 'GRangesList' object
.groupGeneExons <- function(features, aggregateexons) {
    if (aggregateexons) {
        if (!all(features$isTE) & !any(mcols(features)$type == "exon")) {
            stop(".groupGeneExons: no genes with value 'exon' in 'type' column of the metadata of the 'GRanges' or 'GRangesList' object with gene annotations.")
        }
        yesexon <- rep(TRUE, length(features))
        if (!is.null(mcols(features)$type)) {
            yesexon <- mcols(features)$type == "exon"
            yesexon[is.na(yesexon)] <- FALSE
        }
        features <- features[mcols(features)$isTE | yesexon]
        featuressplit <- split(x = features, f = names(features))
    } else {
        features_g <- features[!features$isTE]
        features_t <- features[features$isTE]
        features_t_grl <- split(x = features_t, f = seq_along(features_t))
        names(features_t_grl) <- names(features_t)
        if (!any(mcols(features_g)$type == "exon")) {
            stop(".groupGeneExons: no genes with value 'exon' in 'type' column of the metadata of the 'GRanges' or 'GRangesList' object with gene annotations.")
        }
        features_g <- features_g[mcols(features_g)$type == "exon"]
        featuressplit <- split(x = features_g, f = names(features_g))
        featuressplit <- c(features_t_grl, featuressplit)
    }
    featuressplit
}


## private function .consolidateFeatures()
## builds a 'GRanges' or 'GRangesList' object
## grouping TE features, if necessary, and
## adding gene features, if available.
## parameters: x - TEtranscriptsParam object
##             fnames - feature names vector to which
##                      consolidated features should match

#' @importFrom methods is
#' @importFrom S4Vectors split
#' @importFrom IRanges IRanges IRangesList CharacterList unique
#' @importFrom GenomicRanges GRangesList GRanges
.consolidateFeatures <- function(x, fnames, whnofeat=integer(0)) {
    
    if (length(whnofeat) > 0 && !is(x, "TelescopeParam") &&
        !is(x, "atenaParam"))
        stop(paste("internal error: call to .consolidateFeatures() with",
                   "a non-empty whnofeat and the wrong parameter class."))

    cfeatures <- features(x)
    if (length(x@aggregateby) > 0) {
        iste <- mcols(features(x))$isTE
        teFeatures <- features(x)
        if (!is.null(iste) && any(iste))
            teFeatures <- features(x)[iste]
    
        f <- .factoraggregateby(teFeatures, x@aggregateby)
        md <- mcols(teFeatures)
        if (all(c("Status", "RelLength", "Class") %in% colnames(md))) {
            astatus <- unique(CharacterList(split(md$Status, f)))
            if (max(lengths(astatus)) == 1)
              astatus <- unlist(astatus)
            arlen <- sapply(split(md$RelLength, f), mean)
            aclass <- unique(CharacterList(split(md$Class, f)))
            if (max(lengths(aclass)) == 1)
              aclass <- unlist(aclass)
            md <- DataFrame(Status=astatus, RelLength=arlen, Class=aclass)
        } else
            md <- DataFrame(matrix(character(0), nrow=length(unique(f))))
        
        md$isTE <- rep(TRUE, nrow(md))

        if (is(teFeatures, "GRangesList")) {
            f <- rep(f, times=lengths(teFeatures))
            teFeatures <- unlist(teFeatures)
        }
        cfeatures <- split(teFeatures, f)
        mcols(cfeatures) <- md
    
        if (!is.null(iste) && any(!iste)) {
            geneFeatures <- features(x)[!iste]
            cfeatures <- c(cfeatures, geneFeatures)
        }
        stopifnot(length(cfeatures) == length(fnames)) ## QC
    }
    
    stopifnot(length(cfeatures) == length(fnames)) ## QC
    mt <- match(fnames, names(cfeatures))
    stopifnot(all(!is.na(mt))) ## QC
    cfeatures <- cfeatures[mt]
    
    if (length(whnofeat) > 0) {
        nofeat_gr <- GRanges(seqnames="chrNofeature", 
                             ranges=IRanges(start=1, end=1),
                             isTE=FALSE)
        if (length(whnofeat) == 1) {
            names(nofeat_gr) <- "no_feature"
            
        } else {
            nofeat_gr <- rep(nofeat_gr, length(whnofeat))
            names(nofeat_gr) <- paste0("no_feature", seq_along(whnofeat))
        }
        
        seqlev <- unique(c(seqlevels(cfeatures), seqlevels(nofeat_gr)))
        seqlevels(cfeatures) <- seqlev
        seqlevels(nofeat_gr) <- seqlev
        lencfeatwonofeat <- length(cfeatures)
        cfeatures <- c(cfeatures, nofeat_gr)
        rng <- (lencfeatwonofeat+1):length(cfeatures)
        mcols(cfeatures)$isTE[rng] <- FALSE
    }
    cfeatures
}


## private function .factoraggregateby()
## builds a factor with as many values as the
## length of the input annotations in 'ann', where
## every value is made by pasting the columns in
## 'aggby' separated by ':'.
## parameters: ann - GRanges object with annotations
##             aggby - names of metadata columns in 'ann'
##                     to be pasted together

#' @importFrom GenomicRanges mcols
.factoraggregateby <- function(ann, aggby) {
    if (is(ann,"GRangesList")) {
        anngrl <- ann
        ann <- unlist(ann)
    }
    stopifnot(all(aggby %in% colnames(mcols(ann)))) ## QC
    if (length(aggby) == 1) {
        f <- mcols(ann)[, aggby]
    } else {
        spfstr <- paste(rep("%s", length(aggby)), collapse=":")
        f <- do.call("sprintf", c(spfstr, as.list(mcols(ann)[, aggby])))
    }
    
    if (exists("anngrl")) {
      # Using the aggby of the 1st GRanges in each element of the GRangesList
      f <- unlist(lapply(relist(f, anngrl), function(x) x[1]))
    }
    f
}

## private function .getReadFunction()
## borrowed from GenomicAlignments/R/summarizeOverlaps-methods.R
.getReadFunction <- function(singleEnd, fragments) {
    if (singleEnd) {
        FUN <- readGAlignments
    } else {
        if (fragments)
            FUN <- readGAlignmentsList
        else
            FUN <- readGAlignmentPairs
    }
    
    FUN
}

## private function .appendHits()
## appends the second Hits object to the end of the first one
## assuming they have identical right nodes

#' @importFrom S4Vectors nLnode nRnode isSorted from to Hits
.appendHits <- function(hits1, hits2) {
    stopifnot(nRnode(hits1) == nRnode(hits2))
    stopifnot(isSorted(from(hits1)) == isSorted(from(hits2)))
    hits <- c(Hits(from=from(hits1), to=to(hits1),
                    nLnode=nLnode(hits1)+nLnode(hits2),
                    nRnode=nRnode(hits1), sort.by.query=isSorted(from(hits1))),
            Hits(from=from(hits2)+nLnode(hits1), to=to(hits2),
                    nLnode=nLnode(hits1)+nLnode(hits2),
                    nRnode=nRnode(hits2), sort.by.query=isSorted(from(hits2))))
    hits
}


#' @importFrom Seqinfo seqlevels<- seqlevels
#' @importFrom GenomeInfoDb seqlevelsStyle seqlevelsStyle<-
#' @importFrom GenomicRanges mcols<-
.joinTEsGenes <- function(teFeatures, geneFeatures, geneFeaturesobjname) {
    
    if (!is(geneFeatures, "GRanges") && !is(geneFeatures, "GRangesList"))
        stop(sprintf("gene features object '%s' should be either a 'GRanges' or a 'GRangesList' object.",
                    geneFeaturesobjname))
    
    if (is.null(names(geneFeatures)))
        stop(sprintf("gene features object '%s' has no 'names()'",
                     geneFeaturesobjname))
    
    if (any(names(geneFeatures) %in% names(teFeatures)))
        stop("gene features have some common identifiers with the TE features.")
    
    if (length(geneFeatures) == 0)
        stop(sprintf("gene features object '%s' is empty.", geneFeaturesobjname))
    
    seqlevelsStyle(geneFeatures) <- seqlevelsStyle(teFeatures)[1]
    slev <- unique(c(seqlevels(teFeatures), seqlevels(geneFeatures)))
    seqlevels(teFeatures) <- slev
    seqlevels(geneFeatures) <- slev
    features <- c(teFeatures, geneFeatures)
    temask <- Rle(rep(FALSE, length(teFeatures) + length(geneFeatures)))
    temask[seq_along(teFeatures)] <- TRUE
    mcols(features)$isTE <- temask
    features
}

#' @importFrom S4Vectors queryHits subjectHits
#' @importFrom Matrix Matrix
.buildOvValuesMatrix <- function(x, ov, values, aridx, ridx, fidx) {
    stopifnot(class(values) %in% c("logical", "integer", "numeric")) ## QC
    ovmat <- Matrix(do.call(class(values), list(1)),
                    nrow=length(ridx), ncol=length(fidx))
    mt1 <- match(aridx[queryHits(ov)], ridx)
    mt2 <- match(subjectHits(ov), fidx)
    
    if (is(x, "TelescopeParam") | is(x, "atenaParam")) {
        mtov <- cbind(mt1, mt2)
        mtalign <- match(paste(mtov[,1],mtov[,2],sep = ":"),
                         unique(paste(mtov[,1],mtov[,2], sep = ":")))
        s <- split(x = values[queryHits(ov)], f = mtalign)
        if (is(x, "TelescopeParam")) {
          saln <- unlist(lapply(s, max), use.names = FALSE)
        } else {
          saln <- unlist(lapply(s, sum), use.names = FALSE)
        }
        values <- saln[mtalign]
    } else {
        values <- values[queryHits(ov)]
    }
    
    ovmat[cbind(mt1, mt2)] <- values
    
    ovmat
}


.getMaskUniqueAln <- function(alnreadids) {
    maskuniqaln <- !(duplicated(alnreadids) |
                         duplicated(alnreadids, fromLast = TRUE))
    maskuniqaln
}


.checkOvandsaln <- function(ov, salnmask) {
    if (length(ov) == 0) {
        stop(".qtex: no overlaps were found between reads and features")
    }
    if (!any(salnmask)) {
        warning("secondary alignments are not present in the SAM/BAM file. The quantification of features will proceed without taking into account overlaps of secondary alignments.")
    }
}

## private function .matchSeqinfo()
#' @importFrom Seqinfo seqlengths seqinfo seqinfo<- seqlevels genome genome<-
#' @importFrom GenomeInfoDb keepSeqlevels seqlevelsStyle seqlevelsStyle<-
.matchSeqinfo <- function(gal, features, verbose=TRUE) {
  stopifnot("GAlignments" %in% class(gal) ||
            "GAlignmentPairs" %in% class(gal) ||
            "GAlignmentsList" %in% class(gal) ||
            "GRanges" %in% class(features) || 
            "GRangesList" %in% class(features)) ## QC
  
  if (length(intersect(seqlevelsStyle(gal), seqlevelsStyle(features))) > 0)
    return(gal)

  seqlevelsStyle(gal) <- seqlevelsStyle(features)[1]
  slengal <- seqlengths(gal)
  slenf <- seqlengths(features)
  commonchr <- intersect(names(slengal), names(slenf))
  slengal <- slengal[commonchr]
  slenf <- slenf[commonchr]
  if (any(slengal != slenf)) {
    if (sum(slengal != slenf) == 1 && verbose) {
      message(sprintf(paste("Chromosome %s has different lengths",
                            "between the input BAM and the annotations",
                            "This chromosome will",
                            "be discarded from further analysis",
                            sep=" "),
                      paste(commonchr[which(slengal != slenf)],
                            collapse=", ")))
      
    } else if (verbose) {
      message(sprintf(paste("Chromosomes %s have different lengths",
                            "between the input BAM and the annotations",
                            "These chromosomes",
                            "will be discarded from further analysis",
                            sep=" "),
                      paste(commonchr[which(slengal != slenf)],
                            collapse=", ")))
    }
    if (sum(slengal == slenf) == 0)
      stop(paste("None of the chromosomes in the input BAM file has the",
                 "same length as the chromosomes in the input annotations.",
                 sep = " "))
    gal <- keepSeqlevels(gal, commonchr[slengal == slenf],
                         pruning.mode="coarse")
    commonchr <- commonchr[slengal == slenf]
  }
  
  ## set the seqinfo information to the one of the annotations
  mt <- match(commonchr, seqlevels(gal))
  seqinfo(gal, new2old=mt, pruning.mode="coarse") <- seqinfo(features)[commonchr]
  
  gal
}

library(Rsamtools)
library(Matrix)
library(sparseMatrixStats)

test_new_Telescope_functions <- function(){
  ####################################################################
  ######### Creating all params needed to check functions ############
  ####################################################################
  
  bamfiles <- list.files(system.file("extdata", package="atena"),
                         pattern="*.bam", full.names=TRUE)
  TE_annot <- readRDS(file = system.file("extdata", "Top28TEs.rds",
                                         package="atena"))
  tspar <- TelescopeParam(bamfiles, teFeatures = TE_annot,
                          singleEnd = TRUE, ignoreStrand=TRUE,
                          aggregateby = c("repName"))        
  mode <- ovUnion
  readfun <- atena:::.getReadFunction(tspar@singleEnd, tspar@fragments)
  sbflags <- scanBamFlag(isUnmappedQuery=FALSE,
                         isDuplicate=FALSE,
                         isNotPassingQualityControls=FALSE)
  param <- ScanBamParam(flag=sbflags, tag="AS")
  iste <- as.vector(attributes(tspar@features)$isTE[,1])
  if (any(duplicated(names(tspar@features[iste])))) {
    stop(".qtex_telescope: transposable element annotations do not contain unique names for each element")
  }
  ov <- Hits(nLnode=0, nRnode=length(tspar@features), sort.by.query=TRUE)
  alnreadids <- character(0)
  asvalues <- integer()
  mreadlen <- numeric(0)
  strand_arg <- "strandMode" %in% formalArgs(readfun)
  bf <- tspar@bfl[[1]]
  yieldSize <- 1e6L
  yieldSize(bf) <- yieldSize
  open(bf)
  while (length(alnreads <- do.call(readfun,
                                    c(list(file = bf), list(param=param),
                                      list(strandMode=tspar@strandMode)[strand_arg],
                                      list(use.names=TRUE))))) {
    alnreads <- atena:::.matchSeqinfo(alnreads, tspar@features)
    alnreadids <- c(alnreadids, names(alnreads))
    asvalues <- c(asvalues, mcols(alnreads)$AS)
    mreadlen <- median(width(ranges(alnreads)))
    thisov <- mode(alnreads, tspar@features,
                   ignoreStrand=tspar@ignoreStrand)
    ov <- atena:::.appendHits(ov, thisov)
  }
  on.exit(close(bf))
  maskuniqaln <- !(duplicated(alnreadids) | 
                     duplicated(alnreadids, fromLast = TRUE))
  readids <- unique(alnreadids[queryHits(ov)])
  ov <- atena:::.getNoFeatureOv(maskuniqaln, ov, alnreadids)
  mt <- match(readids, alnreadids)
  readids <- unique(alnreadids[queryHits(ov)]) # updating 'readids'
  cntvec <- rep(0L, length(tspar@features) + 1)
  alnreadidx <- match(alnreadids, readids)
  rd_idx <- sort(unique(alnreadidx[queryHits(ov)]))
  
  tx_idx <- sort(unique(subjectHits(ov)))
  istex <- as.vector(iste[tx_idx])[-length(tx_idx)]  # removing "no_feature"
  
  asvalues <- (asvalues-min(asvalues)+1) / (max(asvalues)+1 - min(asvalues))
  QmatTS <- atena:::.buildOvValuesMatrix(tspar, ov, asvalues, alnreadidx,
                                         rd_idx, tx_idx)
  QmatTS@x <- QmatTS@x / rowSums2(QmatTS)[QmatTS@i +1]
  maskmulti <- ifelse(rowSums2(QmatTS > 0) == 1, 0, 1)
  PiTS <- rep(1 / length(tx_idx), length(tx_idx))
  Theta <- rep(1 / length(tx_idx), length(tx_idx))
  
  ####################################################################
  ############################  TEST 1 ###############################
  ####################################################################
  
  
  old_tsEstep <- function() {
    X <- t(t(QmatTS) * PiTS)
    wh <- which(X*maskmulti > 0, arr.ind = TRUE)
    X[wh] <- t(t(X[wh]) * Theta[wh[, "col"]])
    X <- X[rowSums(X)>0,, drop=FALSE]
    X <- X / rowSums(X)
    X
  }
  
  new_tsEstep <- function() {
    X <- QmatTS
    j <- rep(1:ncol(X), diff(X@p))
    X@x <- X@x * PiTS[j]
    wh <- which(X@x * maskmulti[X@i + 1] > 0)
    X@x[wh] <- X@x[wh] * Theta[j][wh]
    X <- X[rowSums2(X) > 0, , drop = FALSE]
    X <- X/rowSums2(X)
    X
  }
  
  checkEquals(
    old_tsEstep(),
    new_tsEstep()
  )
  
  ####################################################################
  ############################  TEST 2 ###############################
  ####################################################################
  
  X <- new_tsEstep()
  maxbyrow <- rowMaxs(X)
  
  old_Xind <- function(){
    Xind <- (X / maxbyrow) == 1
  }
  
  new_Xind <- function(){
    Xind <- as(as(as(X, "lMatrix"), "generalMatrix"), "CsparseMatrix")
    Xind@x <- (X@x /maxbyrow[X@i+1]) == 1
    Xind
  }
  
  checkEquals(
    old_Xind(),
    new_Xind()
  )
}
  

library(GenomicRanges)
library(Rsamtools)
library(Matrix)
library(sparseMatrixStats)

test_new_TEtranscript_functions <- function(){
  
  ####################################################################
  ######### Creating all params needed to check functions ############
  ####################################################################
  
  bamfiles <- list.files(system.file("extdata", package="atena"),
                         pattern="*.bam", full.names=TRUE)
  TE_annot <- readRDS(file = system.file("extdata", "Top28TEs.rds",
                                         package="atena"))
  ttpar <- TEtranscriptsParam(bamfiles, teFeatures = TE_annot,
                              singleEnd = TRUE, ignoreStrand=TRUE,
                              aggregateby = c("repName"))        
  mode=match.fun(ovUnion)
  readfun <- atena:::.getReadFunction(ttpar@singleEnd, ttpar@fragments)
  sbflags <- scanBamFlag(isUnmappedQuery=FALSE, isDuplicate=FALSE,
                         isNotPassingQualityControls=FALSE)
  param <- ScanBamParam(flag=sbflags, tag="AS")
  ## iste <- as.vector(attributes(ttpar@features)$isTE[,1])
  iste <- mcols(features(ttpar))$isTE
  if (any(duplicated(names(features(ttpar)[iste])))) {
    stop(".qtex_tetranscripts: transposable element annotations
         do not contain unique names for each element")
  }
  ov <- Hits(nLnode=0, nRnode=length(ttpar@features), sort.by.query=TRUE)
  alnreadids <- character(0)
  avgreadlen <- integer(0)
  d <- numeric(0)
  strand_arg <- "strandMode" %in% formalArgs(readfun)
  yieldSize <- 1e6L
  bf <- ttpar@bfl[[1]]
  yieldSize(bf) <- yieldSize
  open(bf)
  while (length(alnreads <- do.call(readfun, 
                                    c(list(file = bf), list(param=param),
                                      list(strandMode=ttpar@strandMode)[strand_arg],
                                      list(use.names=TRUE))))) {
    alnreads <- atena:::.matchSeqinfo(alnreads, ttpar@features)
    avgreadlen <- c(avgreadlen, atena:::.getAveLen(ttpar, alnreads))
    if (ttpar@singleEnd == FALSE) {
      d <- c(d, abs(start(first(alnreads)) - start(second(alnreads))))
    }
    alnreadids <- c(alnreadids, names(alnreads))
    thisov <- mode(alnreads, ttpar@features,
                   ignoreStrand=ttpar@ignoreStrand)
    ov <- atena:::.appendHits(ov, thisov)
  }
  on.exit(close(bf))
  
  maskuniqaln <- !(duplicated(alnreadids) | duplicated(alnreadids, 
                                                       fromLast = TRUE))
  readids <- unique(alnreadids[queryHits(ov)])
  tx_idx <- sort(unique(subjectHits(ov)))
  
  ####################################################################
  ############################  TEST 1 ###############################
  ####################################################################
  
  old_buildOvAlignmentsMatrix <- function(ov, arids, rids, fidx){
    oamat <- Matrix(FALSE, nrow = length(rids), ncol = length(fidx))
    mt1 <- match(arids[queryHits(ov)], rids)
    mt2 <- match(subjectHits(ov), fidx)
    oamat[cbind(mt1, mt2)] <- TRUE
    oamat
  }
  
  new_buildOvAlignmentsMatrix <-  function(ov, arids, rids, fidx){
    mt1 <- match(arids[queryHits(ov)], rids)
    mt2 <- match(subjectHits(ov), fidx)
    positions <- cbind(mt1, mt2)
    oamat <- sparseMatrix(positions[,1], positions[,2], x=TRUE)
    oamat
  }
  
  checkEquals(old_buildOvAlignmentsMatrix(ov, alnreadids, readids, tx_idx),
              new_buildOvAlignmentsMatrix(ov, alnreadids, readids, tx_idx))
  
  ####################################################################
  ############################  TEST 2 ###############################
  ####################################################################
  
  ovalnmat <- atena:::.buildOvAlignmentsMatrix(ov, alnreadids, readids, tx_idx)
  mt <- match(readids, alnreadids)
  ## istex <- as.vector(iste[tx_idx])
  ovalnmat <- ovalnmat[!maskuniqaln[mt], iste]
  yesov <- rowSums2(ovalnmat)>0
  ovalnmat <- ovalnmat[yesov,]
  readids <- readids[!maskuniqaln[mt]][yesov]
  
  old_Qmat_generation <- function (){
    Qmat <- Matrix(0, nrow=length(readids), ncol=length(tx_idx[iste]),
                   dimnames=list(readids, NULL))
    Qmat[which(ovalnmat, arr.ind=TRUE)] <- 1
    Qmat
  }
  
  new_Qmat_generation <- function(){
    x <- as.integer(ovalnmat@x)
    Qmat <- sparseMatrix(i=ovalnmat@i, p=ovalnmat@p, x=x, index1=FALSE,
                         dimnames=list(readids, NULL))
  }
  
  checkEquals(old_Qmat_generation(), new_Qmat_generation(), 
              msg="Testing Qmat generation")
  
  ####################################################################
  ############################  TEST 3 ###############################
  ####################################################################
  
  Qmat <- new_Qmat_generation()
  
  old_divide_Qmat <- function(){
    Qmat / rowSums2(ovalnmat)
  }
  
  new_divide_Qmat <- function(){
    qmat <- Qmat
    qmat@x <- Qmat@x / rowSums2(ovalnmat)[Qmat@i +1]
    qmat
  }
  
  checkEquals(old_divide_Qmat(), new_divide_Qmat(),
              msg = "Testing dividing Qmat by its rowSums")
  
  ####################################################################
  ############################  TEST 4 ###############################
  ####################################################################
  
  Pi <- colSums2(Qmat)
  Pi <- Pi / sum(Pi)
  probmassbyread <- as.vector(ovalnmat %*% Pi)
  
  old_cntvecovtx <- function(){
    wh <- which(ovalnmat, arr.ind=TRUE)
    cntvecovtx <- rep(0, ncol(ovalnmat)) #cntvecovtx <- rep(0, length(tx_idx))
    x <- tapply(Pi[wh[, "col"]] / probmassbyread[wh[, "row"]], wh[, "col"],
                FUN=sum, na.rm=TRUE)
    cntvecovtx[as.integer(names(x))] <- x
    cntvecovtx
  }
  
  new_cntvecovtx <- function(){
    ovalnmat2 <- as(as(as(ovalnmat, "dMatrix"), "generalMatrix"), "CsparseMatrix")
    i <- ovalnmat2@i + 1
    ovalnmat2@x <- ovalnmat2@x / probmassbyread[i]
    j <- rep(1:ncol(ovalnmat2), diff(ovalnmat2@p))
    ovalnmat2@x <- ovalnmat2@x * Pi[j]
    cntvecovtx <- sparseMatrixStats::colSums2(ovalnmat2, na.rm=TRUE)
    cntvecovtx
  }
  
  checkEquals(old_cntvecovtx(), new_cntvecovtx(), 
              msg = "Testing generating 'cntvecovtx' vector")
  
  ####################################################################
  ############################  TEST 4 ###############################
  ####################################################################
  
  Q <- rsparsematrix(1000, 200, 0.2)
  Pi <- colSums2(Q)
  
  old_tsEstep <- function(){
    X <- t(t(Q) * Pi)
    X <- X[rowSums(X) > 0, , drop = FALSE]
    X <- X/rowSums(X)
    X
  }
  
  new_tsEstep <- function(){
    X <- Q
    j <- rep(1:ncol(X), diff(X@p))
    X@x <- X@x * Pi[j]
    X <- X[rowSums2(X)>0,, drop=FALSE]
    X <- X / rowSums2(X)
    X
  }
  
  checkEquals(old_tsEstep(), new_tsEstep(), 
              msg = "Testing .tsEstep function")
  
}

#' Get RepeatMasker UCSC annotations
#'
#' The \code{annotaTEs()} function fetches RepeatMasker UCSC transposable
#' element (TE) annotations using 
#' \link[AnnotationHub]{AnnotationHub} and parses them.
#' 
#' @param genome The genome version of the desired RepeatMasker annotations
#'        (e.g. "hg38").
#'
#' @param parsefun A \code{function} to parse the annotations:
#' \itemize{
#'   \item Function \code{rmskidentity} returns RepeatMasker annotations as
#'         present in \link[AnnotationHub]{AnnotationHub}, without
#'         processing them.
#'   \item Function \code{rmskbasicparser} parses annotations by removing
#'         low complexity regions, simple repeats, satellites, rRNA, scRNA,
#'         snRNA, srpRNA and tRNA. Also removes TEs
#'         with a strand different than "+" or "-". Modifies "repFamily" and
#'         "repClass" columns when a "?" is present or when they are defined
#'         as "Unknown" or "Other". Finally, assigns a unique id to each TE
#'         instance by adding the suffix "_dup" plus a number at the end of
#'         the "repName".
#'   \item Function \code{rmskatenaparser} parses RepeatMasker annotations 
#'         reconstructing fragmented TEs by assembling together fragments from
#'         the same TE that are close enough. For LTR class TEs, it tries to
#'         reconstruct full-length and partial TEs following the LTR - internal
#'         region - LTR structure. Input is a \code{GRanges} object and output
#'         is a \code{GRangesList} object.
#'   \item Function \code{OneCodeToFindThemAll} parses annotations following
#'         the 'One code to find them all' method by 
#'         \href{https://doi.org/10.1186/1759-8753-5-13}{(Bailly-Bechet et al. 2014)}. 
#'         Input is a \code{GRanges} object and output is a \code{GRangesList} 
#'         object.
#'   \item User-defined function. Input and output should be \code{GRanges}
#'         objects.
#' }
#' 
#' @param verbose (Default \code{TRUE}) Logical value indicating whether to
#'                report progress.
#'
#' @param AHid AnnotationHub unique identifier, of the form AH12345, of an
#'             object with TE annotations. This is an optional argument to
#'             specify a concrete AnnotationHub resource, for instance
#'             when more there is more than one RepeatMasker annotation
#'             available for a specific genome version. If \code{AHid} is
#'             not specified, the latest RepeatMasker annotation is be used.
#' 
#' @param ... Arguments passed to \code{parsefun}.
#'        
#'
#' @return A [`GRanges`][GenomicRanges::GRanges-class] object with
#'         transposable element annotations.
#'
#' @details
#' Given a specific genome version, the \code{annotaTEs()} function fetches
#' RepeatMasker annotations from UCSC Genome Browser using the 
#' \link[AnnotationHub]{AnnotationHub} package. Since RepeatMasker not only
#' provides TE annotations but also low complexity DNA sequences and other
#' types of repeats, a specific \code{parsefun} can be set to parse these
#' annotations (e.g. \code{rmskbasicparser} or a user-defined function). If no
#' parsing is required, \code{parsefun} can be set to \code{rmskidentity}.
#'
#' @seealso
#' \code{\link[AnnotationHub]{AnnotationHub}}
#' 
#'
#' @examples
#' rmskid <- annotaTEs(genome="hg19", parsefun=rmskidentity)
#' rmskid
#' 
#' 
#' @aliases annotaTEs
#' @rdname annotaTEs
#' @name annotaTEs
#' @importFrom AnnotationHub AnnotationHub query
#' @importFrom cli cli_alert_info cli_alert_success
#' @export
annotaTEs <- function(genome="hg38", parsefun=rmskidentity, verbose=TRUE,
                      AHid=NULL, ...) {
    parsefun <- match.fun(parsefun)

    if (verbose)
        cli_alert_info("Connecting to the AnnotationHub")
    suppressMessages(ah <- AnnotationHub())
    if (!is.null(AHid))
        qah <- ah[AHid]
    else
        suppressMessages(qah <- query(ah, c(genome, "RepeatMasker", "UCSC")))
    
    if (length(qah) == 0) {
        if (is.null(AHid))
            stop(sprintf("UCSC RepeatMasker tracks for genome %s not found",
                         genome))
        else
            stop(sprintf("AnnotationHub resource %s not found", AHid))
    } else if (length(qah) > 1) {
        if (verbose) {
            fstr <- paste("%d UCSC RepeatMasker tracks found for genome %s,",
                          "using the latest one")
            cli_alert_info(sprintf(fstr, length(qah), genome))
        }
        mt <- gregexpr(pattern=" \\([A-Za-z0-9]+\\) ", qah$title)
        qahdates <- substr(qah$title, unlist(mt)+2,
                             unlist(mt)+sapply(mt, attr, "match.length")-3)
        ## non-English locales do not identify, e.g., 1Sep2021, and give NAs
        ## set momentarily LC_TIME locale to 'C'
        lct <- Sys.getlocale("LC_TIME")
        Sys.setlocale("LC_TIME", "C")
        qahdates <- as.Date(paste0("1", qahdates), "%d%b%Y")
        Sys.setlocale("LC_TIME", lct)
        ## in case > 1 RM annotations available, pick the most recent one
        suppressMessages(qah <- qah[which.max(qahdates)])
    }

    if (verbose)
        cli_alert_info("Downloading annotations")
    id <- names(qah)
    suppressMessages(gr <- ah[[id]])
    if (verbose)
        cli_alert_info("Parsing annotations")
    annot <- parsefun(gr, ...)
    if (verbose)
        cli_alert_success("Annotations successfully downloaded and parsed")
    annot
}

#' Parser of RepeatMasker annotations
#' @param gr A [`GRanges`][GenomicRanges::GRanges-class] object with
#'           RepeatMasker annotations from \link[AnnotationHub]{AnnotationHub}
#'
#' @return A [`GRanges`][GenomicRanges::GRanges-class] object.
#'         
#' @details 
#' Parses annotations by removing low complexity regions, simple repeats,
#' satellites, rRNA, scRNA, snRNA, srpRNA and tRNA. Also removes TEs with a
#' strand different than "+" or "-". Modifies "repFamily" and
#' "repClass" columns when a "?" is present or when they are defined
#' as "Unknown" or "Other". Finally, assigns a unique id to each TE instance
#' by adding the suffix "_dup" plus a number at the end of the "repName".
#'
#' @examples
#' rmskba <- annotaTEs(genome="dm6", parsefun=rmskbasicparser)
#' rmskba
#' 
#' @aliases rmskbasicparser
#' @rdname rmskbasicparser
#' @name rmskbasicparser
#' @importFrom GenomicRanges strand
#' @export
rmskbasicparser <- function(gr) {
    to_discard <- c("Low_complexity", "Simple_repeat", "Satellite", "rRNA",
                    "scRNA", "snRNA", "srpRNA", "tRNA")
    gr <- gr[!(gr$repClass %in% to_discard),]
    gr <- gr[strand(gr) == "+" | strand(gr) == "-",]
    # removing '?' from name, class or family
    gr$repName <- gsub(pattern="?", x=gr$repName, replacement="", fixed =TRUE)
    gr$repClass <- gsub(pattern="?", x=gr$repClass, replacement="", fixed=TRUE)
    gr$repFamily <- gsub(pattern="?", x=gr$repFamily, replacement="", 
                         fixed =TRUE)
    gr <- unlist(split(gr, f=gr$repName))
    dupl_te <- duplicated(gr$repName) | duplicated(gr$repName, fromLast = TRUE)
    ndup <- lengths(split(gr, f=gr$repName))
    dup <- unlist(lapply(ndup, FUN=function(x) seq_len(x)))
    repName <- paste(gr$repName, "_dup", dup, sep="")
    repName[!dupl_te] <- gr$repName[!dupl_te]
    names(gr) <- repName
    gr <- sort(gr)

    i <- gr$repFamily %in% c("Unknown", "Other")
    if (any(i)) {
      gr$repFamily[i] <- gr$repName[i]
    }
    gr
}

#' Identity function for parsefun
#' @param gr A [`GRanges`][GenomicRanges::GRanges-class] object.
#'
#' @return A [`GRanges`][GenomicRanges::GRanges-class] object.
#'
#' @details 
#' Identity function: returns the [`GRanges`][GenomicRanges::GRanges-class]
#' object without any modification.
#' 
#' @examples
#' rmskid <- annotaTEs(genome="dm6", parsefun=rmskidentity)
#' rmskid
#'         
#' @aliases rmskidentity
#' @rdname rmskidentity
#' @name rmskidentity
#' @export
rmskidentity <- function(gr) {
  gr
}

#' OneCodeToFindThemAll parser of RepeatMasker annotations
#' @param gr A [`GRanges`][GenomicRanges::GRanges-class] object with
#' RepeatMasker annotations from \link[AnnotationHub]{AnnotationHub}
#' 
#' @param dictionary (Default NULL) When NULL, a dictionary is built based 
#' on names of repeats. If not, a data.frame with equivalences LTR - internal
#' regions created by the user, where first column should be the name of the
#' internal region and the second column should be the LTR(s). When more than
#' one LTR, these should be separated by ":".
#' 
#' @param fuzzy (Default FALSE) A logical; if TRUE, the search for equivalences
#' between internal parts and LTRs to reconstruct LTR class transposable 
#' elements is less stringent, allowing more matches between corresponding
#' subparts. This option can increase the proportion of false positives
#' (incorrectly reconstructed LTR class TEs).
#' 
#' @param strict (Default FALSE) A logical; if TRUE, the 80-80 rule is applied,
#' i.e. only copies with more than 80% identity to the reference
#' and more than 80 bp long are reported.
#' 
#' @param insert (Default -1) An integer. When \code{insert} < 0, two fragments
#' are assembled if the distance separating their furthest extremities is less 
#' than twice the reference length of the element. When \code{insert} > 0,
#' fragments are assembled if the distance between their closest extremities
#' is equal or less than \code{insert}. When \code{insert} = 0, two fragments
#' are assembled if they are in contact next to each other.
#' 
#' @param BPPARAM See \code{?\link[BiocParallel:bplapply]{bplapply}} in the 
#' BiocParallel package. Can be used to run calculations in parallel.
#'
#' @return A [`GRangesList`][GenomicRanges::GRangesList-class] object.
#'         
#' @details 
#' Implementation of One code to find them all 
#' \href{https://doi.org/10.1186/1759-8753-5-13}{(Bailly-Bechet et al. 2014)}.
#' Parses RepeatMasker annotations from UCSC by assembling together fragments
#' from the same transposable elemenet (TE) that are close enough (determined
#' by the \code{insert} parameter). For TEs from the LTR class, the parser
#' tries to reconstruct full-length, when possible, or partial TEs following
#' the LTR - internal region - LTR structure. Equivalences between internal
#' regions and flanking LTRs can be set by the user with the \code{dictionary}
#' parameter or can be obtained by the parser. In this last case, the 
#' \code{fuzzy} parameter determines the level of stringency when searching 
#' for LTR - internal region equivalences.
#'
#' @examples
#' \dontrun{
#' rmskoc <- annotaTEs(genome="dm6", parsefun=OneCodeToFindThemAll,
#'                     fuzzy=FALSE, strict=FALSE)
#' }
#' 
#' @references
#' Bailly-Bechet et al. "One code to find them all": a perl tool to 
#' conveniently parse RepeatMasker output files. Mobile DNA. 2014;5(1):1-15.
#' DOI: \url{https://doi.org/10.1186/1759-8753-5-13}
#' 
#' @aliases OneCodeToFindThemAll
#' @rdname OneCodeToFindThemAll
#' @name OneCodeToFindThemAll
#' @importFrom GenomicRanges strand width mcols "mcols<-"
#' @importFrom BiocParallel bplapply SerialParam
#' @importFrom Seqinfo seqnames
#' @importFrom S4Vectors runValue
#' @importFrom IRanges mean
#' @export
OneCodeToFindThemAll <- function(gr, dictionary=NULL, fuzzy=FALSE,
                                 strict=FALSE, insert=-1, 
                                 BPPARAM=SerialParam(progressbar=TRUE)) {

  if (!is.integer(insert) & !is.numeric(insert))
    stop("'insert' must be an integer value")
  
  if (!is(gr, "GRanges"))
    stop("'gr' should be a GRanges object with RepeatMasker annotations")
  
  if (length(gr) == 0)
    stop("'gr' is empty")
  
  if (!is.null(dictionary)) {
    if (!file.exists(dictionary))
        stop("'dictionary' should be NULL or specify the name of a dictionary",
        " file, see ?OneCodeToFindThemAll")
  }
  
  if (is.null(dictionary)) {
    inout <- .builDictionary(gr, fuzzy)
  } else {
    inside <- dictionary[, 2]
    names(inside) <- dictionary[, 1]
    outsidesp <- strsplit(dictionary[, 2], split=":")
    outside <- rep(dictionary[, 1], lengths(outsidesp))
    names(outside) <- unlist(outsidesp)
    inout <- list(inside = inside, outside=outside)
  }
  
  gr <- .filterNonTEs(gr)
  gr$repStart <- abs(gr$repStart)
  gr$repLeft <- abs(gr$repLeft)
  cons_length <- .getConsLength(gr)
  
  # Parsing by chromosome
  annrec <- bplapply(runValue(seqnames(gr)), .OCTFTA_parser, gr=gr, 
                     cons_length=cons_length, outside=inout$outside, 
                     inside=inout$inside, insert=insert, BPPARAM=BPPARAM)
  
  # 'annrec' is a list of GRangesLists, here we concatenate GRangesList to a
  # single GRangesList
  annrec <- do.call("c", annrec)
  
  # -- Filtering based on 'strict' option
  if (strict) {
    maxdiv <- mean(relist(mcols(unlist(annrec))$milliDiv, annrec)) < 200
    tokeep <- sum(width(annrec)) >= 80 & maxdiv
    annrec <- annrec[tokeep]
  }
  
  # -- Computation relative length --
  # for full-length and partial ERVs, and ERVs with only internal region, the
  # consensus length is considered that of the full-length TE: LTR + int + LTR
  nTEs <- suppressWarnings(do.call("rbind", strsplit(names(annrec), ".", 
                                                     fixed=TRUE)))[,1]
  mcols(annrec)$RelLength <- .getRelLength(nTEs, cons_length, inout, annrec)
  
  # -- Simplifying TE name: subfamily name + number of TE separated by "." --
  names(annrec) <- paste(nTEs, seq_along(annrec), sep = ".")
  
  # -- Including TE class name in GRangesList 
  mcols(annrec)$Class <- unlist(unique(relist(unlist(annrec)$repClass, annrec)))
  
  annrec
}

#' Getter functions of TE classes from parsed RepeatMasker annotations.
#'
#' @param annot A [`GRanges`][GenomicRanges::GRanges-class] or
#'              [`GRangesList`][GenomicRanges::GRangesList-class] object obtained
#'              with the function `annotaTES()`, using either [`OneCodeToFindThemAll`]
#'              or [`rmskatenaparser`] as RepeatMasker parser functions.
#'              Alternatively, if `annot` is a [`QuantifyParam`] or a
#'              [`SummarizedExperiment`][SummarizedExperiment::SummarizedExperiment-class]
#'              object produced by the `qtex()` function, this function will
#'              attempt to extract the corresponding annotations from inside
#'              those objects.
#' 
#' @param relLength (Default 0.9) Numeric value that can take values between 0
#'                  to 1. Sets the minimum relative length required for
#'                  features. Elements with a lower relative length than
#'                  \code{relLength} will be filtered. The relative length
#'                  used is the one obtained by \code{OneCodeToFindThemAll()}
#'                  or \code{rmskatenaparser()}.
#'                  (length of the reconstructed TE / length of the reference).
#' 
#' @param fullLength (Default TRUE) Logical value on whether reconstructed
#'                   full-length LTR TEs (elements with structure 
#'                   LTR - internal region - LTR) should be selected.
#' 
#' @param partial (Default FALSE) Logical value on whether partially
#'                reconstructed LTR TEs should be selected (structure LTR -
#'                internal region or internal region - LTR).
#'                               
#' @param soloLTR (Default FALSE) Logical value on whether solo LTRs should be
#'                selected. Note that only fragments unambiguously identified
#'                as LTRs thanks to the identification of their equivalent
#'                internal region are considered as LTRs.
#' 
#' @param otherLTR (Default FALSE) Logical value on whether other TEs from the
#'                 LTR class, not included in any of the previous three
#'                 categories, should be selected. These include TEs from LTR
#'                 class that cannot be unambiguously identified as LTR o
#'                 internal region, and thus cannot be reconstructed into
#'                 partial or full-length elements; as well as solo internal
#'                 regions. 
#'
#' @param returnMask (Default FALSE) Logical value indicating whether a subset
#'                   of the input annotations should be returned (default) or
#'                   a logical mask of the same length as the input annotations
#'                   where \code{TRUE} values indicate what annotations belong
#'                   to the TE class we want to obtain with the getter function.
#'
#' @return A [`GRangesList`][GenomicRanges::GRangesList-class] object with
#'         annotations from class corresponding to the getter function (LTRs,
#'         LINEs, SINEs or DNA transposons).
#'         
#' @details 
#' Retrieves annotations from the TE class corresponding to the getter function,
#' using RepeatMasker annotations after parsing them with the
#' \code{OneCodeToFindThemAll()} or \code{rmskatenaparser()} function. The
#' \code{relLength} parameter can be used to filter out elements with a lower
#' relative length. Further parameters can be used to fine-tune the type of
#' elements to be reported.
#'
#' @examples
#' rmskat <- annotaTEs(genome="dm6", parsefun=rmskatenaparser,
#'                     strict=FALSE)
#'
#' rmskatLTR <- getLTRs(rmskat, relLength=0.95, fullLength=TRUE,
#'                      partial=TRUE)
#' rmskatLTR
#'  
#' @aliases getLTRs
#' @rdname annotaTEsGetters
#' @name annotateTEsGetters
#' @importFrom GenomicRanges mcols
#' @export
getLTRs <- function(annot, relLength=0.9, fullLength=TRUE, partial=FALSE,
                    soloLTR=FALSE, otherLTR=FALSE, returnMask=FALSE) {
    
    annot <- .fetchAnnotationsFromObject(annot)

    cnames <- c("Status", "RelLength", "Class")
    mask <- cnames %in% colnames(mcols(annot))
    if (any(!mask))
        stop(sprintf("column(s) %s missing from input annotations.",
                     paste(cnames[!mask], collapse=", ")))
    if (!any(c(fullLength, partial, soloLTR, otherLTR)))
        stop("at least one of these arguments should be TRUE: 'fullLength',",
             " 'partial', 'soloLTR', 'otherLTR'")

    mask <- rep(TRUE, length(annot))
    if ("isTE" %in% colnames(mcols(annot))) {
        mask <- mcols(annot)$isTE
        annot <- annot[mcols(annot)$isTE]
    }
  
    keep <- mcols(annot)$Class == "LTR" & 
            mcols(annot)$RelLength > relLength
    LTRtypes <- c("full-lengthLTR"[fullLength], "partialLTR_up"[partial], 
                  "partialLTR_down"[partial], "LTR"[soloLTR],
                  "int"[otherLTR], "noLTR"[otherLTR])
    keep2 <- mcols(annot)$Status %in% LTRtypes

    res <- mask
    if ("isTE" %in% colnames(mcols(annot))) {
        res[mask] <- keep & keep2
        mcols(annot)$isTE <- NULL
    } else
        res <- keep & keep2

    if (!returnMask)
        res <- annot[keep & keep2]

    res
}

#' @examples
#' rmskat_line <- getLINEs(rmskat, relLength=0.95)
#'  
#' @aliases getLINEs
#' @rdname annotaTEsGetters
#' @name annotateTEsGetters
#' @importFrom GenomicRanges mcols
#' @export
getLINEs <- function(annot, relLength=0.9, returnMask=FALSE) {

    annot <- .fetchAnnotationsFromObject(annot)

    cnames <- c("RelLength", "Class")
    mask <- cnames %in% colnames(mcols(annot))
    if (any(!mask))
        stop(sprintf("column(s) %s missing from input annotations.",
                     paste(cnames[!mask], collapse=", ")))

    mask <- rep(TRUE, length(annot))
    if ("isTE" %in% colnames(mcols(annot))) {
        mask <- mcols(annot)$isTE
        annot <- annot[mcols(annot)$isTE]
    }
  
    keep <- mcols(annot)$Class == "LINE" &
            mcols(annot)$RelLength > relLength

    res <- mask
    if ("isTE" %in% colnames(mcols(annot))) {
        res[mask] <- keep
        mcols(annot)$isTE <- NULL
    } else
        res <- keep

    if (!returnMask)
        res <- annot[keep]

    res
}

#' @examples
#' rmskat_sine <- getSINEs(rmskat, relLength=0.95)
#'  
#' @aliases getSINEs
#' @rdname annotaTEsGetters
#' @name annotateTEsGetters
#' @importFrom GenomicRanges mcols
#' @export
getSINEs <- function(annot, relLength=0.9, returnMask=FALSE) {

    annot <- .fetchAnnotationsFromObject(annot)

    cnames <- c("RelLength", "Class")
    mask <- cnames %in% colnames(mcols(annot))
    if (any(!mask))
        stop(sprintf("column(s) %s missing from input annotations.",
                     paste(cnames[!mask], collapse=", ")))

    mask <- rep(TRUE, length(annot))
    if ("isTE" %in% colnames(mcols(annot))) {
        mask <- mcols(annot)$isTE
        annot <- annot[mcols(annot)$isTE]
    }
  
    keep <- mcols(annot)$Class == "SINE" &
            mcols(annot)$RelLength > relLength

    res <- mask
    if ("isTE" %in% colnames(mcols(annot))) {
        res[mask] <- keep
        mcols(annot)$isTE <- NULL
    } else
        res <- keep

    if (!returnMask)
        res <- annot[keep]

    res
}

#' @examples
#' rmskat_DNAtrans <- getDNAtransposons(rmskat, relLength=0.95)
#'  
#' @aliases getDNAtransposons
#' @rdname annotaTEsGetters
#' @name annotateTEsGetters
#' @importFrom GenomicRanges mcols
#' @export
getDNAtransposons <- function(annot, relLength=0.9, returnMask=FALSE) {

    annot <- .fetchAnnotationsFromObject(annot)

    cnames <- c("RelLength", "Class")
    mask <- cnames %in% colnames(mcols(annot))
    if (any(!mask))
        stop(sprintf("column(s) %s missing from input annotations.",
                     paste(cnames[!mask], collapse=", ")))

    mask <- rep(TRUE, length(annot))
    if ("isTE" %in% colnames(mcols(annot))) {
        mask <- mcols(annot)$isTE
        annot <- annot[mcols(annot)$isTE]
    }
  
    keep <- mcols(annot)$Class == "DNA" & 
            mcols(annot)$RelLength > relLength

    res <- mask
    if ("isTE" %in% colnames(mcols(annot))) {
        res[mask] <- keep
        mcols(annot)$isTE <- NULL
    } else
        res <- keep

    if (!returnMask)
        res <- annot[keep]

    res
}

#' @importFrom MatrixGenerics rowRanges
.fetchAnnotationsFromObject <- function(object) {
    annot <- object

    if (is(object, "QuantifyParam"))
        annot <- features(object)
    else if (is(object, "SummarizedExperiment"))
        annot <- rowRanges(object)
    else if (!is(object, "GRangesList") && !is(object, "GRanges"))
        stop(paste("input annotations should be either a 'GenomicRanges',",
                   "'QuantifyParam' or a 'SummarizedExperiment' object."))
    annot
}

#' @importFrom stats setNames
.builDictionary <- function(gr, fuzzy=FALSE) {
  annltr <- gr[gr$repClass=="LTR"]
  # Possible internal id
  intid <- c("int","in", "i")
  intid <- c(paste0("-", intid), paste0("_", intid))
  internal <- grepl(pattern = paste(intid,collapse="|"), annltr$repName, 
                    ignore.case = TRUE)
  ltr <- grepl(pattern = "LTR", annltr$repName, ignore.case = TRUE)
  ltr[internal] <- FALSE # preference goes to internal regions
  # not internal and not ltr: undecided (it could be int or LTR)
  undecid <- !internal & !ltr
  internal[undecid] <- TRUE
  ltr[undecid] <- TRUE
  
  intuniq <- unique(annltr$repName[internal])
  ltruniq <- unique(annltr$repName[ltr])
  
  inside <- character(length(intuniq))
  names(inside) <- intuniq
  outside <- character(length(ltruniq))
  names(outside) <- ltruniq
  
  inout <- .findingEquivalences(ltruniq, intuniq, intid, outside, inside, fuzzy)
  
  namerem <- inout$inside[(names(inout$inside) %in% names(inout$outside))]
  whrem <- !(names(inout$inside) %in% namerem[grep(pattern = "[-_]$",namerem)])
  inout$inside <- inout$inside[whrem]
  whrem <- !(names(inout$outside) %in% namerem[-grep(pattern = "[-_]$",namerem)])
  inout$outside <- inout$outside[whrem]
  
  # When the same LTR is assigned to more than 1 different internal region,
  # the quivalence is added to outside
  whadd <- inout$inside[!(names(inout$inside) %in% inout$outside)]
  inout$outside <- c(inout$outside, setNames(names(whadd), nm = whadd))
  
  # Adding NAs when there is no equivalence for an int region and a LTR
  noteq <- unique(annltr$repName[!(annltr$repName %in% c(inout$inside, inout$outside))])
  noteq <- setNames(rep(NA,length(noteq)), noteq)
  inout$inside <- c(inout$inside, noteq)
  
  inout
}

.findingEquivalences <- function(ltruniq, intuniq, intid, outside, 
                                 inside, fuzzy) {
  # Elements containing "LTR" should have an internal part with a corresponding 
  # I,IN, INT or int suffix
  for (ltr in ltruniq) {
    for (id in c(intid, toupper(intid))) { # c("I","IN", "INT","i","in","int")
      tmpint <- gsub(pattern = "[-_]LTR|LTR", id, ltr)
      if (any(intuniq == tmpint) & ltr != tmpint) {
        inside[tmpint] <- ltr
        outside[ltr] <- tmpint
      }
    }
  }
  # Elements not containing LTR could be matched with a following 
  # "-int", "_int", "_I", "_IN", etc... 
  for (ltr in ltruniq) {
    for (id in c(intid, toupper(intid))) {
      tmpint <- paste0(ltr, id)
      if (any(intuniq == tmpint) & ltr != tmpint) {
        inside[tmpint] <- ltr
        outside[ltr] <- tmpint
      }
    }
  }
  
  ### Combination of the two previous steps
  for (ltr in ltruniq) {
    for (id in c(intid, toupper(intid))) {
      for (idsub in c("I","IN", "INT","i","in","int")) {
        tmpint <- gsub(pattern = "LTR", idsub, ltr)
        tmpint <- paste0(tmpint, id)
        if (any(intuniq == tmpint) & ltr != tmpint) {
          inside[tmpint] <- ltr
          outside[ltr] <- tmpint
        }
      }
    }
  }
  inside <- inside[inside != ""]
  outside <- outside[outside != ""]
  
  # Simplify element names by erasing LTR, I, IN, INT, or symbols like -_, etc...
  elem_liste_ltr <- ltruniq
  ltrpat <- grep(pattern = "^LTR", ltruniq)
  elem_liste_ltr[-ltrpat] <- gsub(pattern = "LTR", ltruniq[-ltrpat], 
                                  replacement = "")
  elem_liste_ltr <- gsub(pattern = "[-_]", replacement = "", elem_liste_ltr)
  names(elem_liste_ltr) <- ltruniq
  elem_liste_int <- gsub(pattern = "int|in", replacement = "", 
                         intuniq, ignore.case = TRUE)
  elem_liste_int <- gsub(pattern = "[-_]I", replacement = "", elem_liste_int, 
                         ignore.case = FALSE)
  elem_liste_int <- gsub(pattern = "[-_]", replacement = "", elem_liste_int)
  names(elem_liste_int) <- intuniq
  whr <- (intuniq %in% ltruniq) & grepl(pattern = "[-_]$", intuniq)
  elem_liste_int <- elem_liste_int[!whr]
  inout <- .simplifyNameSearch(elem_liste_ltr, elem_liste_int, inside, outside)
  
  if (fuzzy) {
    inout <- .getFuzzyEquivalences(inout, elem_liste_int, elem_liste_ltr)
  }
  
  inout
}

.simplifyNameSearch <- function(elem_liste_ltr, elem_liste_int, 
                                inside, outside) {
  ### And we try to match the new names
  el1all <- names(elem_liste_int[is.na(outside[elem_liste_int])])
  el2all <- names(elem_liste_ltr[is.na(outside[elem_liste_ltr]) &
                                   is.na(inside[elem_liste_ltr])])
  for (el1 in el1all) {
    for (el2 in el2all) {
      if (el1 == el2) {
        next
      }
      if (elem_liste_ltr[el2] == elem_liste_int[el1]) {
        if (is.na(inside[el1])) {
          inside[el1] <- el2
        } else {
          inside[el1] <- paste(unique(c(inside[el1], el2)), collapse =":")
        }
        # outside[el2] <- paste(unique(c(inside[el2], el1)), collapse =":")
        if (is.na(outside[el2])) {
          outside[el2] <- el1
        } else {
          outside[el2] <- paste(unique(c(outside[el2], el1)), collapse =":")
        }
      }
    }
  }
  list(inside = inside, outside = outside)
}


.getFuzzyEquivalences <- function(inout, elem_liste_int, elem_liste_ltr) {
  outside <- inout$outside
  inside <- inout$inside
  patterns <- c("[a-z]$","[0-9]$","[a-z0-9][a-z0-9]$")
  for (pat in patterns) {
    for (el1 in sort(names(elem_liste_int))) {
      if (!is.na(outside[el1])) {
        next
      }
      for (el2 in sort(names(elem_liste_ltr))) {
        if ((el1 == el2) | !is.na(outside[el2]) | !is.na(inside[el2])) {
          next
        }
        el1_int <- gsub(pat, "", elem_liste_int[el1], ignore.case = TRUE)
        el2_ltr <- gsub(pat, "", elem_liste_ltr[el2], ignore.case = TRUE)
        if (elem_liste_ltr[el2] == el1_int | elem_liste_int[el1] == el2_ltr) {
          if (is.na(inside[el1])) {
            inside[el1] <- el2
          } else {
            inside[el1] <- paste(unique(c(inside[el1], el2)), collapse =":")
          }
          # outside[el2] <- paste(unique(c(inside[el2], el1)), collapse =":")
          if (is.na(outside[el2])) {
            outside[el2] <- el1
          } else {
            outside[el2] <- paste(unique(c(outside[el2], el1)), collapse =":")
          }
        }
      }
    }
  }
  list(inside = inside, outside = outside)
}


.filterNonTEs <- function(gr) {
  rc <- gr$repClass == "RC" & gr$repFamily == "Helitron"
  gr$repClass[rc] <- "DNA"
  gr$repFamily[rc] <- "RC"
  gr <- gr[gr$repClass %in% c("DNA","SINE","LINE","LTR")]
  gr
}

#' @importFrom GenomicRanges strand
.getConsLength <- function(gr) {
  # Computing length of consensus sequences for each element
  remain <- integer(length(gr))
  remain[as(strand(gr) == "+", "logical")] <- gr$repLeft[as(strand(gr) == "+", "logical")]
  remain[as(strand(gr) == "-", "logical")] <- gr$repStart[as(strand(gr) == "-", "logical")]
  lt <- gr$repEnd + remain
  lt <- split(lt, gr$repName)
  cons_length <- lapply(lt, function(x) as.integer(names(which.max(table(x)))))
  unlist(cons_length)
}

#' @importFrom GenomicRanges seqnames strand
.OCTFTA_parser <- function(chr, gr, cons_length, outside, inside, insert) {
  
  annchrp <- gr[seqnames(gr) == chr & strand(gr) == "+"]
  annchrn <- gr[seqnames(gr) == chr & strand(gr) == "-"]
  if (any(strand(gr) == "*")) { # undefined strand
    annchru <- gr[seqnames(gr) == chr & strand(gr) == "*"]
  }
  
  # Creating GRangesList for the specific chromosome and strand, splitting
  # features based on repName
  annchrp <- split(annchrp, annchrp$repName)
  annchrn <- split(annchrn, annchrn$repName)
  if (any(strand(gr) == "*")) { # undefined strand
    annchru <- split(annchru, annchru$repName)
  }
  
  # checking family is the same for all features with same repName
  # stopifnot(all(unlist(lapply(annchrp, function(gr) length(unique(gr$family)) == 1))))
  # stopifnot(all(unlist(lapply(annchrn, function(gr) length(unique(gr$family)) == 1))))
  
  # Calling .reconstructTEs() to reconstruct TEs
  if (length(annchrp) > 0) {
    annchrp_rec <- .reconstructTEs(annchr=annchrp, outside, inside, 
                                   cons_length, insert, minusStrand=FALSE)
  } else {
    annchrp_rec <- annchrp
  }
  if (length(annchrn) > 0) {
    annchrn_rec <- .reconstructTEs(annchr=annchrn, outside, inside,
                                   cons_length, insert, minusStrand=TRUE)
  } else {
    annchrn_rec <- annchrn
  }
  annrec <- c(annchrp_rec, annchrn_rec)
  
  if (any(strand(gr) == "*")) {
    if (length(annchru) > 0) {
      annchru_rec <- .reconstructTEs(annchr=annchru, outside, inside, 
                                     cons_length, insert, minusStrand=FALSE)
    } else {
      annchru_rec <- annchru
    }
    annrec <- c(annrec, annchru_rec)
  }
  
  annrec
}

#' @importFrom GenomicRanges start mcols "mcols<-"
.reconstructTEs <- function(annchr, outside, inside, cons_length, insert, 
                            minusStrand=FALSE) {
  # For LTRs and internal regions, we check for the presence (in the same 
  # strand) of internal regions and LTRs, respectively, between elements with 
  # the same repName. To do so, we divide features with the same repName
  # (e.g. LTR23) using the start of the corresponding internal region
  # (LTR23-int) as flanking regions.
  
  # Selecting only cases where both the ltr and internal region are in the
  # chromosome
  outsidechr <- outside[(outside %in% names(annchr)) &
                          (names(outside) %in% names(annchr))]
  whint <- which(names(annchr) %in% outsidechr)
  whltr <- which(names(annchr) %in% names(outsidechr))
  # We extract coordinates of int and corresponding ltr (same order)
  annchrint <- annchr[whint]
  annchrltr <- annchr[whltr]
  
  if (length(annchrint) > 0 & length(annchrltr) > 0) {
    splitf2 <- .splitNonContinous(annchrint, annchrltr, outsidechr)
    annchrltrsp <- split(unlist(annchrltr,use.names=FALSE), splitf2$splitfltr2)
    annchrintsp <- split(unlist(annchrint,use.names=FALSE), splitf2$splitfint2)
    annchr <- annchr[-c(whint, whltr)]
    
    # We keep ltr and int separated from the rest of TEs to later perform the
    # reconstruction of full-length or partial ERVs
    
    # Now, features with the same repName are merged together if: they are on
    # the same strand, if the end of the 2nd is closer than 2xconsensus length
    # to the start of the 1st, the start/end of the 2nd is after start of 1st
    # (to avoid merging elements completely contained inside another element),
    # the start/end in the consensus sequence of the 2nd is after the one of 
    # the 1st.
    annchrltrsp2 <- .mergeCloseFeatures(annchrltrsp, cons_length, insert,
                                        minusStrand=minusStrand)
    annchrintsp2 <- .mergeCloseFeatures(annchrintsp, cons_length, insert,
                                        minusStrand=minusStrand)
    
    mcols(annchrltrsp2)$Status <- "LTR"
    mcols(annchrintsp2)$Status <- "int"
  }
  
  # if 'annchrltrsp2' or 'annchrintsp2' are empty, 'annchr2' contains all
  # TEs including LTR and int regions. If 'annchrltrsp2' or 'annchrintsp2' are
  # not empty, still 'annchr' contains LTR and int regions for which no
  # equivalences have been identified (not present in 'outside')
  annchr2 <- .mergeCloseFeatures(annchr, cons_length, insert,
                                 minusStrand=minusStrand)
  
  # Note: All features not identified as LTR or int (from the previously built
  # dictionary with equivalences: 'outside'), are identinfied as "noLTR",
  # however LTRs and int are expected to be inside 'annchr2' since the
  # dictionary 'outside' does not identify all LTR and int regions.
  if (length(annchr2) > 0)
    mcols(annchr2)$Status <- "noLTR"
  
  # ---- Reconstructing full-length and partial ERVs ----
  if (length(annchrint) > 0 & length(annchrltr) > 0) {
    anngrlERVs <- .reconstructERVs(annchrltrsp2, annchrintsp2, outside, inside,
                                   cons_length)
    anngrl <- c(annchr2, anngrlERVs)
  } else {
    anngrl <- annchr2
  }
  opos <- order(min(start(anngrl)), decreasing = FALSE)
  anngrl[opos]
}

#' @importFrom GenomicRanges start mcols "mcols<-"
#' @importFrom Matrix colSums rowSums
.splitNonContinous <- function(annchrint, annchrltr, outsidechr) {
  annchrint2 <- annchrint[outsidechr[names(annchrltr)]]
  annchrltr2 <- split(unlist(annchrltr), 
                      rep(outsidechr[names(annchrltr)], lengths(annchrltr)))
  
  posc_int <- mapply(function(ltrname,intname) 
    outer(start(annchrltr2[[ltrname]]), start(annchrint[[intname]]), "<"),
    ltrname = names(annchrltr2), intname = names(annchrint),
    SIMPLIFY = FALSE)
  
  posc_ltr <- mapply(function(ltrname,intname) 
    outer(start(annchrltr[[ltrname]]), start(annchrint2[[intname]]), "<"),
    ltrname = names(annchrltr), intname = names(annchrint2),
    SIMPLIFY = FALSE)
  
  # 'posc' is a list of matrices in which ltr are rows and int regions are 
  # columns the matrices are TRUE/FALSE depending if the start position of
  # a ltr (row) is smaller than the start position of a int (col)
  # Cases where there is only 1 ltr or internal region in the chr are well
  # addressed
  
  # Creating a factor to split ltr and int according to their position in
  # relation to their equivalent int and ltr, respectively.
  splitfltr <- lapply(posc_ltr, function(x) rowSums(x))
  splitfint <- lapply(posc_int, function(x) colSums(x))
  splitfltr2 <- as.factor(paste(rep(names(splitfltr), lengths(splitfltr)),
                                unlist(splitfltr), sep ="."))
  splitfint2 <- as.factor(paste(rep(names(splitfint), lengths(splitfint)),
                                unlist(splitfint), sep ="."))
  list(splitfltr2 = splitfltr2, splitfint2 = splitfint2)
}

#' @importFrom GenomicRanges GRangesList start width strand end reduce
.mergeCloseFeatures <- function(annchr, cons_length, insert, minusStrand=FALSE) {
  if (insert < 0) {
    # Which elements are separated by less than twice the consensus length?
    annchrwhfirst <- as(lapply(annchr, function(x) x[-1]),"GRangesList")
    stopifnot(identical(names(annchrwhfirst), names(annchr)))
    conslen <- cons_length[do.call("rbind", strsplit(names(annchrwhfirst), 
                                                split = ".", fixed=TRUE))[,1]]
    # yesclose <- (diff(end(annchr)) + width(annchrwhfirst)) < 2*conslen
    yesclose <- (diff(start(annchr)) + width(annchrwhfirst)) < 2*conslen
    
    # Building granges of consensus sequence position
    startcons <- ifelse(strand(unlist(annchr)) == "+",
                        unlist(annchr)$repStart, 
                        unlist(annchr)$repLeft)
    # There are very few cases (for features in "-" strand) where the start is
    # after the end in the consensus sequence. Addressing them by assigning
    # the end position +1 as start 
    disc_start <- startcons > unlist(annchr)$repEnd
    startcons[disc_start] <- unlist(annchr)$repEnd[disc_start] +1
    grcons <- relist(IRanges(start=startcons, 
                             end=unlist(annchr)$repEnd), annchr)
    # Splitting when start/end in chr and consensus sequence of the 1st is not
    # before the 2nd, and elements that are not close enough
    if (minusStrand) {
      yescons <- diff(start(grcons)) < 0 & diff(end(grcons)) < 0
    } else {
      yescons <- diff(start(grcons)) > 0 & diff(end(grcons)) > 0
    }
    yesclosepos <- diff(start(annchr)) > 0 & diff(end(annchr)) > 0 &
      yescons & yesclose
    aggf <- integer(sum(lengths(annchr, use.names = FALSE)))
    whfirst <- c(1, cumsum(lengths(annchr, use.names=FALSE))+1)
    aggf[c(seq_along(aggf))[-whfirst]] <- unlist(cumsum(!yesclosepos))
    annchr2 <- split(unlist(annchr, use.names = FALSE), 
                     paste(rep(names(annchr), lengths(annchr)),aggf, sep ="."))
    annchr2 <- sort(annchr2)
    
    # Note that this implementation does not take into account the start 
    # position of the 1st merged element to see if the distance between the
    # end of a downstream element and the start of an upstream element
    # is smaller than twice the consensus length. Instead, here we take the
    # start of the previous element to compute this distance.
    
  } else {
    insert <- insert + 1 # to adapt to reduce() behaviour
    sp2 <- reduce(annchr, with.revmap=TRUE, min.gapwidth=insert)
    annchr2 <- relist(unlist(annchr, use.names = FALSE), 
                      mcols(unlist(sp2))$revmap)
    names(annchr2) <- paste(rep(names(sp2), lengths(sp2)), 
                            seq_along(annchr2), sep =".")
  }
  annchr2
}

#' @importFrom GenomicRanges start end mcols "mcols<-"
#' @importFrom S4Vectors pc
.reconstructERVs <- function(annchrltrsp2, annchrintsp2, outside, inside,
                             cons_length) {
  ltrtorec <- do.call("rbind", strsplit(names(annchrltrsp2), 
                                        ".", fixed=TRUE))[,1] %in% names(outside)
  inttorec<- do.call("rbind", strsplit(names(annchrintsp2), 
                                       ".", fixed=TRUE))[,1] %in% outside
  # GRangesList with both int and equivalent LTRs
  annchrltrint <- c(annchrltrsp2[ltrtorec], annchrintsp2[inttorec])
  o <- order(min(start(annchrltrint)), decreasing = FALSE)
  annchrltrint <- annchrltrint[o]
  whint <- which(mcols(annchrltrint)$Status == "int")
  namef <- do.call("rbind", strsplit(names(annchrltrint), ".", fixed =TRUE))[,1]
  int <- namef[whint]
  # ltr <- inside[int]
  ltr <- strsplit(inside[int], ":")
  whintup <- whint-1
  whintdo <- whint+1
  # Addressing for when first or last feature in 'annchrltrint' is not an LTR
  # by setting as upstream/downatream feature the int feature itself, causing
  # yesltrup/yesltrdown to be FALSE
  whintup[whintup < 1] <- 1
  whintdo[whintdo > length(annchrltrint)] <- length(annchrltrint)
  # yesltrup <- namef[whintup] == ltr
  # yesltrdown <- namef[whintdo] == ltr
  yesltrup <- .checkEquivalentLTR(namef[whintup], ltr)
  yesltrdown <- .checkEquivalentLTR(namef[whintdo], ltr)
  yesdistup <- (min(start(annchrltrint[whint])) -
                max(end(annchrltrint[whintup]))) < cons_length[namef[whintup]]/2
  # original code uses consensus length of 1st LTR
  yesdistdown <- (min(start(annchrltrint[whintdo])) -
                  max(end(annchrltrint[whint]))) < cons_length[namef[whintup]]/2
  
  # Addressing cases where an LTR is simultaneously a downstream LTR of int1
  # and an upstream LTR of int2. Preference is given to int1.
  whup <- yesltrup & yesdistup
  whdown <- yesltrdown & yesdistdown
  if (sum(whup) > 0 & sum(whdown) > 0) {
    conflictltr <- names(annchrltrint[whintup][whup]) %in%
      names(annchrltrint[whintdo][whdown])
    yesltrup[whup][conflictltr] <- FALSE
    yesdistup[whup][conflictltr] <- FALSE
    whup <- yesltrup & yesdistup
    whdown <- yesltrdown & yesdistdown
  }
  fulllength <- whup & whdown
  partial <- whup | whdown
  partial[fulllength] <- FALSE
  
  anngrlERVs <- .getFulllength_Partial_ERVs(fulllength, partial, annchrltrint, 
                                            whintup, whint, whintdo, yesltrup,
                                            yesdistup, yesltrdown, yesdistdown,
                                            annchrltrsp2, annchrintsp2, ltrtorec,
                                            inttorec)
  anngrlERVs
}

#' @importFrom GenomicRanges start end mcols "mcols<-"
#' @importFrom S4Vectors pc
.getFulllength_Partial_ERVs <- function(fl, pt, annchrltrint, whintup,
                                        whint, whintdo, yesltrup, yesdistup,
                                        yesltrdown, yesdistdown, annchrltrsp2,
                                        annchrintsp2, ltrtorec, inttorec) {
  # Creating GRangesList with full-length ERVs
  if (any(fl)) {
    fulllength_grl <- pc(annchrltrint[whintup][fl], annchrltrint[whint][fl],
                         annchrltrint[whintdo][fl])
    mcols(fulllength_grl)$Status <- "full-lengthLTR"
    # assigning names of int element
    names(fulllength_grl) <- names(annchrltrint[whint][fl])
    # Making sure no partially reconstructed ERV contains fragments from a 
    # full-length ERV
    disc <- names(annchrltrint[whintup][pt]) %in% names(annchrltrint[whintdo][fl])
    disc2 <- names(annchrltrint[whintdo][pt]) %in% names(annchrltrint[whintup][fl])
    pt[pt][disc | disc2] <- FALSE
  } else {
    fulllength_grl <- GRangesList()
  }
  ptup <- pt & yesltrup & yesdistup
  ptdown <- pt & yesltrdown & yesdistdown
  
  # Creating GRangesList of partial ERVs
  if (any(ptup)) {
    partial_grl1 <- pc(annchrltrint[whintup][ptup], annchrltrint[whint][ptup])
    mcols(partial_grl1)$Status <- "partialLTR_up"
    # assigning names of int element
    names(partial_grl1) <- names(annchrltrint[whint][ptup])
    # removing LTR for reconstruction with an upstream int if it has been 
    # previously reconstructed with a downstream int
    disc <- names(annchrltrint[whintdo][ptdown]) %in%
      names(annchrltrint[whintup][ptup])
    pt[ptdown][disc] <- FALSE
    ptdown <- pt & yesltrdown & yesdistdown
  } else {
    partial_grl1 <- GRangesList()
  }
  if (any(ptdown)) {
    partial_grl2 <- pc(annchrltrint[whint][ptdown],annchrltrint[whintdo][ptdown])
    mcols(partial_grl2)$Status <- "partialLTR_down"
    # assigning names of int element
    names(partial_grl2) <- names(annchrltrint[whint][ptdown])
  } else {
    partial_grl2 <- GRangesList()
  }
  # names() of partial_grl2 are already int elements
  partial_grl <- c(partial_grl1, partial_grl2)
  
  # Which are not solo int or LTR (have been reconstructed)?
  whyesrec <- c(whint[fl], c(whintdo)[fl], c(whintup)[fl], whint[ptup], 
                whint[ptdown], c(whintup)[ptup], c(whintdo)[ptdown])
  
  # GRangesList with reconstructed ERVs and solo LTR and int
  anngrlERVs <- c(annchrltrsp2[!ltrtorec], annchrintsp2[!inttorec],
                  fulllength_grl, partial_grl, annchrltrint[-whyesrec])
  anngrlERVs
}

.checkEquivalentLTR <- function(namef, ltr) {
  yesequiv <- logical()
  for (i in seq_along(namef)) {
    yesequiv <- c(yesequiv, namef[i] %in% ltr[[i]])
  }
  yesequiv
}

#' @importFrom GenomicRanges width mcols "mcols<-"
.getRelLength <- function(nTEs, cons_length, inout, annrec) {
  whERVs <- mcols(annrec)$Status %in% c("partialLTR_up", "partialLTR_down",
                                        "full-lengthLTR", "int")
  ltrconsl <- cons_length[names(inout$outside)[match(nTEs[whERVs], inout$outside)]]
  cons_length_full <- cons_length[nTEs[whERVs]] + 2*ltrconsl
  # cons_length_full <- cons_length[nTEs[whERVs]] + 2*cons_length[inside[nTEs[whERVs]]]
  rlenERVs <- sum(width(annrec[whERVs])) / cons_length_full
  
  # for solo LTRs and the rest of TEs the consensus length is obtained from the
  # positions in the consensus sequence
  rlenOther <- sum(width(annrec[!whERVs])) / cons_length[nTEs[!whERVs]]
  
  rlen <- numeric(length(annrec))
  rlen[whERVs] <- rlenERVs
  rlen[!whERVs] <- rlenOther
  # addressing relative lengths > 1 (due to assembled fragments that partially
  # overlap between them or consensus lengths shorter than the single fragment)
  rlen[rlen > 1] <- 1
  
  rlen
}


#' atena annotation parser of RepeatMasker annotations
#' @param gr A [`GRanges`][GenomicRanges::GRanges-class] object with
#' RepeatMasker annotations from \link[AnnotationHub]{AnnotationHub}
#' 
#' @param strict (Default FALSE) A logical; if TRUE, the 80-80 rule is applied,
#' i.e. only copies with more than 80% identity to the reference
#' and more than 80 bp long are reported.
#' 
#' @param insert (Default 1000L) An integer > 0. Fragments are assembled 
#' together if the distance between their closest extremities
#' is equal or less than \code{insert}. When \code{insert} = 0, two fragments
#' are assembled if they are in contact next to each other.
#'
#' @return A [`GRangesList`][GenomicRanges::GRangesList-class] object.
#'         
#' @details 
#' atena annotation parser of RepeatMasker annotations.
#' Parses RepeatMasker annotations from UCSC by assembling together fragments
#' from the same transposable element (TE) that are close enough (determined
#' by the \code{insert} parameter). For TEs from the LTR class, the parser
#' tries to reconstruct full-length, when possible, or partial TEs following
#' the LTR - internal region - LTR structure. Equivalences between LTR and
#' internal regions are found by, first, identifying LTR regions (those with
#' the "LTR" substring in their name) and internal regions (those with a
#' suffix such as "-int", "-I", etc.). Then, LTR are assigned to internal
#' regions for which the comparison of the two names are has a higher number
#' of equal consecutive characters.
#'
#' @examples
#' rmskat <- annotaTEs(genome="dm6", parsefun=rmskatenaparser,
#'                     strict=FALSE)
#' rmskat
#' 
#' @aliases rmskatenaparser
#' @rdname rmskatenaparser
#' @name rmskatenaparser
#' @importFrom GenomicRanges strand width mcols "mcols<-" seqnames
#' @importFrom IRanges mean
#' @importFrom Seqinfo seqlevels
#' @export
rmskatenaparser <- function(gr, strict= FALSE, insert=1000) {
  if (!is(gr, "GRanges"))
    stop("'gr' should be a GRanges object with RepeatMasker annotations")
  
  if (length(gr) == 0)
    stop("'gr' is empty")
  
  if (!is.integer(insert) & !is.numeric(insert))
    stop("'insert' must be an integer value")
  
  inout <- .builDictionary_at(gr)
  
  gr <- .filterNonTEs(gr)
  gr$repStart <- abs(gr$repStart)
  gr$repLeft <- abs(gr$repLeft)
  cons_length <- .getConsLength(gr)
  
  fsplit <- paste(mcols(gr)$repName, seqnames(gr), strand(gr), sep = ";")
  ann <- split(gr, fsplit)
  mcols(ann)$repName <- do.call("rbind", strsplit(names(ann), split = ";", 
                                                  fixed =TRUE))[,1]
  annrec <- .reconstructTEs_at(ann=ann, inout$outside, inout$inside,
                               cons_length, insert)
  # -- Filtering based on 'strict' option
  if (strict) {
    maxdiv <- mean(relist(mcols(unlist(annrec))$milliDiv, annrec)) < 200
    tokeep <- sum(width(annrec)) >= 80 & maxdiv
    annrec <- annrec[tokeep]
  }
  
  # -- Computation relative length --
  # for full-length and partial ERVs, and ERVs with only internal region, the
  # consensus length is considered that of the full-length TE: LTR + int + LTR
  nTEs <- do.call("rbind", strsplit(names(annrec), ";", fixed=TRUE))[,1]
  mcols(annrec)$RelLength <- .getRelLength(nTEs, cons_length, inout, annrec)
  
  # -- Simplifying TE name: subfamily name + number of TE separated by "." --
  names(annrec) <- paste(nTEs, seq_along(annrec), sep = ".")
  
  # -- Including TE class name in GRangesList 
  mcols(annrec)$Class <- unlist(unique(relist(unlist(annrec)$repClass, annrec)))
  
  annrec
}

#' @importFrom stats aggregate
#' @importFrom Matrix colSums
.builDictionary_at <- function(gr, thmatch=5, threl = 0.5) {
  
  # identifying LTR and int regions
  annltr <- gr[gr$repClass=="LTR"]
  # Possible internal id
  intid <- c("int","in", "i")
  intid <- c(paste0("-", intid), paste0("_", intid))
  internal <- grepl(pattern = paste(intid,collapse="|"), annltr$repName, 
                    ignore.case = TRUE)
  ltr <- grepl(pattern = "LTR", annltr$repName, ignore.case = TRUE)
  ltr[internal] <- FALSE # preference goes to internal regions
  # not internal and not ltr: undecided (it could be int or LTR)
  # here we discard undecided elements
  undecid <- !internal & !ltr
  internal[undecid] <- FALSE
  ltr[undecid] <- FALSE
  
  intuniq <- unique(annltr$repName[internal])
  ltruniq <- unique(annltr$repName[ltr])
  
  inside <- character(length(intuniq))
  names(inside) <- intuniq
  outside <- character(length(ltruniq))
  names(outside) <- ltruniq
  
  mm <- NULL
  for (ltrn in names(outside)) {
    nmatch <- vapply(names(inside), function(intn) {
      max(nchar(abbreviate(c(intn, ltrn), minlength = 1, use.classes=FALSE)))
    }, FUN.VALUE = integer(1L))
    nmatch[nmatch < c(thmatch +1)] <- 0
    nmatch <- nmatch / nchar(ltrn)
    mm <- cbind(mm, nmatch)
  }
  colnames(mm) <- names(outside)
  rownames(mm) <- names(inside)
  # at least 50% of characters in LTR name have to consecutively match the
  # internal region name
  mm[mm < threl] <- 0L
  whmin <- colSums(mm) > 0
  outside[whmin] <- names(inside)[apply(mm[,whmin],2, which.max)]
  aginside <- aggregate(names(outside[outside != ""]), 
                        by = list(outside[outside != ""]),
                        FUN= paste, collapse = ":")
  inside[aginside$Group.1] <- aginside$x
  outside <- outside[outside != ""]
  
  list(inside=inside, outside=outside)
}

#' @importFrom GenomicRanges start mcols "mcols<-"
#' @importFrom Seqinfo seqlevels
.reconstructTEs_at <- function(ann, outside, inside, cons_length, insert) {
  # For LTRs and internal regions, we check for the presence (in the same 
  # strand) of internal regions and LTRs, respectively, between elements with 
  # the same repName. To do so, we divide features with the same repName
  # (e.g. LTR23) using the start of the corresponding internal region
  # (LTR23-int) as flanking regions.
  
  # Selecting only cases where both the LTR and internal region are in the
  # chromosome
  stchr <- c(paste(seqlevels(ann), c("+"), sep =";"),
             paste(seqlevels(ann), c("-"), sep =";"))
  outside_stchr <- paste(rep(outside, each = length(stchr)), 
                         rep(stchr, length(outside)), sep =";")
  names(outside_stchr) <-  paste(rep(names(outside), each = length(stchr)), 
                                 rep(stchr, length(outside)), sep =";")
  outsidechr <- outside_stchr[(outside_stchr %in% names(ann)) &
                                (names(outside_stchr) %in% names(ann))]
  whint <- which(names(ann) %in% outsidechr)
  whltr <- which(names(ann) %in% names(outsidechr))
  # We extract coordinates of int and corresponding LTR (same order)
  annint <- ann[whint]
  annltr <- ann[whltr]
  
  if (length(annint) > 0 & length(annltr) > 0) {
    splitf2 <- .splitNonContinous_at(annint, annltr, outsidechr)
    annltrsp <- split(unlist(annltr, use.names = FALSE), splitf2$splitfltr2)
    annintsp <- split(unlist(annint, use.names = FALSE), splitf2$splitfint2)
    ann <- ann[-c(whint, whltr)]
    # We keep LTR and int separated from the rest of TEs to later perform the
    # reconstruction of full-length or partial ERVs
    
    # Now, features with the same repName, strand and chromosome are merged
    # together if they are close enough, according to the 'insert' parameter
    annltrsp2 <- .mergeCloseFeatures_at(annltrsp, cons_length, insert)
    annintsp2 <- .mergeCloseFeatures_at(annintsp, cons_length, insert)
    mcols(annltrsp2)$Status <- "LTR"
    mcols(annintsp2)$Status <- "int"
  }
  ann2 <- .mergeCloseFeatures_at(ann, cons_length, insert)
  
  # Note: All features not identified as LTR or int (from the previously built
  # dictionary with equivalences: 'outside'), are identinfied as "noLTR",
  # however LTRs and int are expected to be inside 'annchr2' since the
  # dictionary 'outside' does not identify all LTR and int regions.
  mcols(ann2)$Status <- "noLTR"
  
  # ---- Reconstructing full-length and partial ERVs ----
  if (length(annint) > 0 & length(annltr) > 0) {
    anngrlERVs <- .reconstructERVs_at(annltrsp2, annintsp2, outsidechr,
                                      outside, inside, cons_length)
    anngrl <- c(ann2, anngrlERVs)
  } else {
    anngrl <- ann2
  }
  chr <- do.call("rbind", strsplit(names(anngrl), ";", fixed=TRUE))[, 2]
  opos <- order(chr, min(start(anngrl)), decreasing=FALSE)
  anngrl[opos]
}

#' @importFrom GenomicRanges start mcols "mcols<-"
#' @importFrom Matrix colSums rowSums
.splitNonContinous_at <- function(annint, annltr, outsidechr) {
  annint2 <- annint[outsidechr[names(annltr)]]
  annltr2 <- split(unlist(annltr), 
                   rep(outsidechr[names(annltr)], lengths(annltr)))
  
  posc_int <- mapply(function(ltrname, intname) 
    outer(start(annltr2[[ltrname]]), start(annint[[intname]]), "<"),
    ltrname = names(annltr2), intname = names(annint),
    SIMPLIFY = FALSE)
  
  posc_ltr <- mapply(function(ltrname, intname) 
    outer(start(annltr[[ltrname]]), start(annint2[[intname]]), "<"),
    ltrname = names(annltr), intname = names(annint2),
    SIMPLIFY = FALSE)
  
  # 'posc' is a list of matrices in which ltr are rows and int regions are 
  # columns the cells are TRUE/FALSE depending if the start position of
  # a ltr (row) is smaller than the start position of a int (col)
  
  # Creating a factor to split ltr and int according to their position in
  # relation to their equivalent int and ltr, respectively.
  splitfltr <- lapply(posc_ltr, function(x) rowSums(x))
  splitfint <- lapply(posc_int, function(x) colSums(x))
  splitfltr2 <- as.factor(paste(rep(names(splitfltr), lengths(splitfltr)),
                                unlist(splitfltr), sep = "."))
  splitfint2 <- as.factor(paste(rep(names(splitfint), lengths(splitfint)),
                                unlist(splitfint), sep = "."))
  list(splitfltr2 = splitfltr2, splitfint2 = splitfint2)
}

#' @importFrom GenomicRanges GRangesList start width strand end reduce
.mergeCloseFeatures_at <- function(annsp, cons_length, insert) {
  insert <- insert + 1 # to adapt to reduce() behaviour
  sp2 <- reduce(annsp, with.revmap=TRUE, min.gapwidth=insert)
  annsp2 <- relist(unlist(annsp, use.names=FALSE), mcols(unlist(sp2))$revmap)
  names(annsp2) <- paste(rep(names(sp2), lengths(sp2)), 
                         seq_along(annsp2), sep =".")
  annsp2
}

#' @importFrom GenomicRanges start end mcols "mcols<-"
#' @importFrom S4Vectors pc
.reconstructERVs_at <- function(annltrsp2, annintsp2, outsidechr, 
                                outside, inside, cons_length) {
  ltrtorec <- do.call("rbind", strsplit(names(annltrsp2), 
                                        ".", fixed=TRUE))[,1] %in% names(outsidechr)
  inttorec<- do.call("rbind", strsplit(names(annintsp2), 
                                       ".", fixed=TRUE))[,1] %in% outsidechr
  # GRangesList with both int and equivalent LTRs
  annltrint <- c(annltrsp2[ltrtorec], annintsp2[inttorec])
  
  # Setting order based on strand, chr and start position of element 
  chr <- do.call("rbind", strsplit(names(annltrint), ";", fixed=TRUE))[,2]
  st <- do.call("rbind", strsplit(names(annltrint), ";", fixed=TRUE))[,3]
  st <- do.call("rbind", strsplit(st, ".", fixed=TRUE))[,1]
  o <- order(chr, st, min(start(annltrint)), decreasing = FALSE)
  annltrint <- annltrint[o]
  whint <- which(mcols(annltrint)$Status == "int")
  namef <- do.call("rbind", strsplit(names(annltrint), ".", fixed =TRUE))[,1]
  int <- namef[whint]
  ltr <- split(names(outsidechr), f = outsidechr)
  ltr <- ltr[int]
  whintup <- whint-1
  whintdo <- whint+1
  # Addressing for when first or last feature in 'annltrint' is not an LTR
  # by setting as upstream/downatream feature the int feature itself, causing
  # yesltrup/yesltrdown to be FALSE
  whintup[whintup < 1] <- 1
  whintdo[whintdo > length(annltrint)] <- length(annltrint)
  yesltrup <- .checkEquivalentLTR(namef[whintup], ltr)
  yesltrdown <- .checkEquivalentLTR(namef[whintdo], ltr)
  namef_simp <- do.call("rbind", strsplit(namef, ";", fixed =TRUE))[,1]
  yesdistup <- (min(start(annltrint[whint])) - max(end(annltrint[whintup]))) <
    cons_length[namef_simp[whintup]]/2
  yesdistdown <- (min(start(annltrint[whintdo])) - max(end(annltrint[whint]))) <
    cons_length[namef_simp[whintup]]/2
  
  # Addressing cases where an LTR is simultaneously a downstream LTR of int1
  # and an upstream LTR of int2. Preference is given to int1.
  whup <- yesltrup & yesdistup
  whdown <- yesltrdown & yesdistdown
  if (sum(whup) > 0 & sum(whdown) > 0) {
    conflictltr <- names(annltrint[whintup][whup]) %in%
      names(annltrint[whintdo][whdown])
    yesltrup[whup][conflictltr] <- FALSE
    yesdistup[whup][conflictltr] <- FALSE
    whup <- yesltrup & yesdistup
    whdown <- yesltrdown & yesdistdown
  }
  fulllength <- whup & whdown
  partial <- whup | whdown
  partial[fulllength] <- FALSE
  
  anngrlERVs <- .getFulllength_Partial_ERVs(fulllength, partial, annltrint,
                                            whintup, whint, whintdo, yesltrup,
                                            yesdistup, yesltrdown, yesdistdown,
                                            annltrsp2, annintsp2, ltrtorec,
                                            inttorec)
  anngrlERVs
}

#' Build a TEtranscripts parameter object
#'
#' Build an object of the class \code{TEtranscriptsParam}
#'
#' @param bfl a character string vector of BAM file names.
#'
#' @param teFeatures A \code{GRanges} or \code{GRangesList} object with the
#' TE annotated features to be quantified. Elements in this object should have
#' names, which are used as a grouping factor for genomic ranges forming a
#' common locus, unless other metadata column names are specified in the
#' \code{aggregateby} parameter.
#'
#' @param aggregateby Character vector with column names from the annotation
#' to be used to aggregate quantifications. By default, this is an empty
#' vector, which means that the names of the input \code{GRanges} or
#' \code{GRangesList} object given in the \code{teFeatures} parameter are used
#' to aggregate quantifications.
#' 
#' @param ovMode Character vector indicating the overlapping mode. Available
#' options are: "ovUnion" (default) and "ovIntersectionStrict",
#' which implement the corresponding methods from HTSeq
#' (\url{https://htseq.readthedocs.io/en/release_0.11.1/count.html}).
#' Ambiguous alignments (alignments overlapping > 1 feature) are addressed
#' as in the original TEtranscripts method.
#'
#' @param geneFeatures (Default NULL) A \code{GRanges} or 
#' \code{GRangesList} object with the
#' gene annotated features to be quantified. Following the TEtranscripts
#' algorithm, overlaps with unique reads are first tallied with respect to
#' these gene features. Elements should have names indicating the gene name/id.
#' In case that \code{geneFeatures} is a \code{GRanges} and contains a 
#' metadata column named
#' \code{type}, only the elements with \code{type} = \code{exon} are considered
#' for the analysis. Then, exon counts are summarized to the gene level. If
#' NULL, gene expression is not quantified.
#'
#' @param singleEnd (Default TRUE) Logical value indicating if reads are single
#' (\code{TRUE}) or paired-end (\code{FALSE}).
#'
#' @param strandMode (Default 1) Numeric vector which can take values 0, 1 or
#' 2.
#' The strand mode is a per-object switch on
#' \code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
#' objects that controls the behavior of the strand getter. See
#' \code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
#' class for further detail. If \code{singleEnd = TRUE}, then \code{strandMode}
#' is ignored.
#'
#' @param ignoreStrand (Default FALSE) Logical value that defines if the strand
#' should be taken into consideration when computing the overlap between reads
#' and annotated features. When \code{ignoreStrand = FALSE}, an aligned read
#' is considered to be overlapping an annotated feature as long as they
#' have a non-empty intersecting genomic range on the same strand, while when
#' \code{ignoreStrand = TRUE} the strand is not considered.
#'
#' @param fragments (Default TRUE) Logical value applied to paired-end data
#' only. In both cases (\code{fragments=FALSE} and \code{fragments=TRUE}), the
#' read-counting method discards not properly paired reads. Moreover,
#' when \code{fragments=FALSE}, only non-ambiguous properly paired reads are
#' counted. When \code{fragments=TRUE}, ambiguous reads are also counted 
#' (see "Pairing criteria" in \code{\link[GenomicAlignments]{readGAlignments}()}). 
#' \code{fragments=TRUE} is equivalent to
#' the behavior of the TEtranscripts algorithm. For further details see
#' \code{\link[GenomicAlignments]{summarizeOverlaps}()}.
#' 
#' @param tolerance A positive numeric scalar storing the minimum tolerance
#' above which the SQUAREM algorithm (Du and Varadhan, 2020) keeps iterating.
#' Default is \code{1e-4} and this value is passed to the \code{tol} parameter
#' of the \code{\link[SQUAREM]{squarem}()} function.
#'
#' @param maxIter A positive integer scalar storing the maximum number of
#' iterations of the SQUAREM algorithm (Du and Varadhan, 2020). Default
#' is 100 and this value is passed to the \code{maxiter} parameter of the
#' \code{\link[SQUAREM]{squarem}()} function.
#'
#' @param verbose (Default \code{TRUE}) Logical value indicating whether to
#' report progress.
#'
#' @details
#' This is the constructor function for objects of the class
#' \code{TEtranscriptsParam-class}. This type of object is the input to the
#' function \code{\link{qtex}()} for quantifying expression of transposable
#' elements using the TEtranscripts method
#' \href{https://doi.org/10.1093/bioinformatics/btv422}{Jin et al. (2015)}. The
#' TEtranscripts algorithm quantifies TE expression by using an EM algorithm
#' to optimally distribute ambiguously mapped reads.
#' 
#' @return A \linkS4class{TEtranscriptsParam} object.
#'
#' @examples
#' bamfiles <- list.files(system.file("extdata", package="atena"),
#'                        pattern="*.bam", full.names=TRUE)
#' \dontrun{
#' ## use the following two instructions to fetch annotations, they are here
#' ## commented out to enable running this example quickly when building and
#' ## checking the package
#' rmskat <- annotaTEs(genome="dm6", parsefun=rmskatenaparser,
#'                     strict=FALSE, insert=500)
#' rmskLTR <- getLTRs(rmskat, relLength=0.8,
#'                    fullLength=TRUE,
#'                    partial=TRUE,
#'                    otherLTR=TRUE)
#' }
#'
#' ## DO NOT TYPE THIS INSTRUCTION, WHICH JUST LOADS A PRE-COMPUTED ANNOTATION
#' ## YOU SHOULD USE THE INSTRUCTIONS ABOVE TO FETCH ANNOTATIONS
#' rmskLTR <- readRDS(system.file("extdata", "rmskatLTRrlen80flenpartoth.rds",
#'                                package="atena"))
#'
#'
#' library(TxDb.Dmelanogaster.UCSC.dm6.ensGene)
#' txdb <- TxDb.Dmelanogaster.UCSC.dm6.ensGene
#' txdb_genes <- genes(txdb)
#'
#' ## build a parameter object for TEtranscripts
#' ttpar <- TEtranscriptsParam(bamfiles,
#'                             teFeatures=rmskLTR,
#'                             geneFeatures=txdb_genes,
#'                             singleEnd=TRUE,
#'                             ignoreStrand=TRUE,
#'                             aggregateby="repName")
#' ttpar
#'
#' @references
#' Jin Y et al. TEtranscripts: a package for including transposable elements
#' in differential expression analysis of RNA-seq datasets.
#' Bioinformatics. 2015;31(22):3593-3599. DOI:
#' \url{https://doi.org/10.1093/bioinformatics/btv422}
#'
#' @importFrom methods is new
#' @importFrom Rsamtools BamFileList
#' @importFrom cli cli_alert_info cli_alert_success
#' @export
#' @rdname TEtranscriptsParam-class
TEtranscriptsParam <- function(bfl, teFeatures, aggregateby=character(0),
                               ovMode="ovUnion",
                               geneFeatures=NULL,
                               singleEnd=TRUE,
                               ignoreStrand=FALSE,
                               strandMode=1L,
                               fragments=TRUE,
                               tolerance=0.0001,
                               maxIter=100L,
                               verbose=TRUE) {

    if (verbose)
        cli_alert_info("Locating BAM files")
    bfl <- .checkBamFileListArgs(bfl, singleEnd, fragments)

    if (!ovMode %in% c("ovUnion","ovIntersectionStrict"))
      stop("'ovMode' should be one of 'ovUnion', 'ovIntersectionStrict'")
    
    if (verbose)
        cli_alert_info("Processing features")
    teFeaturesobjname <- deparse(substitute(teFeatures))
    geneFeaturesobjname <- deparse(substitute(geneFeatures))
    features <- .processFeatures(teFeatures, teFeaturesobjname,
                                 geneFeatures, geneFeaturesobjname,
                                 aggregateby, aggregateexons=TRUE)

    obj <- new("TEtranscriptsParam", bfl=bfl, features=features,
               aggregateby=aggregateby, ovMode=ovMode,
               singleEnd=singleEnd, ignoreStrand=ignoreStrand,
               strandMode=as.integer(strandMode), fragments=fragments,
               tolerance=tolerance, maxIter=as.integer(maxIter))
    if (verbose)
        cli_alert_success("Parameter object successfully created")
    obj
}

#' @param object A \linkS4class{TEtranscriptsParam} object.
#'
#' @importFrom Seqinfo seqlevels
#' @export
#' @aliases show,TEtranscriptsParam-method
#' @rdname TEtranscriptsParam-class
setMethod("show", "TEtranscriptsParam",
        function(object) {
            cat(class(object), "object\n")
            cat(sprintf("# BAM files (%d): %s\n", length(object@bfl),
                        .pprintnames(names(object@bfl))))
            cat(sprintf("# features (%s length %d): %s\n",
                        class(object@features),
                        length(object@features),
                        ifelse(is.null(names(object@features)),
                                paste("on",
                                    .pprintnames(seqlevels(features(object)))),
                                .pprintnames(names(features(object))))))
            if (length(object@aggregateby) > 0)
                cat(sprintf("# aggregated by: %s\n",
                            paste(object@aggregateby, collapse=", ")))
            cat(sprintf("# %s; %s",
                        ifelse(object@singleEnd, "single-end", "paired-end"),
                        ifelse(object@ignoreStrand, "unstranded", "stranded")))
            if (!object@ignoreStrand)
                cat(sprintf(" (strandMode=%d)", object@strandMode))
            if (!object@singleEnd)
                cat(sprintf("; %s",
                        ifelse(object@fragments, "counting properly paired, same-strand pairs, singletons, reads with unmapped pairs and other fragments",
                                "counting properly paired reads")))
            cat("\n")
        })

#' @importFrom BiocParallel SerialParam bplapply
#' @importFrom SummarizedExperiment SummarizedExperiment
#' @export
#' @aliases qtex
#' @aliases qtex,TEtranscriptsParam-method
#' @rdname qtex
setMethod("qtex", "TEtranscriptsParam",
        function(x, phenodata=NULL, mode=ovUnion, yieldSize=1e6L,
                    BPPARAM=SerialParam(progressbar=TRUE)) {
            .checkPhenodata(phenodata, length(x@bfl))

            cnt <- bplapply(x@bfl, .qtex_tetranscripts, ttpar=x, mode=x@ovMode,
                            yieldSize=yieldSize, BPPARAM=BPPARAM)
            cnt <- do.call("cbind", cnt)
            cdata <- .createColumnData(cnt, phenodata)
            colnames(cnt) <- rownames(cdata)

            features <- .consolidateFeatures(x, rownames(cnt))

            SummarizedExperiment(assays=list(counts=cnt),
                                 rowRanges=features,
                                 colData=cdata)
        })

#' @importFrom stats setNames aggregate median
#' @importFrom Rsamtools scanBamFlag ScanBamParam yieldSize yieldSize<-
#' @importFrom GenomicRanges width
#' @importFrom GenomicAlignments readGAlignments readGAlignmentsList
#' @importFrom GenomicAlignments readGAlignmentPairs
#' @importFrom S4Vectors Hits queryHits subjectHits
#' @importFrom Matrix Matrix rowSums colSums t which
#' @importFrom IRanges ranges
.qtex_tetranscripts <- function(bf, ttpar, mode, yieldSize=1e6L) {
    mode <- match.fun(mode)
    readfun <- .getReadFunction(ttpar@singleEnd, ttpar@fragments)
    sbflags <- .getScanBamFlag_tt(ttpar@singleEnd)
    param <- ScanBamParam(flag=sbflags, what="flag", tag="AS")
    
    iste <- mcols(features(ttpar))$isTE
    if (any(duplicated(names(features(ttpar)[iste]))))
        stop(".qtex_tetranscripts: duplicated names in TE annotations.")

    ov <- Hits(nLnode=0, nRnode=length(ttpar@features), sort.by.query=TRUE)
    alnreadids <- character(0)
    avgreadlen <- integer(0)
    d <- numeric(0)
    salnmask <- logical(0)
    strand_arg <- "strandMode" %in% formalArgs(readfun)
    yieldSize(bf) <- yieldSize
    open(bf)
    while (length(alnreads <- do.call(readfun, 
                                c(list(file = bf), list(param=param),
                                list(strandMode=ttpar@strandMode)[strand_arg],
                                list(use.names=TRUE))))) {
        alnreads <- .matchSeqinfo(alnreads, ttpar@features)
        avgreadlen <- c(avgreadlen, .getAveLen(ttpar, alnreads))
        if (is(alnreads, "GAlignmentPairs")) {
            d <- c(d, abs(start(first(alnreads)) - start(second(alnreads))))
        } else if (is(alnreads, "GAlignmentsList")) {
            d <- c(d, max(abs(diff(start(alnreads)))))
            d[d < 0] <- 0
        }
        salnmask <- c(salnmask, any(.secondaryAlignmentMask(alnreads)))
        alnreadids <- c(alnreadids, names(alnreads))
        thisov <- mode(alnreads, ttpar@features,
                       ignoreStrand=ttpar@ignoreStrand, inter.feature=FALSE)
        ov <- .appendHits(ov, thisov)
    }
    on.exit(close(bf))
    .checkOvandsaln(ov, salnmask)
    ## get uniquely aligned-reads
    maskuniqaln <- .getMaskUniqueAln(alnreadids)
    if (ttpar@singleEnd == TRUE) {
        # unique + multi-mapping reads (only once) are considered
        avgreadlen <- mean(avgreadlen[!duplicated(alnreadids)])
    } else {
        # only unique reads are considered
        avgreadlen <- mean(avgreadlen[(d <= 500) & maskuniqaln])
    }
    ## fetch all different read identifiers from the overlapping alignments
    readids <- unique(alnreadids[queryHits(ov)])
    ## fetch all different transcripts from the overlapping alignments
    tx_idx <- sort(unique(subjectHits(ov)))
    
    cntvec <- .ttQuantExpress(ov, alnreadids, readids, tx_idx, ttpar, iste,
                              maskuniqaln, avgreadlen)
    ## Original TEtranscripts coerces fractional counts to integer
    setNames(as.integer(cntvec), names(cntvec))
}

#' @importFrom GenomicAlignments extractAlignmentRangesOnReference
#' @importFrom GenomicAlignments cigar start qwidth first second
.getAveLen <- function(ttpar, alnreads) {
    if (is(alnreads, "GAlignments")) {
        avgreadlenaln <- qwidth(alnreads)
    } else if (is(alnreads, "GAlignmentPairs")) {
        d <- abs(start(first(alnreads)) - start(second(alnreads)))
        d2 <- qwidth(second(alnreads))
        avgreadlenaln <- d + d2
    } else if (is(alnreads, "GAlignmentsList")) {
        d <- max(abs(diff(start(alnreads))))
        d[d<0] <- 0
        l <- lengths(alnreads)
        d2 <- unlist(qwidth(alnreads))[cumsum(l)]
        avgreadlenaln <- d + d2
    } else {
        stop(sprintf(".getAveLen: wrong class %s\n", class(alnreads)))
    }
    avgreadlenaln
}

.ttQuantExpress <- function(ov, alnreadids, readids, tx_idx, ttpar, iste,
                            maskuniqaln, avgreadlen) {
    
    ## build a matrix representation of the overlapping alignments
    ovalnmat <- .buildOvAlignmentsMatrix(ov, alnreadids, readids, tx_idx)
    
    mt <- match(readids, alnreadids)
    multigcnt <- rep(0L, length(ttpar@features))
    istex <- as.vector(iste[tx_idx])
    
    if (!all(iste)) {
        ## Correcting for preference of unique/multi-mapping reads to genes or TEs
        ovalnmat <- .correctPreference(ovalnmat, maskuniqaln, mt, istex)
    }
    
    # Getting counts from unique reads
    uniqcnt <- .countUniqueRead(ttpar, ovalnmat, maskuniqaln, mt, tx_idx, istex)
    
    if (!all(iste) & any(!maskuniqaln)) {
        ## Getting gene counts where a multimapping read maps to > 1 gene
        multigcnt <- .countMultiReadsGenes(ttpar, ovalnmat, maskuniqaln, mt,
                                           iste, istex, tx_idx, readids, 
                                           alnreadids, ov, uniqcnt)
    }
    
    # Adjusting 'ovalnmat' when alignment from multimapping read maps > 1 TE
    ovalnmat <- .adjustalnmultiovTE(iste, ov, alnreadids, readids, tx_idx, ovalnmat)
    
    cntvec <- rep(0L, length(ttpar@features))
    cntvec <- .ttEMstep(maskuniqaln, mt, ovalnmat, istex, tx_idx, readids, ttpar,
                        avgreadlen, cntvec)
    ## Summarize counts of unique and multimapping reads
    cntvec <- .summarizeCounts(iste, cntvec, uniqcnt, multigcnt, ttpar)
    cntvec
}



#' @importFrom Matrix Matrix sparseMatrix t which
#' @importFrom sparseMatrixStats colSums2 rowSums2
#' @importFrom SQUAREM squarem
#' @importFrom IRanges ranges
.ttEMstep <- function(maskuniqaln, mt, ovalnmat, istex, tx_idx, readids, ttpar,
                      avgreadlen, cntvec) {
if (sum(!maskuniqaln[mt]) > 0) { ## multi-mapping reads
    ## TEtranscripts doesn't use uniquely-aligned reads to inform the
    ## procedure of distributing multiple-mapping reads, as explained in
    ## Jin et al. (2015) pg. 3594, "to reduce potential bias to certain TEs."
    ## Hence, once counted, we discard unique alignments
    ovalnmat <- ovalnmat[!maskuniqaln[mt], istex]
    yesov <- rowSums2(ovalnmat)>0
    ovalnmat <- ovalnmat[yesov,]
    readids <- readids[!maskuniqaln[mt]][yesov]
    ## the Qmat matrix stores row-wise the probability that read i maps to
    ## a transcript j, assume uniform probabilities by now
    Qmat <- Matrix(0, nrow=length(readids), ncol=length(tx_idx[istex]),
                    dimnames=list(readids, NULL))
    Qmat[which(ovalnmat>0, arr.ind=TRUE)] <- 1
    # Qmat <- Qmat / rowSums2(ovalnmat)
    
    ## Pi, corresponding to rho in Equations (1), (2) and (3) in Jin et al.
    ## (2015) stores probabilities of expression for each transcript, corrected
    ## for its effective length as defined in Eq. (1) of Jin et al. (2015)
    # Pi <- colSums2(Qmat)
    # Pi <- colSums2(ovalnmat / rowSums2(ovalnmat))
    Pi <- colSums2(ovalnmat) / sum(ovalnmat)
    
    if (is(ttpar@features,"GRangesList")) {
        elen <- as.numeric(sum(width(ttpar@features[tx_idx][istex]))) - avgreadlen+1
    } else {
        elen <- width(ttpar@features[tx_idx][istex]) - avgreadlen + 1
    }
    Pi <- .correctForTxEffectiveLength(Pi, elen)
    
    ## use the SQUAREM algorithm to achieve faster EM convergence
    emres <- squarem(par=Pi, Q=Qmat, elen=elen,
                    fixptfn=.ttFixedPointFun,
                    control=list(tol=ttpar@tolerance, maxiter=ttpar@maxIter))
    Pi <- emres$par
    Pi[Pi < 0] <- 0 ## Pi estimates are sometimes negatively close to zero
    Pi <- Pi / sum(Pi)
    ## use the estimated transcript expression probabilities
    ## to finally distribute ambiguously mapping reads
    # probmassbyread <- as.vector(ovalnmat %*% Pi) 
    probmassbyread <- as.vector(Qmat %*% Pi)
    
    # Version 1
    # cntvecovtx <- rowSums(t(ovalnmat / probmassbyread) * Pi, na.rm=TRUE)
    # Version 2
    # wh <- which(ovalnmat, arr.ind=TRUE)
    cntvecovtx <- rep(0, ncol(ovalnmat)) #cntvecovtx <- rep(0, length(tx_idx))
    # x <- tapply(Pi[wh[, "col"]] / probmassbyread[wh[, "row"]], wh[, "col"],
    #             FUN=sum, na.rm=TRUE)
    wh <- which(Qmat==1, arr.ind=TRUE)
    x <- tapply(Pi[wh[, "col"]] / probmassbyread[wh[, "row"]], wh[, "col"],
                FUN=sum, na.rm=TRUE)
    
    cntvecovtx[as.integer(names(x))] <- x
    cntvec[tx_idx][istex] <- cntvecovtx
}
cntvec
}


## private function .buildOvAlignmentsMatrix()
## builds a matrix representation of the overlapping alignments
## with reads on the rows and transcripts on the columns and
## a cell (i, j) set to TRUE if read i aligns to transcript j.
## therefore, when a read i aligns more than once to the same
## transcript j this is represented only once in the matrix
## parameters: ov - Hits object with the overlaps between
##                  alignments and features
##             arids - alignment read identifiers
##             rids - unique read identifiers
##             fidx - features index

#' @importFrom Matrix sparseMatrix
#' @importFrom S4Vectors queryHits subjectHits
.buildOvAlignmentsMatrix <- function(ov, arids, rids, fidx) {
    mt1 <- match(arids[queryHits(ov)], rids)
    mt2 <- match(subjectHits(ov), fidx)
    positions <- cbind(mt1, mt2)
    oamat <- sparseMatrix(positions[,1], positions[,2], x=TRUE, 
                        dims = c(length(rids), length(fidx)))
    oamat
}

## private function .correctForTxEffectiveLength()
## corrects input transcript expression probabilities by
## the transcript effective length as defined in Eq. (1) of
## Jin et al. (2015)
## parameters: x - transcript expression probabilities
##             elen - effective length of each transcript
.correctForTxEffectiveLength <- function(x, elen) {
    x[elen > 0] <- x[elen > 0] / elen[elen > 0]
    x[elen <= 0] <- 0
    if (sum(x) > 0) {
        x <- x / sum(x)
    }
    x
}

## private function .ttEstep()
## E-step of the EM algorithm of TEtranscripts
.ttEstep <- function(Q, Pi) {
    ## this first part is a sparse optimization of doing
    ## X <- t(t(Q) * Pi)
    X <- Q
    j <- rep(seq_len(ncol(X)), diff(X@p))
    X@x <- X@x * Pi[j]
    ## End of first part
    X <- X[rowSums2(X)>0,, drop=FALSE]
    X <- X / rowSums2(X)
    X
}

## private function .ttEstep()
## M-step of the EM algorithm of TEtranscripts
.ttMstep <- function(X) {
    Pi <- colSums2(X) / sum(X@x)
    Pi
}

## private function .ttFixedPointFun()
## fixed point function of the EM algorithm of TEtranscripts
.ttFixedPointFun <- function(Pi, Q, elen) {
    X <- .ttEstep(Q, Pi)
    Pi2 <- .ttMstep(X)
    Pi2 <- .correctForTxEffectiveLength(Pi2, elen)
    Pi2
}

## private function .ttLlh()
## log-likelihood function of the TEtranscripts model
## (currently not used)
.ttLlh <- function(X, Pi, Q) {
    z <- t(t(Q) * Pi)
    mask <- z == 0
    z[mask] <- NA
    sum(X * log(z), na.rm=TRUE)
}


## private function .correctPreference()
## Corrects ovalnmat for preference of unique/multi-mapping reads to
## genes/TEs, respectively
.correctPreference <- function(ovalnmat, maskuniqaln, mt, istex) {
    indx <- (rowSums2(ovalnmat[,istex]) > 0) & (rowSums2(ovalnmat[,!istex]) > 0)
    
    ## Assigning unique reads mapping to both a TE and a gene as gene counts
    # Which unique reads overlap to both genes and TEs?
    idxu <- indx & maskuniqaln[mt]
    if (any(idxu)) {
        # ovalnmat[idxu,istex] <- FALSE
        whu <- which(ovalnmat[idxu,istex], arr.ind = TRUE)
        if (is(whu, "integer")) {
            whu <- as.matrix(data.frame(row = 1, col = whu))
        }
        whudf <- cbind(which(idxu)[whu[,"row"]], which(istex)[whu[,"col"]])
        ovalnmat[whudf] <- FALSE
    }
    
    ## Removing overlaps of multi-mapping reads to genes if at least one
    ## alignment of the read overlaps a TE
    idxm <- indx & !maskuniqaln[mt]
    if (any(idxm)) {
        # ovalnmat[idxm,!istex] <- FALSE
        whm <- which(ovalnmat[idxm,!istex], arr.ind = TRUE)
        if (is(whm, "integer")) {
            whm <- as.matrix(data.frame(row = 1, col = whm))
        }
        whmdf <- cbind(which(idxm)[whm[,"row"]], which(!istex)[whm[,"col"]])
        ovalnmat[whmdf] <- FALSE
    }
    
    ovalnmat
}


## private function .countMultiReadsGenes()
## Counts multi-mapping reads mapping to multiple genes by counting fraction
## counts
.countMultiReadsGenes <- function(ttpar, ovalnmat, maskuniqaln, mt, iste,
                                  istex, tx_idx, readids, alnreadids, ov,
                                  uniqcnt) {
    ovalnmat_multig <- ovalnmat[!maskuniqaln[mt], !istex, drop=FALSE]
    yesg <- rowSums2(ovalnmat_multig)>0
    
    ## Computing counts for reads with multiple alignments mapping to different
    ## reads and also for multimapping reads with only 1 alignment mapping to
    ## a gene
    ovalnmat_multig <- ovalnmat_multig[yesg,,drop=FALSE]
    
    # Getting the num of different alignments mapping to a gene for each read
    alnreadids_multig <-alnreadids[unique(queryHits(
                                            ov[!iste[subjectHits(ov)]]))]
    nalnperread <- table(alnreadids_multig) # getting only overlaps from genes
    readids_multig <- readids[!maskuniqaln[mt]][yesg]
    mt_multig <- match(readids_multig, names(nalnperread))
    
    # Counts provided by unique reads for genes to which multimapping 
    # reads map to
    matmultiguniqc <- t(t(ovalnmat_multig)*uniqcnt[tx_idx][!istex])
    rsum <- rowSums2(matmultiguniqc)
    rsum0 <- which(rsum == 0)
    
    # Reads for which the genes to which the read aligns have > 0 counts
    # matmultiguniqc[!rsum0,, drop=FALSE] <- matmultiguniqc[!rsum0,,drop=FALSE]/
    #     rsum[!rsum0]
    j <- matmultiguniqc@i + 1
    matmultiguniqc@x[!(j %in% rsum0)] <- matmultiguniqc@x[!(j %in% rsum0)] / 
        rsum[j[!(j %in% rsum0)]]
    
    
    # Reads for which the genes to which the read aligns have 0 counts
    # matmultiguniqc[rsum0,, drop=FALSE] <- (ovalnmat_multig[rsum0,, drop=FALSE] /
    #     rowSums2(ovalnmat_multig[rsum0,, drop=FALSE]))
    
    matmultiguniqc@x[(j %in% rsum0)] <- ovalnmat_multig@x[(j %in% rsum0)] / 
        rowSums2(ovalnmat_multig)[j[j %in% rsum0]]
    
    # Adjusting for number of alignments
    matmultiguniqc <- matmultiguniqc/as.numeric(nalnperread[mt_multig])
    
    multigcnt <- rep(0L, length(ttpar@features))
    multigcnt[tx_idx][!istex] <- colSums2(matmultiguniqc)
    multigcnt
}

## private function .countUniqueRead()
## Counts unique reads mapping to TEs and genes (if present)
.countUniqueRead <- function(ttpar, ovalnmat, maskuniqaln, mt, tx_idx, istex) {
    uniqcnt <- rep(0L, length(ttpar@features))
    ovalnmatuniq_g <- ovalnmat[maskuniqaln[mt], !istex, drop=FALSE]
    ovalnmatuniq_te <- ovalnmat[maskuniqaln[mt], istex, drop=FALSE]
    ovmultig <- rowSums2(ovalnmatuniq_g) > 1
    ovmultite <- rowSums2(ovalnmatuniq_te) > 1
    
    # Addressing gene counts: count divided by the number of different genes
    if (any(ovmultig)) {
        # Counting reads overlapping only 1 element
        uniqcnt[tx_idx][!istex] <- colSums2(ovalnmatuniq_g[!ovmultig,,drop=FALSE])
        # Counting reads overlapping more than 1 element
        mg <- ovalnmatuniq_g[ovmultig,,drop=FALSE] /
            rowSums2(ovalnmatuniq_g[ovmultig,,drop=FALSE])
        uniqcnt[tx_idx][!istex] <- uniqcnt[tx_idx][!istex] + colSums2(mg)
    
    } else {
        uniqcnt[tx_idx][!istex] <- colSums2(ovalnmatuniq_g)
    }
    
    # Addressing TE counts: count divided by the number of different TEs
    # proportionally to the expression level of each TE provided by unique counts
    if (any(ovmultite)) {
        # Counting reads overlapping only 1 element
        uniqcnt[tx_idx][istex] <- colSums2(ovalnmatuniq_te[!ovmultite,,drop=FALSE])
        # Counting reads overlapping more than 1 element
        mte <- t(t(
            ovalnmatuniq_te[ovmultite,,drop=FALSE])*uniqcnt[tx_idx][istex])
        nocounts <- which(rowSums2(mte) == 0)
        mte[nocounts,,drop=FALSE] <- ovalnmatuniq_te[
            ovmultite,,drop=FALSE][nocounts,]
        mte <- mte/rowSums2(mte)
        uniqcnt[tx_idx][istex] <- uniqcnt[tx_idx][istex] + colSums2(mte)
    
    } else {
        uniqcnt[tx_idx][istex] <- colSums2(ovalnmatuniq_te)
    }
    
    uniqcnt
}


## private function .summarizeCounts()
## Counts unique reads mapping to TEs and genes (if present)
.summarizeCounts <- function(iste, cntvec, uniqcnt, multigcnt, ttpar) {
    
    ## add multi-mapping and unique-mapping counts
    if (!all(iste)) {
        cntvec <- cntvec + uniqcnt + multigcnt
    } else {
        cntvec <- cntvec + uniqcnt
    }
    names(cntvec) <- names(ttpar@features)
    
    cntvec_t <- cntvec[iste]
    ## aggregate TE quantifications if necessary
    if (length(ttpar@aggregateby) > 0) {
        f <- .factoraggregateby(ttpar@features[iste], ttpar@aggregateby)
        stopifnot(length(f) == length(cntvec_t)) ## QC
        cntvec_t <- tapply(cntvec_t, f, sum, na.rm=TRUE)
    }
    
    ## aggregating exon counts to genes
    if (!all(iste)) {
        cntvec <- c(cntvec_t, cntvec[!iste])
    } else {
        cntvec <- cntvec_t
    }
    
    cntvec <- round(cntvec, digits = 8)
    cntvec
}

## private function .adjustalnmultiovTE()
## Adjusts ovalnmat when a multimapping reads have alignments mapping to > 1
## TE, by dividing the count value by the number of different TEs the
## alignment maps to

#' @importFrom S4Vectors queryHits subjectHits
.adjustalnmultiovTE <- function(iste, ov, alnreadids, readids, tx_idx,
                                ovalnmat) {
    isteov <- iste[subjectHits(ov)]
    multialn <- !(.getMaskUniqueAln(queryHits(ov)))
    ovte_multialn <- ov[isteov & multialn]
    mt1_tem <- match(alnreadids[queryHits(ovte_multialn)], readids)
    mt2_tem <- match(subjectHits(ovte_multialn), tx_idx)
    pos_tem <- cbind(mt1_tem, mt2_tem)
    novperaln <- table(queryHits(ovte_multialn))
    mt_multig <- match(queryHits(ovte_multialn), names(novperaln))
    ovalnmat <- as(as(as(ovalnmat, "dMatrix"), "generalMatrix"), "CsparseMatrix")
    ovalnmat[pos_tem] <- ovalnmat[pos_tem] / as.numeric(novperaln[mt_multig])
    ovalnmat
}

#' @importFrom Rsamtools scanBamFlag
.getScanBamFlag_tt <- function(singleEnd) {
    if (singleEnd == TRUE) {
        sbflags <- scanBamFlag(isUnmappedQuery=FALSE, isDuplicate=FALSE,
                               isNotPassingQualityControls=FALSE)
    } else {
        sbflags <- scanBamFlag(isUnmappedQuery=FALSE, isDuplicate=FALSE,
                               isNotPassingQualityControls=FALSE,
                               isProperPair=TRUE)
    }
  sbflags
}

test_TEtranscripts <- function() {
    bamfiles <- list.files(system.file("extdata", package="atena"),
                           pattern="*.bam", full.names=TRUE)
    TE_annot <- readRDS(file=system.file("extdata", "Top28TEs.rds",
                                         package="atena"))
    ttpar <- TEtranscriptsParam(bamfiles, teFeatures=TE_annot,
                                singleEnd=TRUE, ignoreStrand=TRUE,
                                aggregateby="repName")
    ttSE <- qtex(ttpar)
    
    checkEqualsNumeric(dim(ttSE), c(1, 2))
    checkEqualsNumeric(sort(assay(ttSE), decreasing=TRUE), c(150, 123))
}


test_tt_input1 <- function() {
    
    TE_annot <- readRDS(file=system.file("extdata", "Top28TEs.rds",
                                         package="atena"))
    checkException(TEtranscriptsParam(teFeatures=TE_annot, singleEnd=TRUE),
                    "An error prompts when no input BAM file is specified",
                    silent=TRUE)
}


test_tt_input2 <- function() {
    
    bamfiles <- list.files(system.file("extdata", package="atena"),
                           pattern="*.bam", full.names=TRUE)
    checkException(TEtranscriptsParam(bfl=bamfiles, singleEnd=TRUE),
                    "An error prompts when TE annotations are not specified",
                    silent=TRUE)
}


test_exon_summarization <- function() {
    
    bamfiles <- list.files(system.file("extdata", package="atena"),
                           pattern="*.bam", full.names=TRUE)
    TE_annot <- readRDS(file=system.file("extdata", "Top28TEs.rds",
                                         package="atena"))
    
    # Creating an example of gene annotations
    annot_gen <- GRanges(seqnames=rep("2L",10),
                        ranges=IRanges(start=c(1,20,45,80,110,130,150,170,200,220),
                                       width=c(10,20,35,10,5,15,10,25,5,20)),
                        strand="*", 
                        type=rep("exon",10))
    # Setting gene ids
    names(annot_gen) <- paste0("gene",c(rep(1,3),rep(2,4),rep(3,3)))
    annot_gen
    ttpar_gen <- TEtranscriptsParam(bamfiles, teFeatures=TE_annot,
                                    geneFeatures=annot_gen, singleEnd=TRUE,
                                    ignoreStrand=TRUE)
    
    checkTrue(is(features(ttpar_gen), "GRangesList"),
              "Exons are summarized at the gene level in a GRangesList object")
}

#' Build an ERVmap parameter object
#'
#' Build an object of the class \code{ERVmapParam}
#'
#' @param bfl A \code{BamFile} or \code{BamFileList} object, or a character
#' string vector of BAM filenames.
#'
#' @param teFeatures A \code{GRanges} or \code{GRangesList} object with the
#' transposable element (TE) annotated features to be quantified. Elements in 
#' this object should have names, which are used as a grouping factor for 
#' genomic ranges forming a common locus, unless other metadata column names 
#' are specified in the \code{aggregateby} parameter.
#'
#' @param aggregateby Character vector with column names in the annotation
#' to be used to aggregate quantifications. By default, this is an empty
#' vector, which means that the names of the input \code{GRanges} or 
#' \code{GRangesList} object given in the \code{teFeatures} parameter are used
#' to aggregate quantifications.
#' 
#' @param ovMode Character vector indicating the overlapping mode. Available
#' options are: "ovUnion" (default) and "ovIntersectionStrict",
#' which implement the corresponding methods from HTSeq
#' (\url{https://htseq.readthedocs.io/en/release_0.11.1/count.html}).
#' Ambiguous alignments (alignments overlapping > 1 feature) are addressed
#' as in the original ERVmap algorithm.
#'
#' @param geneFeatures (Default NULL) A \code{GRanges} or 
#' \code{GRangesList} object with the
#' gene annotated features to be quantified. Overlaps with unique reads are 
#' first tallied with respect to these gene features. Elements should have 
#' names indicating the gene name/id. In case that \code{geneFeatures} 
#' is a \code{GRanges} and contains
#' a metadata column named \code{type}, only the elements with 
#' \code{type} = \code{exon} are considered for the analysis. Then, exon
#' counts are summarized to the gene level. If NULL, gene expression is
#' not quantified.
#'
#' @param singleEnd (Default TRUE) Logical value indicating if reads are single
#' (\code{TRUE}) or paired-end (\code{FALSE}).
#'
#' @param strandMode (Default 1) Numeric vector which can take values 0, 1 or 
#' 2.
#'   The strand mode is a per-object switch on
#'   \code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
#'   objects that controls the behavior of the strand getter. See
#'   \code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
#'   class for further detail. If \code{singleEnd = TRUE}, then
#'   \code{strandMode} is ignored.
#'
#' @param ignoreStrand (Default TRUE) A logical which defines if the strand
#' should be taken into consideration when computing the overlap between reads
#' and annotated features. When \code{ignore_strand = FALSE}, an aligned read
#' is considered to be overlapping an annotated feature as long as they
#' have a non-empty intersecting genomic range on the same strand, while when
#' \code{ignoreStrand = TRUE} the strand is not considered.
#'
#' @param fragments (Default not \code{singleEnd}) A logical; applied to
#' paired-end data only. When \code{fragments=TRUE}, the read-counting
#' method in the original ERVmap algorithm is applied: each mate of a
#' paired-end read is counted (including ambiguous and not properly paired 
#' reads). When
#' \code{fragments=FALSE}, if the two mates of a paired-end read map to the
#' same element, they are counted as a single hit and singletons, reads with
#' unmapped pairs and other ambiguous or not properly paired fragments are 
#' not counted (see "Pairing criteria" in 
#' \code{\link[GenomicAlignments]{readGAlignments}()}).
#'
#' @param maxMismatchRate (Default 0.02) Numeric value storing the maximum
#' mismatch rate employed by the ERVmap algorithm to discard aligned reads
#' whose rate of sum of hard and soft clipping or whose rate of the edit
#' distance over the genome reference to the length of the read is above this
#' threshold.
#'
#' @param suboptimalAlignmentTag (Default "auto") Character string storing the
#' tag name in the BAM files that stores the suboptimal alignment score used in
#' the third filter of ERVmap; see
#' \href{https://doi.org/10.1073/pnas.1814589115}{Tokuyama et al. (2018)}.
#' The default, \code{suboptimalAlignmentTag="auto"}, first extracts the name
#' of the read mapper software from one or more BAM files. If BAM files were
#' generated by BWA, the suboptimal alignment scores are obtained from a tag
#' called \code{XS}. For other read mappers, the suboptimal alignment score
#' is considered to be missing since, except from BWA, no other aligner
#' provides a tag with suboptimal alignment scores. In this case, the available
#' secondary alignments are used to implement an analogous approach to that
#' of the third ERVmap filter. When \code{suboptimalAlignmentTag="none"}, it
#' also performs the latter approach even when the tag \code{XS} is available.
#' When this parameter is different from \code{"auto"} and \code{"none"}, a tag
#' with the given name is used to extract the suboptimal alignment score.
#'
#' @param suboptimalAlignmentCutoff (Default 5) Numeric value storing the
#' cutoff above which the difference between the alignment score and the
#' suboptimal alignment score is considered sufficiently large to retain the
#' alignment. When this value is set to \code{NA}, the filtering step based on
#' suboptimal alignment scores is skipped.
#' 
#' @param geneCountMode (Default \code{"all"}) Character string indicating if 
#' the ERVmap read filters applied to quantify TEs expression should also be 
#' applied when quantifying gene expression (\code{"ervmap"}) or not 
#' (\code{"all"}), in which case all primary alignments mapping to genes are 
#' counted.
#'
#' @param verbose (Default \code{TRUE}) Logical value indicating whether to
#' report progress.
#'
#' @details
#' This is the constructor function for objects of the class
#' \code{ERVmapParam-class}. This type of object is the input to the
#' function \code{\link{qtex}()} for quantifying expression of transposable
#' elements using the ERVmap method
#' \href{https://doi.org/10.1073/pnas.1814589115}{Tokuyama et al. (2018)}. The
#' ERVmap algorithm processes reads following conservative filtering criteria
#' to provide reliable raw count data for each TE.
#' 
#' @return A \linkS4class{ERVmapParam} object.
#'
#' @examples
#' bamfiles <- list.files(system.file("extdata", package="atena"),
#'                        pattern="*.bam", full.names=TRUE)
#' \dontrun{
#' rmskat <- annotaTEs(genome="dm6", parsefun=rmskatenaparser,
#'                     strict=FALSE, insert=500)
#' rmskLTR <- getLTRs(rmskat, relLength=0.8,
#'                    fullLength=TRUE,
#'                    partial=TRUE,
#'                    otherLTR=TRUE)
#' }
#'
#' ## DO NOT TYPE THIS INSTRUCTION, WHICH JUST LOADS A PRE-COMPUTED ANNOTATION
#' ## YOU SHOULD USE THE INSTRUCTIONS ABOVE TO FETCH ANNOTATIONS
#' rmskLTR <- readRDS(system.file("extdata", "rmskatLTRrlen80flenpartoth.rds",
#'                                package="atena"))
#'
#' ## build a parameter object for ERVmap
#' empar <- ERVmapParam(bamfiles,
#'                      teFeatures=rmskLTR,
#'                      singleEnd=TRUE,
#'                      ignoreStrand=TRUE,
#'                      suboptimalAlignmentCutoff=NA)
#' empar
#'
#' @references
#' Tokuyama M et al. ERVmap analysis reveals genome-wide transcription of human
#' endogenous retroviruses. PNAS. 2018;115(50):12565-12572. DOI:
#' \url{https://doi.org/10.1073/pnas.1814589115}
#'
#' @importFrom methods is new
#' @importFrom BiocGenerics path
#' @importFrom cli cli_alert_info cli_alert_success
#' @export
#' @rdname ERVmapParam-class
ERVmapParam <- function(bfl, teFeatures, aggregateby=character(0),
                        ovMode="ovUnion",
                        geneFeatures=NULL,
                        singleEnd=TRUE,
                        ignoreStrand=TRUE,
                        strandMode=1L,
                        fragments=!singleEnd,
                        maxMismatchRate=0.02,
                        suboptimalAlignmentTag="auto",
                        suboptimalAlignmentCutoff=5,
                        geneCountMode="all",
                        verbose=TRUE) {
    
    if (verbose)
        cli_alert_info("Locating BAM files")
    bfl <- .checkBamFileListArgs(bfl, singleEnd, fragments)
    
    if (!ovMode %in% c("ovUnion","ovIntersectionStrict"))
      stop("'ovMode' should be one of 'ovUnion', 'ovIntersectionStrict'")
    
    readmapper <- .checkBamReadMapper(path(bfl))
    
    if (verbose)
        cli_alert_info("Processing features")
    teFeaturesobjname <- deparse(substitute(teFeatures))
    geneFeaturesobjname <- deparse(substitute(geneFeatures))
    features <- .processFeatures(teFeatures, teFeaturesobjname,
                                 geneFeatures, geneFeaturesobjname,
                                 aggregateby, aggregateexons=TRUE)
    
    obj <- new("ERVmapParam", bfl=bfl, features=features,
               aggregateby=aggregateby, ovMode=ovMode,
               singleEnd=singleEnd, ignoreStrand=ignoreStrand,
               strandMode=as.integer(strandMode), fragments=fragments,
               maxMismatchRate=maxMismatchRate, readMapper=readmapper,
               suboptimalAlignmentTag=suboptimalAlignmentTag,
               suboptimalAlignmentCutoff=as.numeric(suboptimalAlignmentCutoff),
               geneCountMode=geneCountMode)
    if (verbose)
        cli_alert_success("Parameter object successfully created")
    obj
}

#' @param object A \linkS4class{ERVmapParam} object.
#'
#' @importFrom methods show
#' @importFrom Seqinfo seqlevels
#'
#' @export
#' @aliases show,ERVmapParam-method
#' @rdname ERVmapParam-class
setMethod("show", "ERVmapParam",
        function(object) {
            cat(class(object), "object\n")
            cat(sprintf("# BAM files (%d): %s\n", length(object@bfl),
                        .pprintnames(names(object@bfl))))
            cat(sprintf("# features (%d): %s\n", length(features(object)),
                        ifelse(is.null(names(features(object))),
                               paste("on",
                                     .pprintnames(seqlevels(features(object)))),
                               .pprintnames(names(features(object))))))
            cat(sprintf("# %s, %s",
                        ifelse(object@singleEnd, "single-end", "paired-end"),
                        ifelse(object@ignoreStrand, "unstranded", "stranded")))
            if (!object@ignoreStrand)
                cat(sprintf(" (strandMode=%d)", object@strandMode))
            if (!object@singleEnd)
                cat(sprintf(", %s",
                            ifelse(object@fragments, 
                                "counting each paired-end mate",
                                "counting both paired-end mates")))
            cat("\n")
        })

#' @importFrom BiocParallel SerialParam bplapply
#' @importFrom SummarizedExperiment SummarizedExperiment
#' @export
#' @aliases qtex
#' @aliases qtex,ERVmapParam-method
#' @rdname qtex
setMethod("qtex", "ERVmapParam",
        function(x, phenodata=NULL, mode=ovUnion, yieldSize=1e6L, verbose=1,
            BPPARAM=SerialParam(progressbar=ifelse(verbose==1, TRUE, FALSE))) {
            .checkPhenodata(phenodata, length(x@bfl))
            
            cnt <- bplapply(x@bfl, .qtex_ervmap, empar=x, mode=x@ovMode,
                            yieldSize=yieldSize, verbose=verbose,
                            BPPARAM=BPPARAM)
            cnt <- do.call("cbind", cnt)
            cdata <- .createColumnData(cnt, phenodata)
            colnames(cnt) <- rownames(cdata)
            
            features <- .consolidateFeatures(x, rownames(cnt))
            
            SummarizedExperiment(assays=list(counts=cnt),
                                 rowRanges=features,
                                 colData=cdata)
        })



#' @importFrom BiocGenerics basename path
#' @importFrom Rsamtools ScanBamParam yieldSize yieldSize<-
#' @importFrom methods formalArgs
#' @importFrom S4Vectors mcols
#' @importFrom IRanges findOverlapPairs
#' @importFrom GenomicRanges pintersect
.qtex_ervmap <- function(bf, empar, mode, yieldSize=1000000, verbose) {
    iste <- mcols(features(empar))$isTE
    if (any(duplicated(names(features(empar)[iste]))))
        stop(".qtex_ervmap: duplicated names in TE annotations.")

    mode <- match.fun(mode)
    readfun <- .getReadFunction(empar@singleEnd, empar@fragments)
    
    ## suboptimalAlignmentTag set to 'auto' implies following the original
    ## ERVmap algorithm; when the read mapper is BWA, it is assumed that
    ## the tag 'XS' stores the suboptimal alignment score.
    soatag <- .getsoatag(empar@suboptimalAlignmentTag, empar@readMapper)
    avtags <- .getavtags(empar, bf, soatag)
    avgene <- !all(iste)
    sbflags <- .setBamFlags(bf, empar, avtags, avgene)
    avsoas <- .soasAvail(empar@suboptimalAlignmentCutoff, soatag, avtags, bf,
                         verbose)
    if ((!is.na(empar@suboptimalAlignmentCutoff)) & 
        isTRUE(soatag %in% avtags)) {
            empar@suboptimalAlignmentTag <- soatag
    }
    param <- ScanBamParam(flag=sbflags, what="flag", tag=avtags)
    
    ## 'ov' is a 'Hits' object with the overlaps between reads and features
    ov <- Hits(nLnode=0, nRnode=length(features(empar)), sort.by.query=TRUE)
    ## 'ovdiscard' is a 'Hits' object with the overlaps of discarded reads
    ovdiscard <- Hits(nLnode=0, nRnode=length(features(empar)[!iste]),
                      sort.by.query=TRUE)
    ## 'salnmask' is a logical mask flagging whether an alignment is secondary
    salnmask <- logical(0)
    salnbestAS <- integer(0)
    ## 'alnAS' stores the alignment score
    alnAS <- integer(0)
    ## 'alnNH' stores the NH tag
    alnNH <- integer(0)
    ## 'readids' stores unique read identifiers only once, while 'alnreadidx'
    ## stores for each read index of the corresponding unique read identifier
    readids <- character(0) 
    alnreadidx <- integer(0) # index in 'readids' for alignm. passing filters
    readidx <- integer(0) # index in 'readids' for all alignm. (passing & not)
    thisalnAS <- integer(0)
    alnreadids <- character(0)
    n <- nfiltered <- 0
    
    strand_arg <- "strandMode" %in% formalArgs(readfun)
    yieldSize(bf) <- yieldSize
    cntvec <- .ervmapQuantExpress(bf, empar, mode, readfun, iste, param,avsoas,
                                  strand_arg, avgene, n, ov, ovdiscard,
                                  salnmask, salnbestAS, alnAS, alnNH, readids,
                                  alnreadidx, readidx, thisalnAS, alnreadids,
                                  nfiltered, verbose, avtags)
    cntvec
}


#' @importFrom methods formalArgs as
#' @importFrom IRanges findOverlapPairs
#' @importFrom GenomicRanges pintersect
#' @importFrom GenomicAlignments findOverlaps
#' @importFrom S4Vectors countQueryHits countSubjectHits
.ervmapQuantExpress <- function(bf, empar, mode, readfun, iste, param, avsoas,
                                strand_arg, avgene, n, ov, ovdiscard, salnmask,
                                salnbestAS, alnAS, alnNH, readids, alnreadidx,
                                readidx, thisalnAS, alnreadids, nfiltered,
                                verbose, avtags) {
    open(bf)
    while (length(alnreads <- do.call(readfun,
                                c(list(file=bf), list(param=param),
                                 list(strandMode=empar@strandMode)[strand_arg],
                                 list(use.names=(!avsoas || avgene)))))) {
        alnreads <- .matchSeqinfo(alnreads, features(empar))
        n <- n + length(alnreads)
        thissalnmask <- .secondaryAlignmentMask(alnreads) ## second. aln mask
        mask <- .pass2Filters(alnreads, empar) ## which alignments pass flt 1&2
    
        if (!is.na(empar@suboptimalAlignmentCutoff)) {
            thisalnAS <- .getAlignmentTagScore(alnreads, tag="AS")
            mask <- .pass3Filter(mask, alnreads, empar, thisalnAS, thissalnmask,
                                avsoas)
        }
    
        if (avgene || (!is.na(empar@suboptimalAlignmentCutoff) & !avsoas)) {
            alnreadids <- .getAlnreadids(alnreads, empar@fragments)
        }
    
        mask[thissalnmask] <- TRUE # Setting TRUE in mask to secondary alignments
        alnreadsdiscard <- .getreadsdiscard(empar, alnreads, mask, thissalnmask)
        alnreads <- .filteralnreads(empar,alnreads,mask,thissalnmask,avsoas,avgene)
        thisov <- mode(alnreads, features(empar), ignoreStrand=empar@ignoreStrand,
                       inter.feature=FALSE)
        ov <- .appendHits(ov, thisov)
    
        if (avgene && empar@geneCountMode == "all") {
            ## calculate and store overlaps between the discarded reads & genes
            thisovdiscard <- mode(alnreadsdiscard, features(empar)[!iste],
                                  ignoreStrand=empar@ignoreStrand, 
                                  inter.feature=FALSE)
            ovdiscard <- .appendHits(ovdiscard, thisovdiscard)
        }
    
        ## Reading NH tag if secondary alignments are not available
        alnNH <- .fetchNHtag(thissalnmask, avgene, avtags, alnreads)
        salnmask <- c(salnmask, thissalnmask[mask]) # secondary alignment mask
    
        if (!avsoas && !is.na(empar@suboptimalAlignmentCutoff)) {
            ## store best secondary alignment score
            salnbestAS <- .findSuboptAlignScore(thisalnAS, thissalnmask, 
                                                alnreadids, salnbestAS)
            ## filter and store alignment scores 
            alnAS <- c(alnAS, thisalnAS[mask])
        }
    
        ## filter now read identifiers
        listreadids <- .filterReadIDs(avgene, empar, avsoas, alnreadids,
                                      readids, readidx, alnreadidx, mask)
        readids <- listreadids$readids
        readidx <- listreadids$readidx
        alnreadidx <- listreadids$alnreadidx
    
        if (verbose > 1) {
            nfiltered <- nfiltered + sum(mask & !thissalnmask)
            message(sprintf("%s: %d alignments processed (%d are primary and pass filters).\n",
                            basename(path(bf)), n, nfiltered))
        }
    }
    on.exit(close(bf))
    if (length(ov) == 0) {
        stop(".ervmapQuantExpress: no overlaps were found between reads and features")
    }
    ## Expression quantification
    cntvec <- .ervmapGetCounts(avsoas, salnmask, empar, avgene, ov, alnreadidx,
                               alnAS, salnbestAS, readidx, mask, alnNH, readids,
                               bf, ovdiscard, n, verbose, iste)
    cntvec
}


## Function to know which tags are present in the BAM file (single-end).
## Input is a BamFile object. Returns a data.frame with tags as columns and the
## BAM file as row, indicating if each tag is present in the BAM file (TRUE) or
## not (FALSE).

#' @importFrom Rsamtools scanBamFlag ScanBamParam yieldSize yieldSize<-
#' @importFrom S4Vectors mcols
.get_tags_in_BAM_singleend <- function(bf, soatag="XS") {
    
    yieldSize(bf) <- 1000
    sbflags <- scanBamFlag(isUnmappedQuery = FALSE,
                            isDuplicate = FALSE,
                            isNotPassingQualityControls = FALSE,
                            isSupplementaryAlignment = FALSE,
                            ## XS is only set by BWA for primary alignments
                            isSecondaryAlignment=FALSE)
    
    param <- ScanBamParam(flag = sbflags,
                            tag = c("nM", "NM", "AS", "NH", soatag))
    
    tags <- param@tag
    tags_df <- data.frame(matrix(nrow = length(bf), ncol = length(tags)))
    colnames(tags_df) <- tags
    
    open(bf)
    r_test <- readGAlignments(bf, param = param)
    close(bf)
    
    # Testing, for each sample, if the files contain the different tags
    tags_df[, "NH"] <- ifelse(all(is.na(mcols(r_test)$NH)), FALSE, TRUE)
    tags_df[, "NM"] <- ifelse(all(is.na(mcols(r_test)$NM)), FALSE, TRUE)
    tags_df[, "nM"] <- ifelse(all(is.na(mcols(r_test)$nM)), FALSE, TRUE)
    tags_df[, "AS"] <- ifelse(all(is.na(mcols(r_test)$AS)), FALSE, TRUE)
    if (!is.null(soatag))
        tags_df[, soatag] <- ifelse(all(grepl("\\d", 
                                                (mcols(r_test)[[soatag]]))),
                                    TRUE, FALSE)
    
    return(tags_df)
}



## Function to know which tags are present in the BAM file (paired-end).
## Input is a BamFile object. Returns a data.frame with tags as columns and the
## BAM file as row, indicating if each tag is present in the BAM file (TRUE) or
## not (FALSE).

#' @importFrom Rsamtools scanBamFlag ScanBamParam yieldSize yieldSize<- asMates asMates<-
#' @importFrom S4Vectors mcols
.get_tags_in_BAM_pairedend <- function(bf, empar, soatag="XS") {
    yieldSize(bf) <- 1000
    asMates(bf) <- TRUE
    sbflags <- scanBamFlag(isUnmappedQuery = FALSE,
                            isProperPair = TRUE,
                            isDuplicate = FALSE,
                            isNotPassingQualityControls = FALSE,
                            isSupplementaryAlignment = FALSE,
                            ## XS is only set by BWA for primary alignments
                            isSecondaryAlignment=FALSE)
    
    param <- ScanBamParam(flag = sbflags,
                            tag = c("nM", "NM", "AS", "NH", soatag))
    
    tags <- param@tag
    tags_df <- data.frame(matrix(nrow = length(bf), ncol = length(tags)))
    colnames(tags_df) <- tags
    
    open(bf)
    r_test <- readGAlignmentPairs(bf, param = param, 
                                    strandMode = empar@strandMode)
    close(bf)
    
    # Testing, for each sample, if the files contain the different tags
    tags_df[, "NH"] <- ifelse(all(is.na(mcols(first(r_test))$NH)), FALSE, TRUE)
    tags_df[, "NM"] <- ifelse(all(is.na(mcols(first(r_test))$NM)), FALSE, TRUE)
    tags_df[, "nM"] <- ifelse(all(is.na(mcols(first(r_test))$nM)), FALSE, TRUE)
    tags_df[, "AS"] <- ifelse(all(is.na(mcols(first(r_test))$AS)), FALSE, TRUE)
    if (!is.null(soatag))
        tags_df[, soatag] <- ifelse(all(grepl("\\d", 
                                            (mcols(first(r_test))[[soatag]]))),
                                    TRUE, FALSE)
    
    return(tags_df)
}




#' @importFrom Rsamtools scanBamFlag
.setBamFlags <- function(bf, empar, avtags, avgene) {
    sbflags <- scanBamFlag()
    
    if (empar@singleEnd == TRUE | empar@fragments == TRUE) {
        ## if we use the 'XS' tag then do not read secondary alignments
        if ("XS" %in% avtags && empar@readMapper == "bwa" &&
            empar@suboptimalAlignmentTag %in% c("auto", "XS") && !avgene)
            sbflags <- scanBamFlag(isUnmappedQuery=FALSE,
                                    isDuplicate=FALSE,
                                    isNotPassingQualityControls=FALSE,
                                    isSupplementaryAlignment=FALSE,
                                    isSecondaryAlignment=FALSE)
        
        else ## otherwise do not exclude secondary alignments
            sbflags <- scanBamFlag(isUnmappedQuery=FALSE,
                                    isDuplicate=FALSE,
                                    isNotPassingQualityControls=FALSE,
                                    isSupplementaryAlignment=FALSE)
    } else {
        ## if we use the 'XS' tag then do not read secondary alignments
        if ("XS" %in% avtags && empar@readMapper == "bwa" &&
            empar@suboptimalAlignmentTag %in% c("auto", "XS") && !avgene)
            sbflags <- scanBamFlag(isUnmappedQuery=FALSE,
                                   isDuplicate=FALSE,
                                   isNotPassingQualityControls=FALSE,
                                   isSupplementaryAlignment=FALSE,
                                   isSecondaryAlignment=FALSE,
                                   isProperPair=TRUE)
        
        else ## otherwise do not exclude secondary alignments
            sbflags <- scanBamFlag(isUnmappedQuery=FALSE,
                                   isDuplicate=FALSE,
                                   isNotPassingQualityControls=FALSE,
                                   isSupplementaryAlignment=FALSE,
                                   isProperPair=TRUE)
    }
    sbflags
}




#' @importFrom S4Vectors queryHits subjectHits
#' @importFrom Matrix rowSums
#' @importFrom sparseMatrixStats rowMaxs colSums2 rowSums2
.suboptimalAlignmentFilteringCounts <- function(empar, ov, alnreadidx, alnAS,
                                                salnmask, salnbestAS) {
    ## suboptimal alignment filtering using the available secondary
    ## alignments. according to Tokuyama et al. (2018) pg. 12571.,
    ## reads/fragments are kept if the difference between the alignment
    ## score from BWA (AS) and the suboptimal alignment score from BWA
    ## (field XS) >= 5. here we calculate the suboptimal alignment score
    ## by searching for the secondary alignment with the highest score.
    
    ## fetch all different read identifiers from the overlapping alignments
    rd_idx <- sort(unique(alnreadidx[queryHits(ov)]))
    
    ## fetch all different transcripts from the overlapping alignments
    tx_idx <- sort(unique(subjectHits(ov)))
    
    ## build matrices with primary alignment flags, alignment scores (AS) 
    ## and best secondary alignment AS
    palnmat <- .buildOvValuesMatrix(empar, ov, !salnmask, alnreadidx,
                                    rd_idx, tx_idx)
    asmat <- .buildOvValuesMatrix(empar, ov, alnAS, alnreadidx,
                                    rd_idx, tx_idx)
    salnbestasmat <- .buildOvValuesMatrix(empar, ov, salnbestAS, alnreadidx,
                                            rd_idx, tx_idx)
    
    ## AS from primary alignments
    asprimaryaln <- rowMaxs(palnmat * asmat)
    ## fetch maximum AS from secondary alignments
    salnmaxas <- rowMaxs(salnbestasmat)
    salnmaxas[salnmaxas == 0] <- -99
    mask <- ((asprimaryaln - salnmaxas) >= empar@suboptimalAlignmentCutoff)
    
    palnmat <- palnmat[mask, ]
    cntvec <- rep(0, length(features(empar)))
    cntvec[tx_idx] <- colSums2(palnmat)
    
    cntvec
}

#' @importFrom S4Vectors mcols first second
#' @importFrom Rsamtools bamFlagTest
.secondaryAlignmentMask <- function(aln) {
    mask <- NULL
    if (is(aln, "GAlignments"))
        mask <- bamFlagTest(mcols(aln)$flag, "isSecondaryAlignment")
    else if (is(aln, "GAlignmentPairs")) {
        mask <- bamFlagTest(mcols(first(aln))$flag, "isSecondaryAlignment") |
            bamFlagTest(mcols(second(aln))$flag, "isSecondaryAlignment")
    } else if (is(aln, "GAlignmentsList")) {
        mask <- bamFlagTest(mcols(unlist(aln, use.names=FALSE))$flag,
                            "isSecondaryAlignment")
    } else
        stop(sprintf(".secondaryAlignmentMask: wrong class %s\n", class(aln)))
    
    mask
}

#' @importFrom S4Vectors mcols first second
.getAlignmentTagScore <- function(aln, tag, pairedsummaryfun=pmin.int) {
    if (is(aln, "GAlignments"))
        score <- as.integer(mcols(aln)[[tag]])
    else if (is(aln, "GAlignmentPairs")) {
        ## take the minimum score from both mates
        score <- pairedsummaryfun(as.integer(mcols(first(aln))[[tag]]), 
                                    mcols(second(aln))[[tag]])
    } else if (is(aln, "GAlignmentsList")) {
        score <- as.integer(mcols(unlist(aln, use.names=FALSE))[[tag]])
    } else
        stop(sprintf(".getAlignmentTagScore: wrong class %s\n", class(aln)))
    
    as.integer(score)
}


#' @importFrom S4Vectors mcols
.fetchNMtag <- function(aln) {
    nmtag <- "NM"
    if (is(aln, "GAlignments")) {
        if (is.null(mcols(aln)[[nmtag]]) || all(is.na(mcols(aln)[[nmtag]]))) {
            nmtag <- "nM"
            if (is.null(mcols(aln)[[nmtag]]) || all(is.na(mcols(aln)[[nmtag]])))
                stop("no NM or nM tag in BAM file.")
        }
    } else if (is(aln, "GAlignmentPairs")) {
        if (is.null(mcols(first(aln))[[nmtag]]) || 
            all(is.na(mcols(first(aln))[[nmtag]]))) {
            nmtag <- "nM"
            if (is.null(mcols(first(aln))[[nmtag]]) || 
                all(is.na(mcols(first(aln))[[nmtag]])))
                stop("no NM or nM tag in BAM file.")
        }
    } else if (is(aln, "GAlignmentsList")) {
        if (is.null(mcols(unlist(aln))[[nmtag]]) || 
            all(is.na(mcols(unlist(aln))[[nmtag]]))) {
            nmtag <- "nM"
            if (is.null(mcols(unlist(aln))[[nmtag]]) || 
                all(is.na(mcols(unlist(aln))[[nmtag]])))
                stop("no NM or nM tag in BAM file.")
        }
    } else
        stop(sprintf(".fetchNMtag: wrong class %s\n", class(aln)))
    
    nmtag
}

#' @importFrom S4Vectors mcols first second
#' @importFrom IRanges relist
.getAlignmentMismatches <- function(aln) {
    mm <- NULL
    nmtag <- .fetchNMtag(aln)
    
    if (is(aln, "GAlignments"))
        mm <- mcols(aln)[[nmtag]]
    else if (is(aln, "GAlignmentPairs")) {
        ## take the celing of the mean of mismatches from both mates
        mm1 <- as.integer(mcols(first(aln))[[nmtag]])
        mm2 <- as.integer(mcols(second(aln))[[nmtag]])
        mm1[is.na(mm1)] <- mm2[is.na(mm1)]
        mm2[is.na(mm2)] <- mm1[is.na(mm2)]
        mm <- ceiling((mm1 + mm2) / 2)
    } else if (is(aln, "GAlignmentsList")) {
        mm <- mcols(unlist(aln, use.names=FALSE))[[nmtag]]
    } else
        stop(sprintf(".getAlignmentMismatches: wrong class %s\n", class(aln)))
    
    as.integer(mm)
}

#' @importFrom S4Vectors mcols first second
#' @importFrom IRanges relist
#' @importFrom GenomicAlignments qwidth
#' @importClassesFrom GenomicAlignments GAlignments
.getAlignmentQueryWidth <- function(aln) {
    qw <- NULL
    if (is(aln, "GAlignments"))
        qw <- qwidth(aln)
    else if (is(aln, "GAlignmentPairs")) {
        ## take the ceiling of the mean of query widths from both mates
        qw1 <- qwidth(first(aln))
        qw2 <- qwidth(second(aln))
        qw1[is.na(qw1)] <- qw2[is.na(qw1)]
        qw2[is.na(qw2)] <- qw1[is.na(qw2)]
        qw <- ceiling((qw1 + qw2) / 2)
    } else if (is(aln, "GAlignmentsList")) {
        qw <- qwidth(unlist(aln, use.names=FALSE))
    } else
        stop(sprintf(".getAlignmentQueryWidth: wrong class %s\n", class(aln)))
    
    as.integer(qw)
}

#' @importFrom S4Vectors mcols first second List
#' @importFrom GenomicAlignments cigar explodeCigarOpLengths
.getAlignmentSumClipping <- function(aln) {
    sc <- NULL
    if (is(aln, "GAlignments"))
        sc <- vapply(explodeCigarOpLengths(cigar(aln), ops=c("H", "S")), 
                    FUN.VALUE = integer(1), FUN = sum)
    else if (is(aln, "GAlignmentPairs")) {
        ## take the ceiling of the mean of query widths from both mates
        sc1 <- vapply(explodeCigarOpLengths(cigar(first(aln)),ops=c("H", "S")), 
                    FUN.VALUE = integer(1), FUN = sum)
        sc2 <- vapply(explodeCigarOpLengths(cigar(second(aln)), 
                                            ops=c("H", "S")), 
                        FUN.VALUE = integer(1), FUN = sum)
        sc1[is.na(sc1)] <- sc2[is.na(sc1)]
        sc2[is.na(sc2)] <- sc1[is.na(sc2)]
        sc <- ceiling((sc1 + sc2) / 2)
    } else if (is(aln, "GAlignmentsList")) {
        sc <- vapply(explodeCigarOpLengths(cigar(unlist(aln, use.names=FALSE)),
                                            ops=c("H", "S")),
                        FUN.VALUE = integer(1), FUN = sum)
    } else
        stop(sprintf(".getAlignmentSumClipping: wrong class %s\n", class(aln)))
    
    as.integer(sc)
}

#' @importFrom S4Vectors queryHits subjectHits
#' @importFrom Matrix Matrix
# .buildOvValuesMatrix <- function(ov, values, aridx, ridx, fidx) {
#     stopifnot(class(values) %in% c("logical", "integer", "numeric")) ## QC
#     ovmat <- Matrix(do.call(class(values), list(1)),
#                     nrow=length(ridx), ncol=length(fidx))
#     mt1 <- match(aridx[queryHits(ov)], ridx)
#     mt2 <- match(subjectHits(ov), fidx)
#     ovmat[cbind(mt1, mt2)] <- values[queryHits(ov)]
#     
#     ovmat
# }

## Obtains the name of the tag containing the suboptimal alignment score, if
## present
.getsoatag <- function(suboptimalAlignmentTag, readMapper) {
    soatag <- NULL
    if (suboptimalAlignmentTag == "auto" && readMapper == "bwa") {
        soatag <- "XS"
    } else if (nchar(suboptimalAlignmentTag) == 2) {
        soatag <- suboptimalAlignmentTag
    }
    soatag
}

## Calls internal functions to obtain the name of the available tags depending
## on whether the data is single- or paired-end
#' @importFrom BiocGenerics basename path
.getavtags <- function(empar, bf, soatag) {
    avtags <- character(0)
    if (empar@singleEnd)
        avtags <- .get_tags_in_BAM_singleend(bf, soatag=soatag)
    else
        avtags <- .get_tags_in_BAM_pairedend(bf, empar=empar, soatag=soatag)
    
    avtags <- colnames(avtags)[unlist(avtags)]
    if (!empar@suboptimalAlignmentTag %in% c("auto", "none", avtags))
        stop(sprintf("suboptimal alignment tag '%s' not available in BAM file %s.",
                    empar@suboptimalAlignmentTag, basename(path(bf))))
    avtags
}


## Looks if the suboptimal alignment score tag reported by the user is present
## in the BAM file.
#' @importFrom BiocGenerics basename path
.soasAvail <- function(suboptimalAlignmentCutoff, soatag, avtags, bf, verbose) {
    avsoas <- FALSE 
    if (!is.na(suboptimalAlignmentCutoff)) {
        if (isTRUE(soatag %in% avtags)) {
            avsoas <- TRUE
            if (verbose > 1) {
                fstr <- "%s: using suboptimal alignment scores from tag '%s'.\n"
                message(sprintf(fstr, basename(path(bf)), soatag))
            }
        } else if (verbose > 1) {
            fstr <- paste("%s: suboptimal alignment filtering based on",
                          "secondary alignments.\n")
            message(sprintf(fstr, basename(path(bf))))
        }
    } else if (verbose > 1) {
        msg <- paste("suboptimal alignment cutoff value is 'NA', skipping",
                     "suboptimal alignment filtering.\n")
        message(msg)
    }
    
    avsoas
}


## Obtains the read identifier of the alignment and, when fragments = TRUE,
## sets a different name to each mate.
.getAlnreadids <- function(alnreads, fragments) {
    alnreadids <- character(0)
    alnreadids <- names(alnreads)
    if (fragments) ## with fragments each mate gets a different read identifier
        alnreadids <- paste(rep(make.names(names(alnreads)),lengths(alnreads)),
                            sequence(lengths(alnreads)), sep=".")
    alnreadids
}


## Find the suboptimal alignment score from the AS of the secondary alignments.
.findSuboptAlignScore <- function(thisalnAS, thissalnmask, alnreadids,
                                    salnbestAS) {
    maxsaasbyreadid <- split(thisalnAS[thissalnmask], alnreadids[thissalnmask])
    maxsaasbyreadid <- vapply(maxsaasbyreadid, FUN.VALUE = integer(1), FUN=max)
    mt <- match(names(maxsaasbyreadid), names(salnbestAS))
    if (any(is.na(mt))) {
        tmpas <- rep(NA_integer_, sum(is.na(mt)))
        names(tmpas) <- names(maxsaasbyreadid)[is.na(mt)]
        salnbestAS <- c(salnbestAS, tmpas)
    }
    mt <- match(names(maxsaasbyreadid), names(salnbestAS))
    stopifnot(all(!is.na(mt))) ## QC
    salnbestAS[mt] <- pmax.int(salnbestAS[mt], maxsaasbyreadid, na.rm=TRUE)
    salnbestAS
}



#' @importFrom Matrix rowSums
#' @importFrom sparseMatrixStats rowSums2
.adjustcommonTEgene <- function(palnmatfilt, readidx, mask, ov, maskthird, 
                                empar, alnreadidx, rd_idx, tx_idx, alnNH,
                                iste) {
    
    ## Identifying multi-mapping reads
    if (length(alnNH) > 0) {
        maskMulti <- alnNH > 2
    } else {
        maskMulti <- duplicated(readidx) | duplicated(readidx, fromLast = TRUE)
        maskMulti <- maskMulti[mask]
    }
    
    maskMultimat <- .buildOvValuesMatrix(empar, ov, maskMulti, alnreadidx,
                                         rd_idx, tx_idx)
    maskMultimat <- maskMultimat[maskthird, ]
    maskUniqmat <- .buildOvValuesMatrix(empar, ov, !maskMulti, alnreadidx,
                                        rd_idx, tx_idx)
    maskUniqmat <- maskUniqmat[maskthird, ]
    
    istex <- iste[tx_idx]
    ## which reads overlap both genes and TEs?
    idx <- (rowSums2(palnmatfilt[,istex]) > 0) &
        (rowSums2(palnmatfilt[,!istex])>0)
    ## of these, keep only the overlap with the TE in case of multimapping read
    palnmatfilt[idx,istex, drop = FALSE] <- as(as(as(
        maskMultimat[idx,istex, drop=FALSE]*palnmatfilt[idx,istex, drop=FALSE],
        "lMatrix"), "generalMatrix"), "CsparseMatrix")
    ## of these, keep only the overlap with the gene in case of a unique read
    palnmatfilt[idx,!istex, drop = FALSE] <- as(as(as(
        maskUniqmat[idx, !istex, drop=FALSE]*palnmatfilt[idx, !istex, drop=FALSE],
        "lMatrix"), "generalMatrix"), "CsparseMatrix")
    # sum((rowSums(palnmatfilt[,istex]) > 0) & (rowSums(palnmatfilt[,!istex]) > 0)) # [1] 0 (QC) 
    
    ## when multi-mapping / unique reads cannot be identified because there is
    ## no NH and no secondary alignments in the BAM file, 'maskMulti' is FALSE
    ## for all reads, which causes all reads aligned to overlapping genes and
    ## TEs to be assigned to genes.
    
    palnmatfilt
}


#' @importFrom S4Vectors queryHits subjectHits
#' @importFrom Matrix rowSums
#' @importFrom sparseMatrixStats rowMaxs colSums2 rowSums2
.countbymatrix <- function(empar, ov, alnreadidx, salnmask, alnAS, salnbestAS, 
                            avgene, applysoasfilter, readidx, mask, alnNH, 
                            iste) {
    ## fetch all different read identifiers from the overlapping alignments
    rd_idx <- sort(unique(alnreadidx[queryHits(ov)]))
    ## fetch all different transcripts from the overlapping alignments
    tx_idx <- sort(unique(subjectHits(ov)))
    ## build matrix with overlaps of primary alignments
    palnmat <- .buildOvValuesMatrix(empar, ov, !salnmask, alnreadidx,
                                    rd_idx, tx_idx)
    
    maskthird <- rep(TRUE, nrow(palnmat))
    
    if (isTRUE(applysoasfilter)) {
        ## build matrices with alignment scores (AS) and best secondary AS
        asmat <- .buildOvValuesMatrix(empar, ov, alnAS, alnreadidx, rd_idx,
                                        tx_idx)
        salnbestasmat <- .buildOvValuesMatrix(empar, ov, salnbestAS,
                                                alnreadidx, rd_idx, tx_idx)
        ## AS from primary alignments
        asprimaryaln <- rowMaxs(palnmat * asmat)
        ## fetch maximum AS from secondary alignments
        salnmaxas <- rowMaxs(salnbestasmat)
        salnmaxas[salnmaxas == 0] <- -99
        maskthird <- ((asprimaryaln - salnmaxas) >= 
                        empar@suboptimalAlignmentCutoff)
    }
    
    cntvec <- rep(0, length(features(empar)))
    
    if (isTRUE(avgene)) {
        palnmatadj <- .adjustcommonTEgene(palnmatfilt = palnmat[maskthird, ], 
                                        readidx, mask, ov, maskthird, empar,
                                        alnreadidx, rd_idx, tx_idx, alnNH,iste)
        cntvec[tx_idx] <- colSums2(palnmatadj)
        ## counting counts of discarded reads mapping to genes
        if (empar@geneCountMode == "all" && applysoasfilter) {
            cntvecdisc <- rep(0, length(features(empar)))
            cntvecdisc[tx_idx] <- colSums2(palnmat[!maskthird, ])
            cntvec <- cntvec + cntvecdisc
        }
        
    } else {
        cntvec[tx_idx] <- colSums2(palnmat[maskthird, ])
    }
    
    cntvec
}


.filteralnreads <- function(empar, alnreads, mask, thissalnmask,
                                avsoas, avgene) {
    
    if (empar@fragments) {
        if ((avsoas || is.na(empar@suboptimalAlignmentCutoff)) && !avgene) {
            alnreads <- unlist(alnreads)[mask & !thissalnmask]
        } else {
            alnreads <- unlist(alnreads)[mask]
        }
    } else {
        if ((avsoas || is.na(empar@suboptimalAlignmentCutoff)) && !avgene) {
            alnreads <- alnreads[mask & !thissalnmask]
        } else {
            alnreads <- alnreads[mask]
        }
    }
    alnreads
}

#' @importFrom BiocGenerics basename path
.ervmapGetCounts <- function(avsoas, salnmask, empar, avgene, ov, alnreadidx,
                            alnAS, salnbestAS, readidx, mask, alnNH, readids,
                            bf, ovdiscard, n, verbose, iste) {
    ## if suboptimal alignment scores are available or they are not but
    ## we do not have secondary alignments either, then simply count filtered
    ## reads. In the latter case, we issue a warning.
    if (avsoas || !any(salnmask) || is.na(empar@suboptimalAlignmentCutoff)) {
    
        if (avgene) {
            cntvec <- .countbymatrix(empar, ov, alnreadidx, salnmask, alnAS,
                                salnbestAS, avgene, applysoasfilter = FALSE,
                                readidx, mask, alnNH, iste)
        } else {
            cntvec <- countSubjectHits(ov)
        }
        if (!is.na(empar@suboptimalAlignmentCutoff) && !avsoas) {
            msg <- paste("suboptimal alignment scores and secondary",
                         "alignments not available: skipping suboptimal",
                         "alignment filtering.")
            warning(msg)
        }
    
    } else {
        salnbestAS <- salnbestAS[match(readids[alnreadidx], names(salnbestAS))]
        salnbestAS[is.na(salnbestAS)] <- -99L
        rm(readids)
        gc()
        cntvec <- .countbymatrix(empar, ov, alnreadidx, salnmask, alnAS,
                                salnbestAS, avgene, applysoasfilter = TRUE,
                                readidx, mask, alnNH, iste)
        if (verbose > 1) {
            fstr <- paste("%s: %d alignments processed (%d are primary and",
                          "pass subptimal alignment filtering).\n")
            message(sprintf(fstr, basename(path(bf)), n, sum(cntvec)))
        }
    }
    names(cntvec) <- names(features(empar))
    
    if (isTRUE(avgene) && empar@geneCountMode == "all") {
        ## Adding counts of reads not passing ERVmap filters to genes 
        cntvec[!iste] <- cntvec[!iste] + countSubjectHits(ovdiscard)
    }
    
    if (length(empar@aggregateby) > 0) {
        if (avgene) {
            f <- .factoraggregateby(features(empar)[iste,], empar@aggregateby)
            f <- c(f, names(features(empar))[!iste])
        } else {
            f <- .factoraggregateby(features(empar), empar@aggregateby)
        }
        
        stopifnot(length(f) == length(cntvec)) ## QC
        cntvec <- tapply(cntvec, f, sum, na.rm=TRUE)
    }
    cntvec
}


.getreadsdiscard <- function(empar, alnreads, mask, thissalnmask) {
    
    if (empar@fragments) {
        alnreadsdiscard <- unlist(alnreads)[!mask & !thissalnmask] 
    } else {
        alnreadsdiscard <- alnreads[!mask & !thissalnmask]
    }
    alnreadsdiscard
}


.fetchNHtag <- function(thissalnmask, avgene, avtags, alnreads) {
    
    if (!any(thissalnmask) & avgene) {
        if (isTRUE("NH" %in% avtags)) {
            thisNH <- .getAlignmentTagScore(alnreads, tag="NH")
            alnNH <- c(alnNH,thisNH)
            return(alnNH)
        } else {
            warning("secondary alignments and NH tag not available in BAM file. Unique reads cannot be differentiated from multi-mapping reads. All reads aligned to overlapping regions between genes and TEs will be considered as gene counts.")
        }
    }
}


.pass2Filters <- function(alnreads, empar) {
    
    aqw <- .getAlignmentQueryWidth(alnreads)
    anm <- .getAlignmentMismatches(alnreads)
    asc <- .getAlignmentSumClipping(alnreads)
    ## Tokuyama et al. (2018) pg. 12571. reads/fragments are kept
    ## if: (1) the ratio of sum of hard and soft clipping to the
    ## sequence read length is < 0.02; (2) the ratio of the edit
    ## distance to the sequence read length is < 0.02
    mask <- ((asc / aqw) < empar@maxMismatchRate) & 
        ((anm / aqw) < empar@maxMismatchRate)
    
    mask
}


.pass3Filter <- function(mask, alnreads, empar, thisalnAS, thissalnmask, 
                            avsoas) {
    if (avsoas) {
        sas <- .getAlignmentTagScore(alnreads, tag=empar@suboptimalAlignmentTag)
        ## (3) the difference between the alignment score from BWA (AS)
        ## and the suboptimal alignment score from BWA (field XS) >= 5,
        mask <- mask & (thisalnAS - sas >= empar@suboptimalAlignmentCutoff)
    } else if (!any(thissalnmask)) {
        stop("The BAM file does not contain secondary alignments nor suboptimal alignment scores. Either set suboptimalAlignmentCutoff=NA or provide a BAM file with secondary alignments or suboptimal alignment scores.")
    }
    
    mask
}


.filterReadIDs <- function(avgene, empar, avsoas, alnreadids,
                            readids, readidx, alnreadidx, mask) {
    
    if (avgene || (!is.na(empar@suboptimalAlignmentCutoff) & !avsoas)) {
        
        ## store unique read identifiers and what alignment comes from what read
        ## First multiple alignments from the same read in the same chunk are 
        ## identified
        alnreadids_uniq <- unique(alnreadids)
        mt <- match(alnreadids, alnreadids_uniq) 
        l <- length(alnreadids_uniq)
        alnreadids_uniq <- c(alnreadids_uniq, alnreadids_uniq[is.na(mt)])
        mt[is.na(mt)] <- (l+1):(l+sum(is.na(mt)))
        chunkidx <- mt
        # Second, the index vectors are created
        mt <- match(alnreadids_uniq, readids) 
        l <- length(readids)
        # readids contains just once every readID of all reads in the BAM file
        readids <- c(readids, alnreadids_uniq[is.na(mt)]) 
        mt[is.na(mt)] <- (l+1):(l+sum(is.na(mt)))
        # index for all alignments in the BAM file
        readidx <- c(readidx, mt[chunkidx]) 
        # index for all alignments with mask == TRUE.
        alnreadidx <- c(alnreadidx, mt[chunkidx][mask]) 
    
    }
    
    list(readids = readids, readidx = readidx, alnreadidx = alnreadidx)
}

test_ERVmap <- function() {
    bamfiles <- list.files(system.file("extdata", package="atena"),
                           pattern="*.bam", full.names=TRUE)
    TE_annot <- readRDS(file=system.file("extdata", "Top28TEs.rds",
                                         package="atena"))
    empar <- ERVmapParam(bamfiles, teFeatures=TE_annot, singleEnd=TRUE,
                         ignoreStrand=TRUE, suboptimalAlignmentCutoff=NA)
    emSE <- qtex(empar)
    
    checkEqualsNumeric(dim(emSE), c(27, 2))
    checkEqualsNumeric(head(sort(assay(emSE), decreasing=TRUE)), 
                            c(3, 3, 3, 2, 2, 2))
}

test_em_input1 <- function() {
    
    TE_annot <- readRDS(file = system.file("extdata", "Top28TEs.rds",
                                           package="atena"))
    checkException(ERVmapParam(teFeatures=TE_annot, singleEnd=TRUE),
                   "An error prompts when no input BAM file is specified",
                   silent=TRUE)
}


test_em_input2 <- function() {
    
    bamfiles <- list.files(system.file("extdata", package="atena"),
                           pattern="*.bam", full.names=TRUE)
    checkException(ERVmapParam(bfl=bamfiles, singleEnd=TRUE),
                   "An error prompts when TE annotations are not specified",
                   silent=TRUE)
}



test_combine_seqlevels_annot <- function() {
    
    bamfiles <- list.files(system.file("extdata", package="atena"),
                           pattern="*.bam", full.names=TRUE)
    TE_annot <- GRanges(seqnames=rep("3R",8),
                        ranges=IRanges(start=c(30,42,84,120,134,159,175,250),
                                       width=c(35,10,5,15,10,25,5,20)),
                        strand="*")
    names(TE_annot) <- paste("TE", seq(1,8), sep="")
    gene_annot <- GRanges(seqnames=rep("2L",10),
                        ranges=IRanges(start=c(1,20,45,80,110,130,150,170,200,220),
                                       width=c(10,20,35,10,5,15,10,25,5,20)),
                        strand="*", 
                        type=rep("exon",10))
    names(gene_annot) <- paste("gene", seq(1,10), sep="")
    empar <- ERVmapParam(bamfiles, teFeatures=TE_annot, 
                         geneFeatures=gene_annot, singleEnd=TRUE,
                         ignoreStrand=TRUE, suboptimalAlignmentCutoff=NA)
    
    checkTrue(seqlevels(TE_annot) %in% seqlevels(features(empar)))
    checkTrue(seqlevels(gene_annot) %in% seqlevels(features(empar)))
}

#' QuantifyParam parameter class
#'
#' This is a virtual class from which other classes are derived
#' for storing parameters provided to quantification methods of
#' transposable elements from RNA-seq data.
#'
#' @slot bfl A [`BamFileList`][Rsamtools::BamFileList-class] object.
#'
#' @slot features A [`GRanges`][GenomicRanges::GRanges-class] object.
#'
#' @slot aggregateby Character vector with column names in the annotation
#'       to be used to aggregate quantifications.
#'       
#' @slot ovMode Character vector indicating the overlapping mode. Available
#'       options are: "ovUnion" (default) and "ovIntersectionStrict",
#'       which implement the corresponding methods from HTSeq
#'       (\url{https://htseq.readthedocs.io/en/release_0.11.1/count.html}).
#'       In the TEtranscripts, ERVmap and Telescope methods ambiguous
#'       alignments (alignments overlapping > 1 feature) are addressed
#'       differently depending on the method. In the atena method, those
#'       overlaps are not counted.
#'    
#'
#' @importClassesFrom Rsamtools BamFileList
#' @importClassesFrom GenomicRanges GenomicRanges_OR_GenomicRangesList
#'
#' @seealso
#' \code{\link{ERVmapParam-class}}
#' \code{\link{TelescopeParam-class}}
#' \code{\link{TEtranscriptsParam-class}}
#' \code{\link{atenaParam-class}}
#'
#' @examples
#' bamfiles <- list.files(system.file("extdata", package="atena"),
#'                        pattern="*.bam", full.names=TRUE)
#' \dontrun{
#' ## use the following two instructions to fetch annotations, they are here
#' ## commented out to enable running this example quickly when building and
#' ## checking the package
#' rmskat <- annotaTEs(genome="dm6", parsefun=rmskatenaparser,
#'                     strict=FALSE, insert=500)
#' rmskLTR <- getLTRs(rmskat, relLength=0.8,
#'                    fullLength=TRUE,
#'                    partial=TRUE)
#' }
#'
#' ## DO NOT TYPE THIS INSTRUCTION, WHICH JUST LOADS A PRE-COMPUTED ANNOTATION
#' ## YOU SHOULD USE THE INSTRUCTIONS ABOVE TO FETCH ANNOTATIONS
#' rmskLTR <- readRDS(system.file("extdata", "rmskatLTRrlen80flenpartoth.rds",
#'                                package="atena"))
#'
#' ## build a parameter object for TEtranscripts
#' ttpar <- TEtranscriptsParam(bamfiles,
#'                             teFeatures=rmskLTR,
#'                             singleEnd=TRUE,
#'                             ignoreStrand=TRUE)
#' ## just check that the parameter object belongs to the expected classes
#' is(ttpar, "QuantifyParam")
#' is(ttpar, "TEtranscriptsParam")
#'
#' @name QuantifyParam-class
#' @rdname QuantifyParam-class
#' @exportClass QuantifyParam
setClass("QuantifyParam",
         representation(bfl="BamFileList",
                        features="GenomicRanges_OR_GenomicRangesList",
                        aggregateby="character", ovMode="character"))

#' @param object A [`QuantifyParam`] object.
#'
#' @importFrom BiocGenerics path
#'
#' @return \code{path()}: Filesystem paths to the BAM files in the input
#' parameter object.
#'
#' @export
#' @aliases path,QuantifyParam-method
#' @rdname QuantifyParam-class
setMethod("path", "QuantifyParam",
        function(object) {
            path(object@bfl)
        })


#' @param x A [`QuantifyParam`] object.
#'
#' @importFrom GenomicFeatures features
#'
#' @return \code{features()}: The \code{GenomicRanges} or
#' \code{GenomicRangesList} object with the features in the input parameter
#' object.
#'
#' @export
#'
#' @aliases features,QuantifyParam-method
#' @rdname QuantifyParam-class
setMethod("features", "QuantifyParam",
        function(x) {
            x@features
        })

#' ERVmap parameter class
#'
#' This is a class for storing parameters provided to the ERVmap algorithm.
#' It is a subclass of the [`QuantifyParam`] class.
#'
#' @slot readMapper The name of the software used to align reads, obtained from
#' the BAM file header.
#'
#' @slot singleEnd (Default FALSE) Logical value indicating if reads are single
#' (\code{TRUE}) or paired-end (\code{FALSE}).
#'
#' @slot strandMode (Default 1) Numeric vector which can take values 0, 1 or 2.
#'   The strand mode is a per-object switch on
#'   \code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
#'   objects that controls the behavior of the strand getter. See
#'   \code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
#'   class for further detail. If \code{singleEnd = TRUE}, then
#'   \code{strandMode} #'   is ignored.
#'
#' @slot ignoreStrand (Default TRUE) A logical which defines if the strand
#' should be taken into consideration when computing the overlap between reads
#' and TEs in the annotations. When \code{ignore_strand = FALSE}, only those
#' reads which overlap the TE and are on the same strand are counted. On the
#' contrary, when \code{ignore_strand = TRUE}, any read overlapping an element
#' in \code{teFeatures} is counted regardless of the strand.
#'
#' @slot fragments (Default not \code{singleEnd}) A logical; applied to
#' paired-end data only. When \code{fragments=TRUE}, the read-counting
#' method in the original ERVmap algorithm is applied: each mate of a
#' paired-end read is counted (including ambiguous and not properly paired 
#' reads). When
#' \code{fragments=FALSE}, if the two mates of a paired-end read map to the
#' same element, they are counted as a single hit and singletons, reads with
#' unmapped pairs and other ambiguous or not properly paired fragments are 
#' not counted (see "Pairing criteria" in 
#' \code{\link[GenomicAlignments]{readGAlignments}()}).
#'
#' @slot maxMismatchRate (Default 0.02) Numeric value storing the maximum
#' mismatch rate employed by the ERVmap algorithm to discard aligned reads
#' whose rate of sum of hard and soft clipping, or of the edit distance over
#' the genome reference, to the length of the read is above this threshold.
#'
#' @slot suboptimalAlignmentTag (Default "auto") Character string storing the
#' tag name in the BAM files that stores the suboptimal alignment score used in
#' the third filter of ERVmap; see Tokuyama et al. (2018). The default,
#' \code{suboptimalAlignmentTag="auto"}, assumes that either the BAM files were
#' generated by BWA and include a tag called \code{XS} that stores the
#' suboptimal alignment score or, if the \code{XS} tag is not available, then
#' it uses the available secondary alignments to implement an analogous
#' approach to that of the third ERVmap filter. When
#' \code{suboptimalAlignmentTag="none"}, it also performs the latter approach
#' even when the tag \code{XS} is available.
#' When this parameter is different from \code{"auto"} and \code{"none"}, a tag
#' with the given name is used to extract the suboptimal alignment score.
#' The absence of that tag will prompt an error.
#'
#' @slot suboptimalAlignmentCutoff (Default 5) Numeric value storing the cutoff
#' above which the difference between the alignment score and the suboptimal
#' alignment score is considered sufficiently large to retain the alignment.
#' When this value is set to \code{NA}, then the filtering step based on
#' suboptimal alignment scores is skipped.
#'
#' @slot geneCountMode (Default "all") Character string indicating if the
#' ERVmap read filters applied to quantify TEs expression should also be
#' applied when quantifying gene expression ("ervmap") or not ("all"), in which
#' case all primary alignments mapping to genes are counted.
#'
#' @references
#' Tokuyama M et al. ERVmap analysis reveals genome-wide transcription of human
#' endogenous retroviruses. PNAS. 2018;115(50):12565-12572. DOI:
#' \url{https://doi.org/10.1073/pnas.1814589115}
#'
#' @name ERVmapParam-class
#' @rdname ERVmapParam-class
#' @exportClass ERVmapParam
setClass("ERVmapParam", contains="QuantifyParam",
        representation(singleEnd="logical",
                        ignoreStrand="logical",
                        strandMode="integer",
                        fragments="logical",
                        maxMismatchRate="numeric",
                        readMapper="character",
                        suboptimalAlignmentTag="character",
                        suboptimalAlignmentCutoff="numeric",
                        geneCountMode="character"))

#' Telescope parameter class
#'
#' This is a class for storing parameters
#' provided to the Telescope algorithm.
#'
#' @slot singleEnd (Default TRUE) Logical value indicating if reads are single
#' (\code{TRUE}) or paired-end (\code{FALSE}).
#'
#' @slot strandMode (Default 1) Numeric vector which can take values 0, 1 or 2.
#' The strand mode is a per-object switch on
#' \code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
#' objects that controls the behavior of the strand getter. See
#' \code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
#' class for further detail. If \code{singleEnd = TRUE}, then \code{strandMode}
#' is ignored.
#'
#' @slot ignoreStrand (Default FALSE) A logical which defines if the strand
#' should be taken into consideration when computing the overlap between reads
#' and annotated features. When \code{ignoreStrand = FALSE}, an aligned read
#' is considered to be overlapping an annotated feature as long as they
#' have a non-empty intersecting genomic range on the same strand, while when
#' \code{ignoreStrand = TRUE} the strand is not considered.
#'
#' @slot fragments (Default FALSE) A logical; applied to paired-end data only.
#' When \code{fragments=FALSE}, the read-counting method only counts
#' ‘mated pairs’ from opposite strands (non-ambiguous properly paired reads), 
#' while when \code{fragments=TRUE} same-strand pairs, singletons, reads with 
#' unmapped pairs and other ambiguous or not properly paired fragments
#' are also counted (see "Pairing criteria" in 
#' \code{\link[GenomicAlignments]{readGAlignments}()}). \code{fragments=TRUE} 
#' is equivalent to the original Telescope algorithm. For further details see
#' \code{\link[GenomicAlignments]{summarizeOverlaps}()}.
#' 
#' @slot minOverlFract (Default 0.2) A numeric scalar. \code{minOverlFract}
#' is multiplied by the read length and the resulting value is used to
#' discard alignments for which the overlapping length (number of base
#' pairs the alignment and the feature overlap) is lower. When no minimum 
#' overlap is required, set \code{minOverlFract = 0}.
#'
#' @slot pi_prior (Default 0) A positive integer scalar indicating the prior
#' on pi. This is equivalent to adding n unique reads.
#'
#' @slot theta_prior (Default 0) A positive integer scalar storing the prior
#' on Q. Equivalent to adding n non-unique reads.
#'
#' @slot em_epsilon (Default 1e-7) A numeric scalar indicating the EM
#' Algorithm Epsilon cutoff.
#'
#' @slot maxIter A positive integer scalar storing the maximum number of
#' iterations of the EM SQUAREM algorithm (Du and Varadhan, 2020). Default
#' is 100 and this value is passed to the \code{maxiter} parameter of the
#' \code{\link[SQUAREM]{squarem}()} function.
#' 
#' @slot reassign_mode (Default 'exclude') Character vector indicating
#' reassignment mode after EM step. 
#' Available methods are 'exclude' (reads with more than one best
#' assignment are excluded from the final counts), 'choose' (when reads have
#' more than one best assignment, one of them is randomly chosen), 'average'
#' (the read count is divided evenly among the best assignments) and 'conf'
#' (only assignments that exceed a certain threshold -defined by 
#' \code{conf_prob} parameter- are accepted, then the read count is
#' proportionally divided among the assignments above \code{conf_prob}).
#' 
#' @slot conf_prob (Default 0.9) Minimum probability for high confidence
#' assignment.
#' 
#' 
#' @references
#' Bendall et al. Telescope: characterization of the retrotranscriptome by
#' accurate estimation of transposable element expression.
#' PLOS Comp. Biol. 2019;15(9):e1006453. DOI:
#' \url{https://doi.org/10.1371/journal.pcbi.1006453}
#'
#' @name TelescopeParam-class
#' @rdname TelescopeParam-class
#' @exportClass TelescopeParam
setClass("TelescopeParam", contains="QuantifyParam",
        representation(singleEnd="logical",
                        strandMode="integer",
                        ignoreStrand="logical",
                        fragments="logical",
                        minOverlFract="numeric",
                        pi_prior="integer",
                        theta_prior="integer",
                        em_epsilon="numeric",
                        maxIter="integer",
                        reassign_mode="character",
                        conf_prob="numeric"))

#' TEtranscripts parameter class
#'
#' This is a class for storing parameters provided to the TEtranscripts
#' algorithm. It is a subclass of the [`QuantifyParam`].
#'
#' @slot singleEnd (Default FALSE) Logical value indicating if reads are single
#' (\code{TRUE}) or paired-end (\code{FALSE}).
#'
#' @slot ignoreStrand (Default FALSE) A logical which defines if the strand
#' should be taken into consideration when computing the overlap between reads
#' and annotated features. When \code{ignoreStrand = FALSE}, an aligned read
#' will be considered to be overlapping an annotated feature as long as they
#' have a non-empty intersecting genomic ranges on the same strand, while when
#' \code{ignoreStrand = TRUE} the strand will not be considered.
#'
#' @slot strandMode (Default 1) Numeric vector which can take values 0, 1 or 2.
#' The strand mode is a per-object switch on
#' \code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
#' objects that controls the behavior of the strand getter. See
#' \code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
#' class for further detail. If \code{singleEnd = TRUE}, then use either
#' \code{strandMode = NULL} or do not specify the \code{strandMode} parameter.
#'
#' @slot fragments (Default TRUE) A logical; applied to paired-end data only.
#' In both cases (\code{fragments=FALSE} and \code{fragments=TRUE}), the
#' read-counting method discards not properly paired reads. Moreover,
#' when \code{fragments=FALSE}, only non-ambiguous properly paired reads are
#' counted. When \code{fragments=TRUE}, ambiguous reads are also counted 
#' (see "Pairing criteria" in \code{\link[GenomicAlignments]{readGAlignments}()}). 
#' \code{fragments=TRUE} is equivalent to
#' the behavior of the TEtranscripts algorithm. For further details see
#' \code{\link[GenomicAlignments]{summarizeOverlaps}()}.
#'
#' @slot tolerance A positive numeric scalar storing the minimum tolerance
#' above which the SQUAREM algorithm (Du and Varadhan, 2020) keeps iterating.
#' Default is \code{1e-4} and this value is passed to the \code{tol} parameter
#' of the \code{\link[SQUAREM]{squarem}()} function.
#'
#' @slot maxIter A positive integer scalar storing the maximum number of
#' iterations of the SQUAREM algorithm (Du and Varadhan, 2020). Default
#' is 100 and this value is passed to the \code{maxiter} parameter of the
#' \code{\link[SQUAREM]{squarem}()} function.
#'
#' @references
#' Jin Y et al. TEtranscripts: a package for including transposable elements
#' in differential expression analysis of RNA-seq datasets.
#' Bioinformatics. 2015;31(22):3593-3599. DOI:
#' \url{https://doi.org/10.1093/bioinformatics/btv422}
#'
#' @name TEtranscriptsParam-class
#' @rdname TEtranscriptsParam-class
#' @exportClass TEtranscriptsParam
setClass("TEtranscriptsParam", contains="QuantifyParam",
        representation(singleEnd="logical",
                        ignoreStrand="logical",
                        strandMode="integer",
                        fragments="logical",
                        tolerance="numeric",
                        maxIter="integer"))

#' atena parameter class
#'
#' This is a class for storing parameters to quantify TE (and gene) expression
#' using the atena method. It is a subclass of the [`QuantifyParam`].
#'
#' @slot singleEnd (Default TRUE) Logical value indicating if reads are single
#' (\code{TRUE}) or paired-end (\code{FALSE}).
#'
#' @slot strandMode (Default 1) Numeric vector which can take values 0, 1 or 2.
#' The strand mode is a per-object switch on
#' \code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
#' objects that controls the behavior of the strand getter. See
#' \code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
#' class for further detail. If \code{singleEnd = TRUE}, then \code{strandMode}
#' is ignored.
#'
#' @slot ignoreStrand (Default FALSE) A logical which defines if the strand
#' should be taken into consideration when computing the overlap between reads
#' and annotated features. When \code{ignoreStrand = FALSE}, an aligned read
#' is considered to be overlapping an annotated feature as long as they
#' have a non-empty intersecting genomic range on the same strand, while when
#' \code{ignoreStrand = TRUE} the strand is not considered.
#'
#' @slot fragments (Default TRUE) A logical; applied to paired-end data only.
#' When \code{fragments=FALSE}, the read-counting method only counts
#' ‘mated pairs’ from opposite strands (non-ambiguous properly paired reads), 
#' while when \code{fragments=TRUE} same-strand pairs, singletons, reads with 
#' unmapped pairs and other ambiguous or not properly paired fragments
#' are also counted (see "Pairing criteria" in 
#' \code{\link[GenomicAlignments]{readGAlignments}()}). 
#' For further details see
#' \code{\link[GenomicAlignments]{summarizeOverlaps}()}.
#'
#' @slot pi_prior (Default 0) A positive numeric object indicating the prior
#' on pi. The same prior can be specified for all features setting
#' \code{pi_prior} as a scalar, or each feature can have a specific prior by
#' setting \code{pi_prior} as a vector with \code{names()} corresponding to
#' all feature names. Setting a pi prior is equivalent to adding n unique
#' reads.
#'
#' @slot theta_prior (Default 0) A positive numeric object indicating the
#' prior on Q. The same prior can be specified for all features setting
#' \code{theta_prior} as a scalar, or each feature can have a specific prior by
#' setting \code{theta_prior} as a vector with \code{names()} corresponding to
#' all feature names. Equivalent to adding n non-unique reads.
#'
#' @slot em_epsilon (Default 1e-7) A numeric scalar indicating the EM
#' Algorithm Epsilon cutoff.
#'
#' @slot maxIter A positive integer scalar storing the maximum number of
#' iterations of the EM SQUAREM algorithm (Du and Varadhan, 2020). Default
#' is 100 and this value is passed to the \code{maxiter} parameter of the
#' \code{\link[SQUAREM]{squarem}()} function.
#' 
#' @slot reassign_mode (Default 'exclude') Character vector indicating
#' reassignment mode after EM step. 
#' Available methods are 'exclude' (reads with more than one best
#' assignment are excluded from the final counts), 'choose' (when reads have
#' more than one best assignment, one of them is randomly chosen), 'average'
#' (the read count is divided evenly among the best assignments) and 'conf'
#' (only assignments that exceed a certain threshold -defined by 
#' \code{conf_prob} parameter- are accepted, then the read count is
#' proportionally divided among the assignments above \code{conf_prob}).
#' 
#' @slot conf_prob (Default 0.9) Minimum probability for high confidence
#' assignment.
#' 
#'
#' @name atenaParam-class
#' @rdname atenaParam-class
#' @exportClass atenaParam
setClass("atenaParam", contains="QuantifyParam",
         representation(singleEnd="logical",
                        strandMode="integer",
                        ignoreStrand="logical",
                        fragments="logical",
                        pi_prior="numeric",
                        theta_prior="numeric",
                        em_epsilon="numeric",
                        maxIter="integer",
                        reassign_mode="character",
                        conf_prob="numeric"))

#' Build a Telescope parameter object
#'
#' Build an object of the class \code{TelescopeParam}.
#'
#' @param bfl A \code{BamFile} or \code{BamFileList} object, or a character
#' string vector of BAM filenames.
#'
#' @param teFeatures A \code{GRanges} or \code{GRangesList} object. Elements
#' in this object should have names, which are used as a grouping factor for
#' genomic ranges forming a common locus (equivalent to "locus" column in
#' Telescope). This grouping is performed previous to TE expression
#' quantification, unlike the aggregation of quantifications performed when
#' the \code{aggregateby} parameter is specified, which is performed after
#' individual TE instances are quantified.
#'
#' @param aggregateby Character vector with column names from the annotation
#' to be used to aggregate quantifications. By default, this is an empty
#' vector, which means that the names of the input \code{GRanges} or
#' \code{GRangesList} object given in the \code{teFeatures} parameter are used
#' to aggregate quantifications.
#' 
#' @param ovMode Character vector indicating the overlapping mode. Available
#' options are: "ovUnion" (default) and "ovIntersectionStrict",
#' which implement the corresponding methods from HTSeq
#' (\url{https://htseq.readthedocs.io/en/release_0.11.1/count.html}).
#' Ambiguous alignments (alignments overlapping > 1 feature) are addressed
#' as in the original Telescope method: the overlap with the longest
#' overlapping length is kept.
#'
#' @param geneFeatures (Default NULL) A \code{GRanges} or 
#' \code{GRangesList} object with the
#' gene annotated features to be quantified. The TEtranscripts approach for
#' gene expression quantification is used, in which overlaps with multi-mapping
#' reads are preferentially assigned to TEs. Elements should have names
#' indicating the gene name/id. In case that \code{geneFeatures} 
#' is a \code{GRanges} and contains a
#' metadata column named \code{type}, only the elements with
#' \code{type} = \code{exon} are considered for the analysis. Then, exon counts
#' are summarized to the gene level. If NULL, gene expression is not
#' quantified.
#'
#' @param singleEnd (Default TRUE) Logical value indicating if reads are single
#' (\code{TRUE}) or paired-end (\code{FALSE}).
#'
#' @param strandMode (Default 1) Numeric vector which can take values 0, 1 or
#' 2.
#' The strand mode is a per-object switch on
#' \code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
#' objects that controls the behavior of the strand getter. See
#' \code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
#' class for further detail. If \code{singleEnd = TRUE}, then \code{strandMode}
#' is ignored.
#'
#' @param ignoreStrand (Default FALSE) A logical which defines if the strand
#' should be taken into consideration when computing the overlap between reads
#' and annotated features. When \code{ignoreStrand = FALSE}, an aligned read
#' is considered to be overlapping an annotated feature as long as they
#' have a non-empty intersecting genomic range on the same strand, while when
#' \code{ignoreStrand = TRUE} the strand is not considered.
#'
#' @param fragments (Default FALSE) A logical; applied to paired-end data only.
#' When \code{fragments=FALSE}, the read-counting method only counts
#' ‘mated pairs’ from opposite strands (non-ambiguous properly paired reads), 
#' while when \code{fragments=TRUE} same-strand pairs, singletons, reads with 
#' unmapped pairs and other ambiguous or not properly paired fragments
#' are also counted (see "Pairing criteria" in 
#' \code{\link[GenomicAlignments]{readGAlignments}()}). \code{fragments=TRUE} 
#' is equivalent to the original Telescope algorithm. For further details see
#' \code{\link[GenomicAlignments]{summarizeOverlaps}()}.
#'
#' @param minOverlFract (Default 0.2) A numeric scalar. \code{minOverlFract}
#' is multiplied by the read length and the resulting value is used to
#' discard alignments for which the overlapping length (number of base
#' pairs the alignment and the feature overlap) is lower. When no minimum 
#' overlap is required, set \code{minOverlFract = 0}.
#'
#' @param pi_prior (Default 0) A positive integer scalar indicating the prior
#' on pi. This is equivalent to adding n unique reads.
#'
#' @param theta_prior (Default 0) A positive integer scalar storing the prior
#' on Q. Equivalent to adding n non-unique reads.
#'
#' @param em_epsilon (Default 1e-7) A numeric scalar indicating the EM
#' Algorithm Epsilon cutoff.
#'
#' @param maxIter A positive integer scalar storing the maximum number of
#' iterations of the EM SQUAREM algorithm (Du and Varadhan, 2020). Default
#' is 100 and this value is passed to the \code{maxiter} parameter of the
#' \code{\link[SQUAREM]{squarem}()} function.
#' 
#' @param reassign_mode (Default 'exclude') Character vector indicating
#' reassignment mode after EM step. 
#' Available methods are 'exclude' (reads with more than one best
#' assignment are excluded from the final counts), 'choose' (when reads have
#' more than one best assignment, one of them is randomly chosen), 'average'
#' (the read count is divided evenly among the best assignments) and 'conf'
#' (only assignments that exceed a certain threshold -defined by 
#' \code{conf_prob} parameter- are accepted, then the read count is
#' proportionally divided among the assignments above \code{conf_prob}).
#' 
#' @param conf_prob (Default 0.9) Minimum probability for high confidence
#' assignment.
#' 
#' @param verbose (Default \code{TRUE}) Logical value indicating whether to
#' report progress.
#'
#' @details
#' This is the constructor function for objects of the class
#' \code{TelescopeParam-class}. This type of object is the input to the
#' function \code{\link{qtex}()} for quantifying expression of transposable
#' elements, which will call the Telescope algorithm
#' \href{https://doi.org/10.1371/journal.pcbi.1006453}{Bendall et al. (2019)}
#' with this type of object.
#'
#' @return A \linkS4class{TelescopeParam} object.
#'
#' @examples
#' bamfiles <- list.files(system.file("extdata", package="atena"),
#'                        pattern="*.bam", full.names=TRUE)
#' \dontrun{
#' ## use the following two instructions to fetch annotations, they are here
#' ## commented out to enable running this example quickly when building and
#' ## checking the package
#' rmskat <- annotaTEs(genome="dm6", parsefun=rmskatenaparser,
#'                     strict=FALSE, insert=500)
#' rmskLTR <- getLTRs(rmskat, relLength=0.8,
#'                    fullLength=TRUE,
#'                    partial=TRUE,
#'                    otherLTR=TRUE)
#' }
#'
#' ## DO NOT TYPE THIS INSTRUCTION, WHICH JUST LOADS A PRE-COMPUTED ANNOTATION
#' ## YOU SHOULD USE THE INSTRUCTIONS ABOVE TO FETCH ANNOTATIONS
#' rmskLTR <- readRDS(system.file("extdata", "rmskatLTRrlen80flenpartoth.rds",
#'                                package="atena"))
#'
#' ## build a parameter object for Telescope
#' tspar <- TelescopeParam(bfl=bamfiles,
#'                         teFeatures=rmskLTR,
#'                         singleEnd=TRUE,
#'                         ignoreStrand=TRUE)
#' tspar
#'
#' @references
#' Bendall et al. Telescope: characterization of the retrotranscriptome by
#' accurate estimation of transposable element expression.
#' PLOS Comp. Biol. 2019;15(9):e1006453. DOI:
#' \url{https://doi.org/10.1371/journal.pcbi.1006453}
#'
#' @importFrom methods is new
#' @importFrom Rsamtools BamFileList
#' @importFrom S4Vectors mcols
#' @importFrom cli cli_alert_info cli_alert_success
#' @export
#' @rdname TelescopeParam-class
TelescopeParam <- function(bfl, teFeatures, aggregateby=character(0),
                            ovMode="ovUnion",
                            geneFeatures=NULL,
                            singleEnd=TRUE,
                            strandMode=1L,
                            ignoreStrand=FALSE,
                            fragments=FALSE,
                            minOverlFract=0.2,
                            pi_prior=0L,
                            theta_prior=0L,
                            em_epsilon=1e-7,
                            maxIter=100L,
                            reassign_mode="exclude",
                            conf_prob=0.9,
                            verbose=TRUE) {

    if (verbose)
        cli_alert_info("Locating BAM files")
    bfl <- .checkBamFileListArgs(bfl, singleEnd, fragments)
    
    if (!reassign_mode %in% c("exclude","choose","average","conf"))
      stop("'reassign_mode' should be one of 'exclude', 'choose', 'average' or 'conf'")
    
    if (!ovMode %in% c("ovUnion","ovIntersectionStrict"))
      stop("'ovMode' should be one of 'ovUnion', 'ovIntersectionStrict'")
    
    if (verbose)
        cli_alert_info("Processing features")
    features <- .processFeatures(teFeatures, deparse(substitute(teFeatures)),
                                geneFeatures,deparse(substitute(geneFeatures)),
                                aggregateby, aggregateexons=TRUE)
    
    obj <- new("TelescopeParam", bfl=bfl, features=features,
                aggregateby=aggregateby, ovMode=ovMode,
                singleEnd=singleEnd,ignoreStrand=ignoreStrand,
                strandMode=as.integer(strandMode), fragments=fragments,
                minOverlFract=minOverlFract, pi_prior=pi_prior,
                theta_prior=theta_prior, em_epsilon=em_epsilon,
                maxIter=as.integer(maxIter), reassign_mode=reassign_mode,
                conf_prob=conf_prob)
    if (verbose)
        cli_alert_success("Parameter object successfully created")
    obj
}

#' @param object A \linkS4class{TelescopeParam} object.
#'
#' @importFrom Seqinfo seqlevels
#' @export
#' @aliases show,TelescopeParam-method
#' @rdname TelescopeParam-class
setMethod("show", "TelescopeParam",
        function(object) {
            cat(class(object), "object\n")
            cat(sprintf("# BAM files (%d): %s\n", length(object@bfl),
                        .pprintnames(names(object@bfl))))
            cat(sprintf("# features (%s length %d): %s\n",
                        class(object@features),
                        length(object@features),
                        ifelse(is.null(names(object@features)),
                                paste("on",
                                    .pprintnames(seqlevels(features(object)))),
                                .pprintnames(names(features(object))))))
            if (length(object@aggregateby) > 0)
                cat(sprintf("# aggregated by: %s\n",
                            paste(object@aggregateby, collapse=", ")))
            cat(sprintf("# %s; %s",
                        ifelse(object@singleEnd, "single-end", "paired-end"),
                        ifelse(object@ignoreStrand, "unstranded", "stranded")))
            if (!object@ignoreStrand)
                cat(sprintf(" (strandMode=%d)", object@strandMode))
            if (!object@singleEnd)
                cat(sprintf("; %s",
                        ifelse(object@fragments, "counting properly paired, same-strand pairs, singletons, reads with unmapped pairs and other fragments",
                                "counting properly paired reads")))
            cat("\n")
        })

#' @importFrom BiocParallel SerialParam bplapply
#' @importFrom S4Vectors DataFrame
#' @importFrom SummarizedExperiment SummarizedExperiment
#' @export
#' @aliases qtex
#' @aliases qtex,TelescopeParam-method
#' @rdname qtex
setMethod("qtex", "TelescopeParam",
          function(x, phenodata=NULL, mode=ovUnion, yieldSize=1e6L,
                   auxiliaryFeatures=FALSE,
                   BPPARAM=SerialParam(progressbar=TRUE)) {
            .checkPhenodata(phenodata, length(x@bfl))
            
            cnt <- bplapply(x@bfl, .qtex_telescope, tspar=x, mode=x@ovMode,
                            yieldSize=yieldSize, BPPARAM=BPPARAM)
            cnt <- do.call("cbind", cnt)
            cdata <- .createColumnData(cnt, phenodata)
            colnames(cnt) <- rownames(cdata)
            fnames <- rownames(cnt)[-nrow(cnt)]
            whnofeat <- 1L
            if (!auxiliaryFeatures) {
                cnt <- cnt[-nrow(cnt), ]
                whnofeat <- integer(0)
            }
            features <- .consolidateFeatures(x, fnames, whnofeat)
            SummarizedExperiment(assays=list(counts=cnt),
                                 rowRanges=features,
                                 colData=cdata)
        })


#' @importFrom stats setNames
#' @importFrom Rsamtools scanBamFlag ScanBamParam yieldSize yieldSize<-
#' @importFrom BiocGenerics path
#' @importFrom GenomicAlignments readGAlignments readGAlignmentsList
#' @importFrom GenomicAlignments readGAlignmentPairs
#' @importFrom S4Vectors mcols Hits queryHits subjectHits
#' @importFrom Matrix Matrix t which
#' @importFrom sparseMatrixStats rowSums2 colSums2
#' @importFrom SQUAREM squarem
.qtex_telescope <- function(bf, tspar, mode, yieldSize=1e6L) {
    mode <- match.fun(mode)
    readfun <- .getReadFunction(tspar@singleEnd, tspar@fragments)
    .checkreassignModes(tspar)
    sbflags <- .getScanBamFlag_ts(tspar@singleEnd, tspar@fragments)
    param <- ScanBamParam(flag=sbflags, what="flag", tag="AS")
    ## iste <- as.vector(attributes(tspar@features)$isTE[,1])
    iste <- mcols(features(tspar))$isTE
    if (any(duplicated(names(features(tspar)[iste]))))
        stop(".qtex_telescope: duplicated names in TE annotations.")

    ov <- Hits(nLnode=0, nRnode=length(tspar@features), sort.by.query=TRUE)
    alnreadids <- character(0)
    asvalues <- integer()
    alen <- numeric(0)
    mreadlen <- numeric(0)
    salnmask <- logical(0)
    strand_arg <- "strandMode" %in% formalArgs(readfun)
    yieldSize(bf) <- yieldSize
    open(bf)
    while (length(alnreads <- do.call(readfun,
                                c(list(file = bf), list(param=param),
                                list(strandMode=tspar@strandMode)[strand_arg],
                                list(use.names=TRUE))))) {
        alnreads <- .matchSeqinfo(alnreads, tspar@features)
        alnreadids <- c(alnreadids, names(alnreads))
        asvalues <- c(asvalues, .getAlignmentASScoreTS(alnreads, tag = "AS"))
        readlen <- .getAlignmentLength(alnreads)
        alen <- c(alen, readlen)
        salnmask <- c(salnmask, any(.secondaryAlignmentMask(alnreads)))
        thisov <- mode(alnreads, tspar@features,
                       ignoreStrand=tspar@ignoreStrand, inter.feature=FALSE)
        
        # Selecting the overlaps with min overlap higher than threshold
        ovlength <- .getOverlapLength(alnreads, thisov, tspar)
        yesminov <- ovlength > tspar@minOverlFract*readlen[queryHits(thisov)]
        thisov <- thisov[yesminov]
        ovlength <- ovlength[yesminov]
        
        # Selecting the best overlap for alignments overlapping > 1 feature
        multiov <- (duplicated(queryHits(thisov)) |
                        duplicated(queryHits(thisov), fromLast = TRUE))
        if (any(multiov)) {
          int <- ovlength[multiov]
          intmax <- aggregate(int, by = list(queryHits(thisov)[multiov]),
                              FUN = which.max)
          whpos <- which(!duplicated(queryHits(thisov)[multiov]))
          whpos <- whpos - 1
          intmax <- intmax[!duplicated(intmax$Group.1),]
          thisov <- thisov[-which(multiov)[-(whpos + intmax$x)]]
        }

        ov <- .appendHits(ov, thisov)
    }
    # close(bf)
    on.exit(close(bf))
    .checkOvandsaln(ov, salnmask)
    maskuniqaln <- .getMaskUniqueAln(alnreadids)
    ## fetch all different read identifiers from the overlapping alignments
    readids <- unique(alnreadids[queryHits(ov)])
    ## Adding "no_feature" overlaps to 'ov'
    ov <- .getNoFeatureOv(maskuniqaln, ov, alnreadids)
    mt <- match(readids, alnreadids)
    readids <- unique(alnreadids[queryHits(ov)]) # updating 'readids'
    cntvec <- .tsEMstep(tspar, alnreadids, readids, ov, asvalues,
                        iste, maskuniqaln, mt, alen)
    setNames(as.integer(cntvec), names(cntvec))
}

## private function .getNoFeatureOv(), obtains the overlaps to "no_feature"
#' @importFrom S4Vectors Hits queryHits subjectHits nRnode nLnode from to
.getNoFeatureOv <- function(maskuniqaln, ov, alnreadids) {
    if (!all(maskuniqaln)) {
        maskmultialn_ov <- !(maskuniqaln[queryHits(ov)])
        alnreadidx_all <- match(alnreadids, unique(alnreadids))
        ovid <- unique(alnreadidx_all[unique(queryHits(ov)[maskmultialn_ov])])
        alnall <- which(alnreadidx_all %in% ovid)
        nofeat <- setdiff(alnall, queryHits(ov)[maskmultialn_ov])
        from <- nofeat
        to <- rep(nRnode(ov) + 1, length(nofeat))
        nofeat_hits <- Hits(from = from, to = to, nLnode = nLnode(ov),
                            nRnode = max(to), sort.by.query = TRUE)
        
        # Adding overlaps to "no_feature" to 'ov'
        hits1 <- ov
        hits2 <- nofeat_hits
        
        hits <- c(Hits(from=from(hits1), to=to(hits1),
                        nLnode=nLnode(hits1),
                        nRnode=nRnode(hits2), sort.by.query=FALSE),
                    Hits(from=from(hits2), to=to(hits2),
                        nLnode=nLnode(hits2),
                        nRnode=nRnode(hits2), sort.by.query=FALSE))
        
        ovnofeat <- Hits(from = from(hits), to = to(hits), nLnode = nLnode(hits),
                        nRnode = nRnode(hits), sort.by.query=TRUE)
    } else {
      ovnofeat <- ov
    }
    return(ovnofeat)
}

#' @importFrom S4Vectors Hits queryHits subjectHits
#' @importFrom Matrix Matrix t which
#' @importFrom SQUAREM squarem
#' @importClassesFrom Matrix lgCMatrix
.tsEMstep <- function(tspar, alnreadids, readids, ov, asvalues,
                      iste, maskuniqaln, mt, alen) {
    ## initialize vector of counts derived from multi-mapping reads
    cntvec <- rep(0L, length(tspar@features) + 1)
    
    alnreadidx <- match(alnreadids, readids)
    rd_idx <- sort(unique(alnreadidx[queryHits(ov)]))
    
    ## fetch all different transcripts from the overlapping alignments
    tx_idx <- sort(unique(subjectHits(ov)))
    if (all(maskuniqaln)) {
        istex <- as.vector(iste[tx_idx])
    } else {
        istex <- as.vector(iste[tx_idx])[-length(tx_idx)] # removing no_feature
    }
    # asvalues <- (asvalues-min(asvalues)+1) / (max(asvalues)+1 - min(asvalues))
    asvalues <- .rescaleAS(asvalues, alen = alen)
    QmatTS <- .buildOvValuesMatrix(tspar, ov, asvalues, alnreadidx,
                                    rd_idx, tx_idx)
    
    if (!all(iste)) { ## Correcting for preference of reads to genes/TEs
        if (all(maskuniqaln)) {
          QmatTS <- .correctPreferenceTS(QmatTS, maskuniqaln, mt, istex)
          stopifnot(!any(rowSums2(QmatTS[,istex]) > 0 &
                           rowSums2(QmatTS[,!istex]) > 0))
        } else {
            QmatTS_nofeat <- QmatTS[,ncol(QmatTS), drop = FALSE]
            QmatTS <- .correctPreferenceTS(QmatTS[,-ncol(QmatTS)], maskuniqaln,
                                           mt, istex)
            stopifnot(!any(rowSums2(QmatTS[,istex]) > 0 &
                               rowSums2(QmatTS[,!istex]) > 0))
            QmatTS <- cbind(QmatTS, QmatTS_nofeat)
        }
    }
    
    
    # --- EM-step --- 
    # QmatTS <- QmatTS / rowSums2(QmatTS)
    QmatTS@x <- QmatTS@x / rowSums2(QmatTS)[QmatTS@i +1]
    
    # Getting 'y' indicator of unique and multi-mapping status from QmatTS 
    # maskmulti <- ifelse(rowSums2(QmatTS > 0) == 1, 0, 1)
    maskmulti <- rowSums2(QmatTS > 0) > 1
    
    # Telescope (Bendall et al.(2019)) defines the initial π estimate uniformly
    PiTS <- rep(1 / length(tx_idx), length(tx_idx))
    
    # Telescope defines an additional reassignment parameter (θ) as uniform
    Theta <- rep(1 / length(tx_idx), length(tx_idx))
    
    ## The SQUAREM algorithm to run the EM procedure
    a <- as.numeric(tspar@pi_prior) # 0
    b <- as.numeric(tspar@theta_prior) # 0
    Thetaenv <- new.env()
    assign("Theta", Theta, envir=Thetaenv)
    tsres <- squarem(par=PiTS, Thetaenv=Thetaenv, Q=QmatTS,maskmulti=maskmulti,
                    a=a, b=b, fixptfn=.tsFixedPointFun,
                    control=list(tol=tspar@em_epsilon, maxiter=tspar@maxIter))
    PiTS <- tsres$par
    PiTS[PiTS < 0] <- 0 ## Pi estimates are sometimes negatively close to zero
    # --- end EM-step ---
    
    Theta <- get("Theta", envir=Thetaenv)
    X <- .tsEstep(QmatTS, Theta, maskmulti, PiTS)
    cntvec <- .reassign(X, tspar@reassign_mode, tspar@conf_prob, cntvec, tx_idx)
    
    names(cntvec) <- c(names(tspar@features), "no_feature")
    nofeat <- cntvec["no_feature"]
    cntvec <- .tssummarizeCounts(cntvec[-length(cntvec)], iste, tspar)
    cntvec <- c(cntvec, nofeat)
    cntvec
}



## private function .checkreassignModes()
## Checks that 'reassign_mode' is valid
.checkreassignModes <- function(tspar) {
  if (!(tspar@reassign_mode %in% c("exclude","choose","average","conf")))
    stop("'reassign_mode' must be one of: 'exclude', 'choose', 'average' or 'conf'")
}


## private function .reassign()
## Implements 4 different reassigning methods from Telescope (exclude, choose,
## average and conf)
#' @importFrom sparseMatrixStats rowSums2 colSums2
#' @importFrom Matrix summary
#' @importFrom stats runif
.reassign <- function(X, reassign_mode, conf_prob, cntvec, tx_idx) {
  
  if (reassign_mode %in% c("exclude", "choose", "average")) {
    maxbyrow <- rowMaxs(X)
    # Older versions
    # Xind <- X == maxbyrow
    # Xind <- (X / maxbyrow) == 1
    Xind <- as(as(as(X, "lMatrix"), "generalMatrix"), "CsparseMatrix")
    Xind@x <- (X@x /maxbyrow[X@i+1]) == 1
    nmaxbyrow <- rowSums2(Xind)
    cntvec[tx_idx] <- colSums2(Xind[nmaxbyrow == 1, ])
  
    if (reassign_mode == "choose" & any(nmaxbyrow > 1)) {
      Xind2 <- Xind[nmaxbyrow > 1, ]
      Xind2_s <- summary(Xind2)
      Xind2_s <- Xind2_s[Xind2_s$x == TRUE,]
      Xind2_s <- Xind2_s[order(Xind2_s[,"i"]),]
      bestov <- as.vector(table(Xind2_s[,"i"]))
      # selected_ov <- vapply(bestov, FUN = function(x) sample(x = x, size = 1), 
      #                       FUN.VALUE = integer(1L))
      selected_ov <- ceiling(runif(n=nrow(Xind2))*rowSums2(Xind2))
      ovcol <- table(Xind2_s[c(0,cumsum(bestov)[-length(bestov)]) + selected_ov,"j"])
      whcol <- as.integer(names(ovcol))
      cntvec[tx_idx][whcol] <- cntvec[tx_idx][whcol] + as.integer(ovcol)
    }
    
    if (reassign_mode == "average" & any(nmaxbyrow > 1)) {
      Xind2 <- Xind[nmaxbyrow > 1, ]
      cntvec[tx_idx] <- cntvec[tx_idx] + colSums2(Xind2/rowSums2(Xind2))
    }
    
  } else if (reassign_mode == "conf") {
    whbelow <- X@x < conf_prob
    X@x[whbelow] <- 0
    X_conf <- X[rowSums2(X) > 0,]
    # X_conf@i indicates row and is 0-based
    # Faster sparse implementation of X_conf/rowSums2(X_conf)
    X_conf@x <- X_conf@x / rowSums2(X_conf)[X_conf@i +1] 
    cntvec[tx_idx] <- colSums2(X_conf)
    
  } else {
    stop("'reassign_mode' should be one of 'exclude', 'choose', 'average' or 'conf'")
  }
  
  cntvec
}


.rescaleAS <- function(asvalues, alen) {
  # asrescale <- rescale(asvalues, c(1, max(asvalues) - min(asvalues) + 1))
  asrescale <- asvalues - min(asvalues) + 1
  asrescale <- asrescale + alen
  asrescale <- asrescale/max(asrescale)
  asrescale <- expm1(asrescale*100)
  asrescale
}


## private function .correctPreferenceTS()
## Corrects QmatTS for preference of unique/multi-mapping reads to
## genes/TEs, respectively, in Telescope
.correctPreferenceTS <- function(QmatTS, maskuniqaln, mt, istex) {
    indx <- (rowSums2(QmatTS[,istex]) > 0) & (rowSums2(QmatTS[,!istex]) > 0)
    
    ## Assigning unique reads mapping to both a TE and a gene as gene counts
    # Which unique reads overlap to both genes and TEs?
    idxu <- indx & maskuniqaln[mt]
    if (any(idxu)) {
        # QmatTS[idxu,istex] <- FALSE
        whu <- which(QmatTS[idxu,istex] > 0, arr.ind = TRUE)
        if (is(whu, "integer")) {
            whu <- as.matrix(data.frame(row = 1, col = whu))
        }
        whudf <- cbind(which(idxu)[whu[,"row"]], which(istex)[whu[,"col"]])
        QmatTS[whudf] <- FALSE
    }
    
    ## Removing overlaps of multi-mapping reads to genes if at least one
    ## alignment of the read overlaps a TE
    idxm <- indx & !maskuniqaln[mt]
    if (any(idxm)) {
        # QmatTS[idxm,!istex] <- FALSE
        whm <- which(QmatTS[idxm,!istex] > 0, arr.ind = TRUE)
        if (is(whm, "integer")) {
          whm <- as.matrix(data.frame(row = 1, col = whm))
        }
        whmdf <- cbind(which(idxm)[whm[,"row"]], which(!istex)[whm[,"col"]])
        QmatTS[whmdf] <- FALSE
    }
    
    QmatTS
}

.tssummarizeCounts <- function(cntvec, iste, tspar) {
    cntvec_t <- cntvec[iste]
    ## aggregate TE quantifications if necessary
    if (length(tspar@aggregateby) > 0) {
        f <- .factoraggregateby(tspar@features[iste], tspar@aggregateby)
        stopifnot(length(f) == length(cntvec_t)) ## QC
        cntvec_t <- tapply(cntvec_t, f, sum, na.rm=TRUE)
    }
    
    ## aggregating exon counts to genes
    if (!all(iste)) {
        cntvec <- c(cntvec_t, cntvec[!iste])
    } else {
        cntvec <- cntvec_t
    }
    
    cntvec
}




## private function .tsEstep()
## E-step of the EM algorithm of Telescope
.tsEstep <- function(Q, Theta, maskmulti, Pi) {
    # X <- t(t(Q) * Pi)
    # # X[maskmulti, ] <- t(t(X[maskmulti, ]) * Theta)
    # # quicker computation of previous line
    # wh <- which(X*maskmulti > 0, arr.ind = TRUE)
    # X[wh] <- t(t(X[wh]) * Theta[wh[, "col"]])
    X <- Q
    j <- rep(seq_len(ncol(X)), diff(X@p))
    X@x <- X@x * Pi[j]
    #wh <- which(X@x * maskmulti[X@i + 1] > 0)
    whmulti <- as.integer(maskmulti[X@i + 1])
    wh <- which(X@x * whmulti > 0)
    j <- rep(seq_len(ncol(X)), diff(X@p))
    X@x[wh] <- X@x[wh] * Theta[j][wh]
    X <- X[rowSums2(X) > 0, , drop = FALSE]
    X <- X/rowSums2(X)
    X
}


## private function .tsMstepPi()
## M-step of the EM algorithm of Telescope
.tsMstepPi <- function(X, a) {
  Pi <- (colSums2(X) + a)/(sum(X@x) + a * ncol(X))
    Pi
}

## private function .tsMstepTheta()
## Update the estimate of the MAP value of θ
.tsMstepTheta <- function(X, maskmulti, b) {
    Theta <- ((colSums2(X[maskmulti, , drop=FALSE]) + b) / 
        (sum(maskmulti) + b*ncol(X)))
}

## private function .tsFixedPointFun()
## fixed point function of the EM algorithm of Telescope
.tsFixedPointFun <- function(Pi, Thetaenv, Q, maskmulti, a, b) {
    Theta <- get("Theta", envir=Thetaenv)
    X <- .tsEstep(Q, Theta, maskmulti, Pi)
    Pi2 <- .tsMstepPi(X, a)
    Theta2 <- .tsMstepTheta (X, maskmulti, b)
    assign("Theta", Theta2, envir=Thetaenv)
    
    Pi2
}


#' @importFrom S4Vectors mcols first second
.getAlignmentASScoreTS <- function(aln, tag) {
  if (is(aln, "GAlignments"))
    score <- as.integer(mcols(aln)[[tag]])
  else if (is(aln, "GAlignmentPairs")) {
    ## take the sum of the score of each mate
    score <- as.integer(mcols(first(aln))[[tag]]) + 
        as.integer(mcols(second(aln))[[tag]])
  } else if (is(aln, "GAlignmentsList")) {
    l <- lengths(aln)
    score <- aggregate(mcols(unlist(aln, use.names = FALSE))[[tag]],
                        by = list(rep(seq_along(l),l)), FUN = sum)
    score <- score$x
    mate_status <- mcols(aln)$mate_status == "mated"
    score[!mate_status] <- unlist(lapply(aln[!mate_status], 
                                  function(x) mean(mcols(x)[,tag])*2))
                                  # multiplied by 2 since there are 2 mates
  } else {
    stop(sprintf(".getAlignmentTagScore: wrong class %s\n", class(aln)))
  }
  as.integer(score)
}

#' @importFrom S4Vectors first second
#' @importFrom GenomicRanges granges
#' @importFrom GenomicAlignments seqnames
.getAlignmentLength <- function(alnreads) {
  if (is(alnreads, "GAlignments"))
    readlen <- qwidth(alnreads)
  else if (is(alnreads, "GAlignmentPairs")) {
      ## take the length of the region comprised by the 2 mates
      readlen <- width(granges(alnreads))
  } else if (is(alnreads, "GAlignmentsList")) {
      # readlen <- width(granges(alnreads, ignore.strand=TRUE))
      # # In case of reads with space between the two mates, the read length
      # # assigned corresponds to twice the maximum alignment length, to account
      # # for the length of the two mates, but not the region between them
      # maxlen <- max(qwidth(unlist(alnreads)))
      # readlen[readlen>maxlen] <- maxlen*2
      
      lsum <- sum(width(alnreads))
      # unmated alignments
      unmat <- mcols(alnreads)$mate_status == "unmated"
      # pairs with discordant chr
      unmat <- unmat | lengths(unique(seqnames(alnreads)))>1
      # Unmated alignments the assigned read length is the median read length
      lsum[unmat] <- median(width(unlist(alnreads)))
      ltogether <- integer(length = length(alnreads))
      ltogether[!unmat] <- width(granges(alnreads[!unmat], ignore.strand=TRUE))
      ltogether[unmat] <- median(width(unlist(alnreads)))
      ltogether[ltogether > lsum] <- lsum[ltogether > lsum]
      readlen <- ltogether
      
  } else {
    stop(sprintf(".getAlignmentLength: wrong class %s\n", class(alnreads)))
  }
  readlen
}


#' @importFrom S4Vectors Hits queryHits subjectHits
#' @importFrom GenomicRanges pintersect
#' @importFrom Seqinfo seqlevels<- seqlevels
.getOverlapLength <- function(alnreads, thisov, tspar) {
    features <- tspar@features
    seqlev <- unique(c(seqlevels(features), seqlevels(alnreads)))
    seqlevels(features) <- seqlev
    seqlevels(alnreads) <- seqlev
    
    if (is(alnreads, "GAlignmentsList")) {
        l <- lengths(alnreads)[queryHits(thisov)]
        features_ov <- rep(features[subjectHits(thisov)], l)
        ovlength <- width(pintersect(GRanges(unlist(alnreads[queryHits(thisov)])),
                                      features_ov,
                                      ignore.strand = tspar@ignoreStrand,
                                      strict.strand=FALSE))
        if (is(ovlength, "CompressedIntegerList")) {
          ovlength <- max(ovlength)
        }
        ovlength_ag <- aggregate(ovlength, by = list(rep(seq_along(l), l)), 
                                  FUN = max)
        ovlength <- ovlength_ag$x
    } else {
        ovlength <- width(pintersect(GRanges(alnreads[queryHits(thisov)]),
                                      features[subjectHits(thisov)],
                                      ignore.strand = tspar@ignoreStrand,
                                      strict.strand=FALSE))
        if (is(ovlength, "CompressedIntegerList")) {
          ovlength <- max(ovlength)
        }
    }
    ovlength
}

#' @importFrom Rsamtools scanBamFlag
.getScanBamFlag_ts <- function(singleEnd, fragments) {
  if (singleEnd == TRUE | fragments == TRUE) {
    sbflags <- scanBamFlag(isUnmappedQuery=FALSE,
                           isDuplicate=FALSE,
                           isNotPassingQualityControls=FALSE)
  } else {
    sbflags <- scanBamFlag(isUnmappedQuery=FALSE,
                           isDuplicate=FALSE,
                           isNotPassingQualityControls=FALSE,
                           isProperPair=TRUE)
  }
  sbflags
}

test_Telescope <- function() {
    bamfiles <- list.files(system.file("extdata", package="atena"),
                           pattern="*.bam", full.names=TRUE)
    TE_annot <- readRDS(file=system.file("extdata", "Top28TEs.rds",
                        package="atena"))
    gene_annot <- readRDS(file=system.file("extdata", "Top50genes.rds",
                          package="atena"))
    tspar <- TelescopeParam(bfl=bamfiles, teFeatures=TE_annot,
                            geneFeatures=gene_annot,
                            singleEnd=TRUE, ignoreStrand=TRUE)
    qts <- qtex(tspar, auxiliaryFeatures=TRUE)
    
    checkEqualsNumeric(dim(qts), c(76, 2))
    checkEqualsNumeric(head(sort(assay(qts), decreasing=TRUE)),
                       as.integer(c(149, 144, 6, 1, 0, 0)))
}


test_ts_input1 <- function() {
    
    TE_annot <- readRDS(file=system.file("extdata", "Top28TEs.rds",
                                         package="atena"))
    checkException(TelescopeParam(teFeatures=TE_annot, singleEnd=TRUE),
                   "An error prompts when no input BAM file is specified",
                   silent=TRUE)
}


test_ts_input2 <- function() {
    
    bamfiles <- list.files(system.file("extdata", package="atena"),
                           pattern="*.bam", full.names=TRUE)
    checkException(TelescopeParam(bfl=bamfiles, singleEnd=TRUE),
                   "An error prompts when TE annotations are not specified",
                   silent=TRUE)
}

#' atena: analysis of transposable elements in R and Bioconductor
#' 
#' The atena package provides a complete re-implementation in R of three
#' existing methods for the quantification of transposable element (TE)
#' expression in order to facilitate its integration into Bioconductor
#' workflows for the analysis of RNA-seq data. The three methods are
#' TEtranscripts
#' (\href{https://doi.org/10.1093/bioinformatics/btv422}{Jin et al. (2015)}),
#' ERVmap
#' (\href{https://doi.org/10.1073/pnas.1814589115}{Tokuyama et al. (2018)})
#' and Telescope
#' (\href{https://doi.org/10.1371/journal.pcbi.1006453}{Bendall et al.(2019)}).
#' 
#' 
#' The main functions are:
#'
#' \itemize{
#' \item \code{\link{TEtranscriptsParam}} - build parameter objects of the class \code{TEtranscriptsParam-class} for the TEtranscripts expression quantification method
#' \item \code{\link{ERVmapParam}} - build parameter objects of the class \code{ERVmapParam-class} for the ERVmap expression quantification method
#' \item \code{\link{TelescopeParam}} - build parameter objects of the class \code{TelescopeParam-class} for the Telescope expression quantification method
#' \item \code{\link{qtex}} - call the TE expression quantification method using a previously built parameter object
#' }
#' 
#' For detailed information on usage, see the package vignette, by typing
#' \code{vignette("atena")}.
#' 
#' All questions and bug reports should be posted to the Bioconductor Support 
#' Site:
#' 
#' \url{https://support.bioconductor.org}
#'
#' The code of the development version of the package is available at the
#' GitHub repository:
#'
#' \url{https://github.com/functionalgenomics/atena}
#' 
#' @references
#' 
#' Jin Y et al. TEtranscripts: a package for including transposable elements
#' in differential expression analysis of RNA-seq datasets.
#' Bioinformatics. 2015;31(22):3593-3599. DOI:
#' \url{https://doi.org/10.1093/bioinformatics/btv422}
#' 
#' Tokuyama M et al. ERVmap analysis reveals genome-wide transcription of
#' human endogenous retroviruses. PNAS. 2018;115(50):12565-12572. DOI:
#' \url{https://doi.org/10.1073/pnas.1814589115}
#' 
#' Bendall et al. Telescope: characterization of the retrotranscriptome by
#' accurate estimation of transposable element expression.
#' PLOS Comp. Biol. 2019;15(9):e1006453. DOI:
#' \url{https://doi.org/10.1371/journal.pcbi.1006453}
#' 
#' 
#' 
#' @name atena-package
#' @aliases atena-package
#' @aliases atena
#' @keywords package
"_PACKAGE"

NULL

---
title: "Creating data for the atena package"
author:
- name: Beatriz Calvo-Serra
  affiliation:
  - &id Dept. of Experimental and Health Sciences, Universitat Pompeu Fabra, Barcelona, Spain
  email: beatriz.calvo@upf.edu
- name: Robert Castelo
  affiliation: *id
  email: robert.castelo@upf.edu
vignette: >
  %\VignetteIndexEntry{Creating data for examples in the atena package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
output:
  BiocStyle::html_document:
    toc: true
    toc_float: true
    number_sections: true
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

## Downloading the data

The data used as example in the `atena` vignette and help pages is publicly
available at the National Center for Biotechnology Information (NCBI) 
Gene Expression Omnibus (accession no. GSE47006, 
https://www.ncbi.nlm.nih.gov/geo).
This dataset was analysed in
[Ohtani H et al., 2013](https://doi.org/10.1101/gad.221515.113). 
The two selected samples are: piwi knockdown and control samples (GSM1142845 
and GSM1142844). The raw data files (`.fastq`) were downloaded from Sequence
Read Archive (SRA, https://www.ncbi.nlm.nih.gov/sra) (samples SRX278536 and 
SRX278535, respectively)


## Alignment with BWA

The data was aligned with `BWA mem` using the `-a` option to output all found 
alignments for single-end, thus, including secondary alignments.

The resulting sorted BAM files (control_KD.sorted.bam and piwi_KD.sorted.bam) 
contain secondary alignments.

**Control KD**

```{bash}
## fetch FASTQ file
FASTQFILE=/projects_fg/TEs/OhtaniSaito13_DmGTSF1/control_KD.fastq.gz

## define where the BWA genome index is
wget http://ftp.ensembl.org/pub/release-104/fasta/drosophila_melanogaster/dna/Drosophila_melanogaster.BDGP6.32.dna.toplevel.fa.gz
GENOMEINDEX=/projects_fg/genomes/drosophila/Drosophila_melanogaster.BDGP6.32.dna.toplevel.fa.gz

## define number of threads to be used by BWA
NTHREADS=4

f=${FASTQFILE##*/}
PREFIX=${f%%.fastq.gz}

if [ ! -e BAM ] ; then mkdir -p BAM ; fi

module load BWA/0.7.17
module load SAMtools/1.7

echo $PREFIX
bwa mem -t $NTHREADS $GENOMEINDEX $FASTQFILE | \
  samtools view -b -o BAM/$PREFIX.unsorted.bam -

samtools sort BAM/$PREFIX.unsorted.bam \
  --threads 5 \
  -o BAM/$PREFIX.sorted.bam

samtools index BAM/$PREFIX.sorted.bam BAM/$PREFIX.sorted.bai
```



**Piwi KD**

```{bash}
## fetch FASTQ file
FASTQFILE=/projects_fg/TEs/OhtaniSaito13_DmGTSF1/piwi_KD.fastq.gz

f=${FASTQFILE##*/}
PREFIX=${f%%.fastq.gz}

if [ ! -e BAM ] ; then mkdir -p BAM ; fi

echo $PREFIX
bwa mem -t $NTHREADS $GENOMEINDEX $FASTQFILE | \
  samtools view -b -o BAM/$PREFIX.unsorted.bam -

samtools sort BAM/$PREFIX.unsorted.bam \
  --threads 5 \
  -o BAM/$PREFIX.sorted.bam

samtools index BAM/$PREFIX.sorted.bam BAM/$PREFIX.sorted.bai

```


## Getting TE and gene annotations

**Gene annotations**
```{bash}
# Gene annotations
wget http://ftp.ensembl.org/pub/release-103/gtf/drosophila_melanogaster/Drosophila_melanogaster.BDGP6.32.103.gtf.gz
```

**TE annotations**
```{bash}
# TE annotations: obtained from the RepeatMasker track in UCSC genome browser
# downloaded from https://genome.ucsc.edu/cgi-bin/hgTables (BDGP Release 6 +
# ISO1 MT/dm6; group: Variation and Repeats; track: RepeatMasker; table: rmsk)
## Obtaining RepeatMasker annotations in GTF and BED format
wget https://github.com/dpryan79/ChromosomeMappings/blob/master/BDGP6_UCSC2ensembl.txt?raw=true
mv /genomics/users/bea/HERVs/drosophila_atena/BDGP6_UCSC2ensembl.txt\?raw\=true \
/genomics/users/bea/HERVs/drosophila_atena/BDGP6_UCSC2ensembl.txt

```

```{r}
# Adding the Ensembl nomenclature to RepeatMasker annotations
ann <- read.table("
        /genomics/users/bea/HERVs/drosophila_atena/Dmelanogaster_rmsk_dm6.txt",
        header = TRUE)
chr <- read.table(
        "/genomics/users/bea/HERVs/drosophila_atena/BDGP6_UCSC2ensembl.txt")
colnames(chr) <- c("UCSC","Ensembl")
ann <- merge(ann, chr, by.y = "UCSC", by.x = "genoName")
# rRNA, simple repeat regions, low complexity regions, satellite repeats and
# artefacts will be filtered out from RepeatMasker annotations since they are
# not TEs
repClass <- c("RC", "LINE", "LTR", "DNA", "Unknown", "Other", "RNA")
ann <- ann[ann$repClass %in% repClass,]
# Creating a GRanges object with the annotations
suppressPackageStartupMessages(library(GenomicRanges))
ann_gr <- GRanges(seqnames = Rle(ann$Ensembl), 
                ranges = IRanges(start = ann$genoStart, end = ann$genoEnd),
                strand = Rle(ann$strand),
                mcols = DataFrame(ann[,11:13], 
                            name = paste(ann$repName, 1:nrow(ann), sep = "_")))
names(ann_gr) <- mcols(ann_gr)$mcols.name
colnames(mcols(ann_gr)) <- gsub(colnames(mcols(ann_gr)), 
                                pattern = "mcols.", replacement = "")
rtracklayer::export.gff2(ann_gr, con = "/genomics/users/bea/HERVs/drosophila_atena/Dmelanogaster_rmsk_dm6.gtf")

```



## Subsampling `.bam` files

The resulting `.bam` files were subsampled to smaller files containing just a 
few alignments.

First `.bam` files were sorted by name in order to group alignments coming
from the same read (the primary alignment + secondary alignments in case of
multimapping reads).

```{bash}
samtools sort -n -@ 4 /projects_fg/TEs/OhtaniSaito13_DmGTSF1/BAM/piwi_KD.sorted.bam -o piwi_KD.sortedbyname.bam -O bam
samtools sort -n -@ 4 /projects_fg/TEs/OhtaniSaito13_DmGTSF1/BAM/control_KD.sorted.bam -o control_KD.sortedbyname.bam -O bam

```

Finding which TEs are the 20 most expressed. Also finding which reads from
piwi map to tese 20 most expressed TEs:

```{r}
library(Matrix)
library(Rsamtools)
library(GenomicAlignments)
library(rtracklayer)

#' @importFrom S4Vectors nLnode nRnode isSorted from to Hits
.appendHits <- function(hits1, hits2) {
    stopifnot(nRnode(hits1) == nRnode(hits2))
    stopifnot(isSorted(from(hits1)) == isSorted(from(hits2)))
    hits <- c(Hits(from=from(hits1), to=to(hits1),
                    nLnode=nLnode(hits1)+nLnode(hits2),
                    nRnode=nRnode(hits1), sort.by.query=isSorted(from(hits1))),
            Hits(from=from(hits2)+nLnode(hits1), to=to(hits2),
                    nLnode=nLnode(hits1)+nLnode(hits2),
                    nRnode=nRnode(hits2), sort.by.query=isSorted(from(hits2))))
    hits
}


te_dros <- import(
    "/genomics/users/bea/HERVs/drosophila_atena/Dmelanogaster_rmsk_dm6.gtf")
names(te_dros) <- mcols(te_dros)$ID

bampath <- "/projects_fg/TEs/OhtaniSaito13_DmGTSF1/BAM"
bamfiles <- c("/projects_fg/TEs/OhtaniSaito13_DmGTSF1/BAM/piwi_KD.sorted.bam",
            "/projects_fg/TEs/OhtaniSaito13_DmGTSF1/BAM/control_KD.sorted.bam")

bfl <- BamFileList(bamfiles, yieldSize = 10000000, asMates = FALSE)
sbflags <- scanBamFlag(isUnmappedQuery=FALSE,
                        isDuplicate=FALSE,
                        isNotPassingQualityControls=FALSE)
param <- ScanBamParam(flag=sbflags)

i <- 2
bf <- bfl[[i]]
ov <- Hits(nLnode=0, nRnode=length(te_dros), sort.by.query=TRUE)
alnreadids <- character(0)
l <- 0
open(bf)
while (length(alnreads <- do.call("readGAlignments", c(list(file = bf), 
                                            list(param=param), 
                                            list(use.names=TRUE))))) {
    alnreadids <- c(alnreadids, names(alnreads))
    l <- l+1
    print(l)
    thisov <- findOverlaps(alnreads, te_dros, ignore.strand=FALSE)
    ov <- .appendHits(ov, thisov)
}
close(bf)

.buildOvAlignmentsMatrix <- function(ov, arids, rids, fidx) {
    oamat <- Matrix(FALSE, nrow=length(rids), ncol=length(fidx))
    mt1 <- match(arids[queryHits(ov)], rids)
    mt2 <- match(subjectHits(ov), fidx)
    oamat[cbind(mt1, mt2)] <- TRUE
  
    oamat
}

saveRDS(ov, file = "/genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/piwi_KD_ovAllTEs.rds")
readids <- unique(alnreadids[queryHits(ov)])
tx_idx <- sort(unique(subjectHits(ov)))
ovalnmat <- .buildOvAlignmentsMatrix(ov, alnreadids, readids, tx_idx)
topreads <- readids[order(rowSums(ovalnmat),decreasing = TRUE)[1:175]]
topreads2 <- paste(topreads,"\t",sep="")
write.table(topreads2, 
            file = "/genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/OhtaniSaito13_topreads_piwi.txt",
            quote = FALSE, row.names = FALSE, col.names = FALSE)

ovalnmat_topreads <- ovalnmat[order(rowSums(ovalnmat), decreasing = TRUE)[1:175],]

te_dros_top20 <- te_dros[tx_idx[order(colSums(ovalnmat_topreads), decreasing = TRUE)[1:20]]]
saveRDS(te_dros_top20, 
        file = "/genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/OhtaniSaito13_GRangesTop20TEs_piwi.rds")
```


Finding which genes are the 50 most expressed:
```{r}
suppressPackageStartupMessages(library(GenomicAlignments))
suppressPackageStartupMessages(library(rtracklayer))

# BAM files
bamfiles <- c("/projects_fg/TEs/OhtaniSaito13_DmGTSF1/BAM/piwi_KD.sorted.bam",
            "/projects_fg/TEs/OhtaniSaito13_DmGTSF1/BAM/control_KD.sorted.bam")
bfl <- BamFileList(bamfiles, yieldSize = 100000, asMates = FALSE)
sbflags <- scanBamFlag(isUnmappedQuery=FALSE,
                        isDuplicate=FALSE,
                        isNotPassingQualityControls=FALSE)
param <- ScanBamParam(flag=sbflags)

# Annotations
gene_dros <- import(
    "/projects_fg/genomes/drosophila/Drosophila_melanogaster.BDGP6.32.103.gtf")
gene_dros <- gene_dros[!(gene_dros$gene_biotype == "transposable_element")]
gene_dros <- gene_dros[gene_dros$type == "gene"]
names(gene_dros) <- gene_dros$gene_id

# Counting
overlap_all <- summarizeOverlaps(gene_dros, bfl, singleEnd=TRUE, 
                                param = param, ignore.strand = FALSE)
saveRDS(overlap_all, file = "/genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/OhtaniSaito13gene_counts.rds")

top50_gene <- rownames(assay(overlap_all)[order(rowSums(assay(overlap_all)), 
                                                decreasing = TRUE),])[1:50]
write.table(top50_gene, 
            file = "/genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/OhtaniSaito13gene_counts_top50.txt",
            quote = FALSE, row.names = FALSE, col.names = FALSE)

gene_dros_top50 <- gene_dros[top50_gene]
saveRDS(gene_dros_top50, 
        file = "/genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/OhtaniSaito13_GRangesTop50genes.rds")
```

Finding which reads from control map to the 20 most expressed TEs:
```{r}
suppressPackageStartupMessages(library(GenomicAlignments))
suppressPackageStartupMessages(library(rtracklayer))

# BAM files
bamfiles <- c("/projects_fg/TEs/OhtaniSaito13_DmGTSF1/BAM/piwi_KD.sorted.bam",
              "/projects_fg/TEs/OhtaniSaito13_DmGTSF1/BAM/control_KD.sorted.bam")
bfl <- BamFileList(bamfiles, yieldSize = 1000000, asMates = FALSE)
sbflags <- scanBamFlag(isUnmappedQuery=FALSE,
                       isDuplicate=FALSE,
                       isNotPassingQualityControls=FALSE)
param <- ScanBamParam(flag=sbflags)

## --- Looking now for overlapping reads ---
library(Matrix)

.appendHits <- function(hits1, hits2) {
  stopifnot(nRnode(hits1) == nRnode(hits2))
  stopifnot(isSorted(from(hits1)) == isSorted(from(hits2)))
  hits <- c(Hits(from=from(hits1), to=to(hits1),
                 nLnode=nLnode(hits1)+nLnode(hits2),
                 nRnode=nRnode(hits1), sort.by.query=isSorted(from(hits1))),
            Hits(from=from(hits2)+nLnode(hits1), to=to(hits2),
                 nLnode=nLnode(hits1)+nLnode(hits2),
                 nRnode=nRnode(hits2), sort.by.query=isSorted(from(hits2))))
  hits
}

te_dros_top20 <- readRDS("/genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/OhtaniSaito13_GRangesTop20TEs_piwi.rds")
i <- 2
bf <- bfl[[i]]
ov <- Hits(nLnode=0, nRnode=length(te_dros_top20), sort.by.query=TRUE)
alnreadids <- character(0)
l <- 1
open(bf)
while (length(alnreads <- readGAlignments(bf, param = param, use.names = TRUE))) {
  alnreadids <- c(alnreadids, names(alnreads))
  thisov <- findOverlaps(alnreads, te_dros_top20, ignore.strand=FALSE)
  ov <- .appendHits(ov, thisov)
  l <- l+1
  print(l)
}
close(bf)

readhits <- table(alnreadids[queryHits(ov)])
saveRDS(ov, 
        file = "/genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/OhtaniSaito13_ovTopTEs_control.rds")

topreads<- names(readhits)[order(readhits, decreasing = TRUE)][1:175]
topreads2 <- paste(topreads,"\t",sep="")

write.table(topreads2, 
            file = "/genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/OhtaniSaito13_topreads_control.txt",
            quote = FALSE, row.names = FALSE, col.names = FALSE)

```



Subsampling BAM files by selecting reads based on reads ids which have been 
previously identified to map to the most expressed TEs.

```{bash}
# 175 reads correspond to a high number of alignments, which causes bam files 
# to be too large
# Reducing to 150 reads:
head -n 150 OhtaniSaito13_topreads_piwi.txt > OhtaniSaito13_topreads_piwi2.txt
mv OhtaniSaito13_topreads_piwi2.txt OhtaniSaito13_topreads_piwi.txt

head -n 150 OhtaniSaito13_topreads_control.txt > OhtaniSaito13_topreads_control2.txt
mv OhtaniSaito13_topreads_control2.txt OhtaniSaito13_topreads_control.txt


module load SAMtools

# piwi
samtools view /genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/piwi_KD.sortedbyname.bam | grep -f /genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/OhtaniSaito13_topreads_piwi.txt > /genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/piwi_KD_subsampled_wtheader.sort.sam

samtools view /genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/piwi_KD.sortedbyname.bam -H > header_piwi_KD_subsampled.txt
cat header_piwi_KD_subsampled.txt /genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/piwi_KD_subsampled_wtheader.sort.sam > /genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/piwi_KD_subsampled.sort.sam

samtools sort /genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/piwi_KD_subsampled.sort.sam -o /genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/piwi_KD_subsampled.sort.bam
samtools index /genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/piwi_KD_subsampled.sort.bam /genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/piwi_KD_subsampled.sort.bai

# control
samtools view /genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/control_KD.sortedbyname.bam | grep -f /genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/OhtaniSaito13_topreads_control.txt > /genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/control_KD_subsampled_wtheader.sort.sam

samtools view /genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/control_KD.sortedbyname.bam -H > header_control_KD_subsampled.txt
cat header_control_KD_subsampled.txt /genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/control_KD_subsampled_wtheader.sort.sam > /genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/control_KD_subsampled.sort.sam

samtools sort /genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/control_KD_subsampled.sort.sam -o /genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/control_KD_subsampled.sort.bam
samtools index /genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/control_KD_subsampled.sort.bam /genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/control_KD_subsampled.sort.bai


```


Final files included in `atena` package:
```
/genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/piwi_KD_subsampled.sort.bam
/genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/piwi_KD_subsampled.sort.bai
/genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/control_KD_subsampled.sort.bam
/genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/control_KD_subsampled.sort.bai
/genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/OhtaniSaito13_GRangesTop20TEs_piwi.rds
/genomics/users/bea/HERVs/drosophila_OhtaniSaito13_DmGTSF1/OhtaniSaito13_GRangesTop50genes.rds
```
Which have been renamed to:
```
piwi_KD.bam
piwi_KD.bai
control_KD.bam
control_KD.bai
Top20TEs.rds
Top50genes.rds
```

**Adding 10 top expressed TEs according to the ERVmap method**
ERVmap did not provide any overlap for `Top20TEs.rds`. This is due to the fact
that the alignments which caused these 20 TEs to be the most expressed are
mainly secondary alignments of multi-mapping reads. Since ERVmap does not
count secondary alignments, 0 overlaps are found for these 20 TEs in ERVmap.
For this reason, the following approach was carried out in order to include 10
more TEs in the annotations. These 10 TEs were selected by looking where the
primary alignments of the subsampled BAM files mapped to. 

```{r}
tefile <- "Dmelanogaster_rmsk_dm6.gtf"
te_dros <- import(tefile)

alnreads <- do.call(readfun, c(list(file = bf),
                                list(param=param), 
                                list(strandMode=empar@strandMode)[strand_arg],
                                list(use.names=(!avsoas || avgene))))
ovall <- findOverlaps(te_dros, alnreads)
tes <- countQueryHits(ovall)
te_dros[order(tes, decreasing = TRUE)][1:10]
teall <- c(TE_annot, te_dros[order(tes, decreasing = TRUE)][1:10])
colnames(mcols(teall))[2] <- "feature_type"
saveRDS(teall, 
    file = "bioc-devel/ERVs/OhtaniSaito13_subsampled_for_atena/Top30TEs.rds")

# Finally, 2 duplicated TEs were removed creating the final object Top28TEs.rds

```


**Adding `seqinfo` data**

```{r}
TE_annot <- readRDS(file = "Top28TEs.rds")
gene_annot <- readRDS(file =  "Top50genes.rds")

suppressPackageStartupMessages(library(BSgenome.Dmelanogaster.UCSC.dm6))
suppressPackageStartupMessages(library(GenomeInfoDb))
seqlevelsStyle(gene_annot) <- "UCSC"
seqlevelsStyle(TE_annot) <- "UCSC"

seqlevels(gene_annot, 
          pruning.mode="coarse") <- seqlevels(BSgenome.Dmelanogaster.UCSC.dm6)

seqlevels(TE_annot, 
          pruning.mode="coarse") <- seqlevels(BSgenome.Dmelanogaster.UCSC.dm6)

seqlengths(gene_annot) <- seqlengths(BSgenome.Dmelanogaster.UCSC.dm6)
seqlengths(TE_annot) <- seqlengths(BSgenome.Dmelanogaster.UCSC.dm6)

saveRDS(TE_annot, file = "Top28TEs.rds")
saveRDS(gene_annot, file = "Top50genes.rds")
```

---
title: "An introduction to the atena package"
author:
- name: Beatriz Calvo-Serra
  affiliation:
  - &id Dept. of Experimental and Health Sciences, Universitat Pompeu Fabra, Barcelona, Spain
  email: beatriz.calvo@upf.edu
- name: Robert Castelo
  affiliation: *id
  email: robert.castelo@upf.edu
package: "`r pkg_ver('atena')`"
abstract: >
  The `atena` package provides methods to quantify the expression of transposable elements within R and Bioconductor.
vignette: >
  %\VignetteIndexEntry{An introduction to the atena package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
output:
  BiocStyle::html_document:
    toc: true
    toc_float: true
    number_sections: true
bibliography: bibliography.bib
---

```{r setup, echo=FALSE}
library(knitr)

options(width=80)

knitr::opts_chunk$set(
  collapse=TRUE,
  comment="")
```

# What are transposable elements

Transposable elements (TEs) are autonomous mobile genetic elements. They are
DNA sequences that have, or once had, the ability to mobilize within the genome
either directly or through an RNA intermediate [@payer2019transposable]. TEs
can be categorized into two classes based on the intermediate substrate
propagating insertions (RNA or DNA). Class I TEs, also called retrotransposons,
first transcribe an RNA copy that is then reverse transcribed to cDNA before
inserting in the genome. In turn, these can be divided into long terminal repeat
(LTR) retrotransposons, which refer to endogenous retroviruses (ERVs), and
non-LTR retrotransposons, which include long interspersed element class 1
(LINE-1 or L1) and short interspersed elements (SINEs). Class II TEs, also known
as DNA transposons, directly excise themselves from one location before
reinsertion. TEs are further split into families and subfamilies depending on
various structural features [@goerner2018computational; @guffanti2018novel].

Most TEs have lost the capacity for generating new insertions over their
evolutionary history and are now fixed in the human population. Their insertions
have resulted in a complex distribution of interspersed repeats comprising
almost half (50%) of the human genome [@payer2019transposable].

TE expression has been observed in association with physiological processes in
a wide range of species, including humans where it has been described to be
important in early embryonic pluripotency and development. Moreover, aberrant TE
expression has been associated with diseases such as cancer, neurodegenerative
disorders, and infertility [@payer2019transposable].

# Currently available methods for quantifying TE expression

The study of TE expression faces one main challenge: given their repetitive
nature, the majority of TE-derived reads map to multiple regions of the genome
and these multi-mapping reads are consequently discarded in standard RNA-seq
data processing pipelines. For this reason, specific software packages for the
quantification of TE expression have been developed [@goerner2018computational],
such as TEtranscripts [@jin2015tetranscripts], ERVmap [@tokuyama2018ervmap] and
Telescope [@bendall2019telescope]. The main differences between these three
methods are the following: 

* [TEtranscripts](https://github.com/mhammell-laboratory/TEtranscripts)
  [@jin2015tetranscripts] reassigns multi-mapping reads to TEs proportionally
  to their relative abundance, which is estimated using an
  expectation-maximization (EM) algorithm.

* [ERVmap](https://github.com/mtokuyama/ERVmap) [@tokuyama2018ervmap] is based
  on selective filtering of multi-mapping reads. It applies filters that consist
  in discarding reads when the ratio of sum of hard and soft clipping to the
  length of the read (base pair) is greater than or equal to 0.02, the ratio of
  the edit distance to the sequence read length (base pair) is greater or equal
  to 0.02 and/or the difference between the alignment score from BWA (field AS)
  and the suboptimal alignment score from BWA (field XS) is less than 5.

* [Telescope](https://github.com/mlbendall/telescope) [@bendall2019telescope]
  reassigns multi-mapping reads to TEs using their relative abundance, which
  like in TEtranscripts, is also estimated using an EM algorithm. The main
  differences with respect to TEtranscripts are: (1) Telescope works with an
  additional parameter for each TE that estimates the proportion of
  multi-mapping reads that need to be reassigned to that TE; (2) that
  reassignment parameter is optimized during the EM algorithm jointly with the
  TE relative abundances, using a Bayesian maximum a posteriori (MAP) estimate
  that allows one to use prior values on these two parameters; and (3) using
  the final estimates on these two parameters, multi-mapping reads can be
  flexibly reassigned to TEs using different strategies, where the default one
  is to assign a multi-mapping read to the TE with largest estimated abundance
  and discard those multi-mapping reads with ties on those largest abundances.

Because these tools were only available outside R and Bioconductor, the `atena`
package provides a complete re-implementation in R of these three methods to
facilitate the integration of TE expression quantification into Bioconductor
workflows for the analysis of RNA-seq data.

# TEs annotations

Another challenge in TE expression quantification is the lack of complete TE
annotations due to the difficulty to correctly place TEs in genome assemblies
[@goerner2018computational]. One of the main sources of TE annotations are
RepeatMasker annotations, available for instance at the RepeatMasker track of
the UCSC Genome Browser. `atena` can fetch RepeatMasker annotations with the
function `annotaTEs()` and flexibly parse them by using a parsing function
provided through the parameter `parsefun`. Examples of `parsefun` included in
`atena` are:

* `rmskidentity()`: returns RepeatMasker annotations without any modification.
* `rmskbasicparser()`: filters out non-TE repeats and elements without strand
  information from RepeatMasker annotations. Then assigns a unique id to each 
  elements based on their repeat name.
* `OneCodeToFindThemAll()`: implementation of the "One Code To Find Them All"
  algorithm by @bailly2014one, for parsing RepeatMasker output files.
* `rmskatenaparser()`: attempts to reconstruct fragmented TEs by assembling 
  together fragments from the same TE that are close enough. For LTR class TEs,
  tries to reconstruct full-length and partial TEs following the LTR - internal
  region - LTR structure.

Both, the `rmskatenaparser()` and `OneCodeToFindThemAll()` parser functions
attempt to address the annotation fragmentation present in the output files of
the RepeatMasker software (i.e. presence of multiple hits, such as
homology-based matches, corresponding to a unique copy of an element). This is
highly frequent for TEs of the LTR class, where the consensus sequences are
split separately into the LTR and internal regions, causing RepeatMasker to
also report these two regions of the TE as two separate elements. These two
functions try to identify these and other cases of fragmented annotations and
assemble them together into single elements. To do so, the assembled elements
must satisfy certain criteria. These two parser functions differ in those
criteria, as well as in the approach for finding equivalences between LTR and
internal regions to reconstruct LTR retrotransposons. The `rmskatenaparser()`
function is also much faster than `OneCodeToFindThemAll()`.

## Retrieving and parsing TE annotations

As an example, let's retrieve TE annotations for *Drosophila melanogaster* 
*dm6* genome version. By setting `rmskidentity()` as argument to the
`parsefun` parameter, RepeatMasker annotations are retrieved intact as a
`GRanges` object.

```{r, message=FALSE, warning=FALSE}
library(atena)
library(BiocParallel)

rmskann <- annotaTEs(genome="dm6", parsefun=rmskidentity)
rmskann
```

We can see that we obtained annotations for `r length(rmskann)` elements. Now,
let's fetch the same RepeatMasker annotations, but process them using the
`OneCodeToFindThemAll` parser function [@bailly2014one]. We set the parameter
`strict=FALSE` to avoid applying a filter of minimum 80% identity with the
consensus sequence and minimum 80 bp length. The `insert` parameter is set to
500, meaning that two elements with the same name are merged if they are closer
than 500 bp in the annotations. The `BPPARAM` parameter allows one to run
calculations in parallel using the functionality of the
[BiocParallel](https://bioconductor.org/packages/BiocParallel) Bioconductor
package. In this particular example, we are setting the `BPPARAM` parameter
to `SerialParam(progress=FALSE)` to disable parallel calculations and progress
reporting, but a common setting if we want to run calculations in parallel
would be `BPPARAM=Multicore(workers=ncores, progress=TRUE)`, which would use
`ncores` parallel threads of execution and report progress on the calculations.

```{r, message=FALSE, warning=FALSE}
teann <- annotaTEs(genome="dm6", parsefun=OneCodeToFindThemAll, strict=FALSE,
                   insert=500, BPPARAM=SerialParam(progress=FALSE))
length(teann)
teann[1]
```

As expected, we get a lower number of elements in the annotations, because
repeats that are not TEs have been removed. Furthermore, some fragmented
regions of TEs have been assembled together.

This time, the resulting `teann` object is of class `GRangesList`. Each
element of the list represents an assembled TE containing a `GRanges` object of
length one, if the TE could not be not assembled with another element, or of
length greater than one, if two or more fragments were assembled together into a
single TE.

We can get more information of the parsed annotations by accessing the 
metadata columns with `mcols()`:

```{r}
mcols(teann)
```

There is information about the reconstruction status of the TE (*Status*
column), the relative length of the reconstructed TE (*RelLength*) and the
repeat class of the TE (*Class*). The relative length is calculated by adding
the length (in base pairs) of all fragments from the same assembled TE, and
dividing that sum by the length (in base pairs) of the consensus sequence. For
full-length and partially reconstructed LTR TEs, the consensus sequence length
used is the one resulting from adding twice the consensus sequence length of
the long terminal repeat (LTR) and the one from the corresponding internal
region. For solo-LTRs, the consensus sequence length of the long terminal
repeat is used.

We can get an insight into the composition of the assembled annotations using
the information from the *status* column. Let's look at the absolute
frequencies of the status and class of TEs in the annotations.

```{r comparsedann, message=FALSE, height=6, width=10, out.width="800px", fig.cap="Composition of parsed TE annotations.", echo=FALSE}
library(RColorBrewer)

pal1 <- brewer.pal(6, "Pastel2")
pal2 <- brewer.pal(length(unique(mcols(teann)$Class)), "Set2")

par(mfrow = c(1,2), mar = c(5,4,3,1))
bp1 <- barplot(table(mcols(teann)$Status), col = pal1, border = "black",
        main = "TEs by status", cex.axis=0.8, xaxt = "n")
grid(nx=NA, ny=NULL)
axis(1, at=bp1, labels = FALSE, las=1, line=0, lwd = 0, lwd.ticks = 1) 
par(xpd=TRUE)
text(x= bp1[, 1] - 0.3, y = 10, labels=names(table(mcols(teann)$Status)), 
     srt=40, offset = 1.7, cex = 0.8, pos = 1)
par(xpd=FALSE)

bp2 <- barplot(table(mcols(teann)$Class), col = pal2, border = "black",
        main = "TEs by class", cex.axis=0.8, xaxt = "n",
        ylim = c(0,max(table(mcols(teann)$Status))))
grid(nx=NA, ny=NULL)
axis(1, at=bp2, labels = FALSE, las=1, line=0, lwd = 0, lwd.ticks = 1) 
par(xpd=TRUE)
text(x= bp2[, 1] - 0.1, y = 10, labels=names(table(mcols(teann)$Class)), 
     srt=35, offset = 1.2, cex = 0.8, pos = 1)
par(xpd=FALSE)
```

Here, *full-lengthLTR* are reconstructed LTR retrotransposons following the
LTR - internal region (int) - LTR structure. Partially reconstructed LTR TEs
are *partialLTR_down* (internal region followed by a downstream LTR) and
*partialLTR_up* (LTR upstream of an internal region). *int* and *LTR*
correspond to internal and solo-LTR regions, respectively. Finally, the
*noLTR* refers to TEs of other classes (not LTR), as well as TEs of class LTR
which could not be identified as either internal or long terminal repeat
regions based on their name.

Moreover, the `atena` package provides getter functions to retrieve TEs of a
specific class, using a specific relative length threshold. Those TEs with
higher relative lengths are more likely to have intact open reading frames,
making them more interesting for expression quantification and functional
analyses. For example, to get LINEs with a minimum of 0.9 relative length, we
can use the `getLINEs()` function. We use the TE annotations in `teann` we
obtained before and set the `relLength` to 0.9.

```{r}
rmskLINE <- getLINEs(teann, relLength=0.9)
length(rmskLINE)
rmskLINE[1]
```

To get LTR retrotransposons, we can use the function `getLTRs()`. This function
also allows to get one or more specific types of reconstructed TEs. To get
full-length, partial LTRs and other fragments that could not be reconstructed,
we can:

```{r}
rmskLTR <- getLTRs(teann, relLength=0.8, fullLength=TRUE, partial=TRUE,
                   otherLTR=TRUE)
length(rmskLTR)
rmskLTR[1]
```

To obtain DNA transposons and SINEs, one can use the `getDNAtransposons()` and
`getSINEs()` functions, respectively.

# TE expression quantification

Quantification of TE expression with `atena` consists in the following two
steps:

1. Building of a parameter object for one of the available quantification
   methods.

2. Calling the TE expression quantification method `qtex()` using the
   previously built parameter object.
   
The dataset that will be used to illustrate how to quantify TE expression with
`atena` is a published RNA-seq dataset of _Drosophila melanogaster_ available
at the National Center for Biotechnology Information (NCBI) Gene Expression
Omnibus (GEO) under accession
[GSE47006](https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE47006)). 
The two selected samples are: a piwi knockdown and a piwi control (GSM1142845
and GSM1142844). These files have been subsampled. The piwi-associated
silencing complex (piRISC) silences TEs in the _Drosophila_ ovary, hence the
knockdown of piwi causes the de-repression of TEs. Here we show how the
expression of full-length LTR retrotransposons present in `rmskLTR` can be
easily quantified using `atena`.

## Building a parameter object

A parameter object is build calling a specific function for the quantification
method we want to use. Independenty of each method, all parameter object
constructor functions require that the first two arguments specify the BAM
files and the TE annotation, respectively.

### ERVmap

To use the ERVmap method in `atena` we should first build an object of the
class `ERVmapParam` using the function `ERVmapParam()`. The `singleEnd`
parameter is set to `TRUE` since the example BAM files are single-end. The
`ignoreStrand` parameter works analogously to the same parameter in the
function `summarizeOverlaps()` from package `r Biocpkg("GenomicAlignments")`
and should be set to `TRUE` whenever the RNA library preparation protocol was
stranded.

One of the filters applied by the ERVmap method compares the alignment score of
a given primary alignment, stored in the `AS` tag of a SAM record, to the
largest alignment score among every other secondary alignment, known as the
suboptimal alignment score. The original ERVmap software assumes that input BAM
files are generated using the Burrows-Wheeler Aligner (BWA) software
[@li2009fast], which stores suboptimal alignment scores in the `XS` tag.
Although `AS` is an optional tag, most short-read aligners provide this tag
with alignment scores in BAM files. However, the suboptimal alignment score,
stored in the `XS` tag by BWA, is either stored in a different tag or not
stored at all by other short-read aligner software, such as STAR
[@dobin2013star].

To enable using ERVmap on BAM files produced by short-read aligner software
other than BWA, `atena` allows the user to set the argument
`suboptimalAlignmentTag` to one of the following three possible values:

* The name of a tag different to `XS` that stores the suboptimal alignment
  score.

* The value "none", which will trigger the calculation of the suboptimal
  alignment score by searching for the largest value stored in the `AS` tag
  among all available secondary alignments.

* The value "auto" (default), by which `atena` will first extract the name of
  the short-read aligner software from the BAM file and if that software is
  BWA, then suboptimal alignment scores will be obtained from the `XS` tag.
  Otherwise, it will trigger the calculation previously explained for
  `suboptimalAlignemntTag="none"`.

Finally, this filter is applied by comparing the difference between alignment
and suboptimal alignment scores to a cutoff value, which by default is 5 but
can be modified using the parameter `suboptimalAlignmentCutoff`. The default
value 5 is the one employed in the original ERVmap software that assumes the
BAM file was generated with BWA and for which lower values are interpreted as
"equivalent to second best match has one or more mismatches than the best
match" [@tokuyama2018ervmap, pg. 12571]. From a different perspective, in BWA
the mismatch penalty has a value of 4 and therefore, a
`suboptimalAlignmentCutoff` value of 5 only retains those reads where the
suboptimal alignment has at least 1 mismatch more than the best match.
Therefore, the `suboptimalAlignmentCutoff` value is specific to the short-read
mapper software and we recommend to set this value according to the mismatch
penalty of that software. Another option is to set
`suboptimalAlignmentCutoff=NA`, which prevents the filtering of reads based on
this criteria, as set in the following example.

```{r}
bamfiles <- list.files(system.file("extdata", package="atena"),
                       pattern="*.bam", full.names=TRUE)
empar <- ERVmapParam(bamfiles, 
                     teFeatures=rmskLTR, 
                     singleEnd=TRUE, 
                     ignoreStrand=TRUE, 
                     suboptimalAlignmentCutoff=NA)
empar
```

In the case of paired-end BAM files (`singleEnd=FALSE`), two additional
arguments can be specified, `strandMode` and `fragments`:

* `strandMode` defines the behavior of the strand getter when internally
  reading the BAM files with the `GAlignmentPairs()` function. See the help
  page of `strandMode` in the `r Biocpkg("GenomicAlignments")` package for
  further details.
 
* `fragments` controls how read filtering and counting criteria are applied to
  the read mates in a paired-end read. To use the original ERVmap algorithm
  [@tokuyama2018ervmap] one should set `fragments=TRUE` (default when
  `singleEnd=FALSE`), which filters and counts each mate of a paired-end read
  independently (i.e., two read mates overlapping the same feature count twice
  on that feature, treating paired-end reads as if they were single-end). On
  the other hand, when `fragments=FALSE`, if the two read mates pass the
  filtering criteria and overlap the same feature, they count once on that
  feature. If either read mate fails to pass the filtering criteria, then both
  read mates are discarded.

An additional functionality with respect to the original ERVmap software is the
integration of gene and TE expression quantification. The original ERVmap
software doesn't quantify TE and gene expression coordinately and this can
potentially lead to counting twice reads that simultaneously overlap a gene and
a TE. In `atena`, gene expression is quantified based on the approach used in
the TEtranscripts software [@jin2015tetranscripts]: unique reads are preferably
assigned to genes, whereas multi-mapping reads are preferably assigned to TEs.

In case that a unique read does not overlap a gene or a multi-mapping read does
not overlap a TE, `atena` searches for overlaps with TEs or genes,
respectively. Given the different treatment of unique and multi-mapping reads,
`atena` requires the information regarding the _unique_ or _multi-mapping_
status of a read. This information is obtained from the presence of secondary
alignments in the BAM file or, alternatively, from the `NH` tag in the BAM file
(number of reported alignments that contain the query in the current SAM
record). Therefore, either secondary alignments or the `NH` tag need to be
present for gene expression quantification.

The original ERVmap approach does not discard any read overlapping gene
annotations. However, this can be changed using the parameter `geneCountMode`,
which by default `geneCountMode="all"` and follows the behavior in the original
ERVmap method. On the contrary, by setting `geneCountMode="ervmap"`, `atena`
also applies the filtering criteria employed to quantify TE expression to the
reads overlapping gene annotations.

Finally, `atena` also allows one to aggregate TE expression quantifications. By
default, the names of the input `GRanges` or `GRangesList` object given in the
`teFeatures` parameter are used to aggregate quantifications. However, the
`aggregateby` parameter can be used to specify other column names in the
feature annotations to be used to aggregate TE counts, for example at the
sub-family level.

### Telescope

To use the Telescope method for TE expression quantification, the
`TelescopeParam()` function is used to build a parameter object of the class
`TelescopeParam`.

As in the case of `ERVmapParam()`, the `aggregateby` argument, which should be
a character vector of column names in the annotation, determines the columns to
be used to aggregate TE expression quantifications. This way, `atena` provides
not only quantifications at the subfamily level, but also allows to quantify
TEs at the desired level (family, class, etc.), including locus based
quantifications. For such a use case, the object with the TE annotations should
include a column with unique identifiers for each TE locus and the
`aggregateby` argument should specify the name of that column. When
`aggregateby` is not specified, the `names()` of the object containing TE
annotations are used to aggregate quantifications.

Here, TE quantifications will be aggregated according to the `names()` of the
`rmskLTR` object.

```{r}
bamfiles <- list.files(system.file("extdata", package="atena"),
                       pattern="*.bam", full.names=TRUE)
tspar <- TelescopeParam(bfl=bamfiles, 
                        teFeatures=rmskLTR, 
                        singleEnd=TRUE, 
                        ignoreStrand=TRUE)
tspar
```

In case of paired-end data (`singleEnd=FALSE`), the argument usage is similar
to that of `ERVmapParam()`. In relation to the BAM file, Telescope follows the
same approach as the ERVmap method: when `fragments=FALSE`, only _mated read
pairs_ from opposite strands are considered, while when `fragments=TRUE`,
same-strand pairs, singletons, reads with unmapped pairs and other fragments
are also considered by the algorithm. However, there is one important
difference with respect to the counting approach followed by ERVmap: when
`fragments=TRUE` _mated read pairs_ mapping to the same element are counted
once, whereas in the ERVmap method they are counted twice.

As in the ERVmap method from `atena`, the gene expression quantification method
in Telescope is based on the approach of the TEtranscripts software
[@jin2015tetranscripts]. This way, `atena` provides the possibility to
integrate TE expression quantification by Telescope with gene expression
quantification. As in the case of the ERVmap method implemented in `atena`,
either secondary alignments or the `NH` tag are required for gene expression
quantification.

### TEtranscripts

Finally, the third method available is TEtranscripts. First, the
`TEtranscriptsParam()` function is called to build a parameter object of the
class `TEtranscriptsParam`. The usage of the `aggregateby` argument is the same
as in `TelescopeParam()` and `ERVmapParam()`. Locus based quantifications in
the TEtranscripts method from `atena` is possible because the TEtranscripts
algorithm actually computes TE quantifications at the locus level and then sums
up all instances of each TE subfamily to provide expression at the subfamily
level. By avoiding this last step, `atena` can provide TE expression
quantification at the locus level using the TEtranscripts method. For such a
use case, the object with the TE annotations should include a column with
unique identifiers for each TE and the `aggregateby` argument should specify
the name of that column.

In this example, the `aggregateby` argument will be set to
`aggregateby="repName"` in order to aggregate quantifications at the repeat
name level. Moreover, gene expression will also be quantified. To do so, gene
annotations are loaded from a *TxDb* object.

```{r, message=FALSE}
library(TxDb.Dmelanogaster.UCSC.dm6.ensGene)

txdb <- TxDb.Dmelanogaster.UCSC.dm6.ensGene
gannot <- exonsBy(txdb, by="gene")
length(gannot)
```

```{r}
bamfiles <- list.files(system.file("extdata", package="atena"),
                       pattern="*.bam", full.names=TRUE)
ttpar <- TEtranscriptsParam(bamfiles, 
                            teFeatures=rmskLTR,
                            geneFeatures=gannot,
                            singleEnd=TRUE, 
                            ignoreStrand=TRUE, 
                            aggregateby="repName")
ttpar
```

For paired-end data, where would set `singleEnd=FALSE`, the `fragments`
parameter has the same purpose as in `TelescopeParam()`. We can also
extract the TEs and gene combined feature set using the `features()`
function on the parameter object. A metadata column called `isTE` is added
to enable distinguishing TEs from gene annotations.

```{r}
features(ttpar)
mcols(features(ttpar))
table(mcols(features(ttpar))$isTE)
```

Regarding gene expression quantification, `atena` has implemented the approach
of the original TEtranscripts software [@jin2015tetranscripts]. As in the case
of the ERVmap and Telescope methods from `atena`, either secondary alignments
or the `NH` tag are required.

Following the gene annotation processing present in the TEtranscripts
algorithm, in case that `geneFeatures` contains a metadata column named "type",
only the elements with `type="exon"` are considered for quantification. If
those elements are grouped through a `GRangesList` object, then counts are
aggregated at the level of those `GRangesList` elements, such as genes or
transcripts. This also applies to the ERVmap and Telescope methods implemented
in `atena` when gene features are present. Let's see an example of this
processing:

```{r}
## Create a toy example of gene annotations
geneannot <- GRanges(seqnames=rep("2L", 8),
                     ranges=IRanges(start=c(1,20,45,80,110,130,150,170),
                                    width=c(10,20,35,10,5,15,10,25)),
                     strand="*", 
                     type=rep("exon",8))
names(geneannot) <- paste0("gene",c(rep(1,3),rep(2,4),rep(3,1)))
geneannot
ttpar2 <- TEtranscriptsParam(bamfiles, 
                             teFeatures=rmskLTR, 
                             geneFeatures=geneannot, 
                             singleEnd=TRUE, 
                             ignoreStrand=TRUE)
mcols(features(ttpar2))
features(ttpar2)[!mcols(features(ttpar2))$isTE]
```

### Quantifying expression

Finally, to quantify TE expression we call the `qtex()` method using one of the
previously defined parameter objects (`ERVmapParam`, `TEtranscriptsParam` or
`TelescopeParam`) according to the quantification method we want to use. As with
the `OneCodeToFindThemAll()` function described before, here we can also use the
`BPPARAM` parameter to perform calculations in parallel.

The `qtex()` method returns a `SummarizedExperiment` object containing the
resulting quantification of expression in an assay slot. Additionally, when a
`data.frame`, or `DataFrame`, object storing phenotypic data is passed to the
`qtex()` function through the `phenodata` parameter, this will be included as
column data in the resulting `SummarizedExperiment` object and the row names of
these phenotypic data will be set as column names in the output
`SummarizedExperiment` object.

In the current example, the call to quantify TE expression using the ERVmap
method would be the following:

```{r, results='hide'}
emq <- qtex(empar)
```
```{r}
emq
colSums(assay(emq))
```

In the case of the Telescope method, the call would be as follows:

```{r, results='hide'}
tsq <- qtex(tspar)
```
```{r}
tsq
colSums(assay(tsq))
```

For the TEtranscripts method, TE expression is quantified by using the
following call:

```{r, results='hide'}
ttq <- qtex(ttpar)
```
```{r}
ttq
colSums(assay(ttq))
```

As mentioned, TE expression quantification is provided at the repeat name
level.

# Accesing expression quantifications and metadata

The `qtex()` function returns a `SummarizedExperiment` object that, on the
one hand, stores the quantified expression in its assay data.

```{r}
head(assay(ttq))
```

On the other hand, it contains metadata about the features that may be useful
to select subsets of the quantified data and extract and explore the feature
annotations, using the function `rowData()` on this `SummarizedExperiment`
object.

```{r}
rowData(ttq)
```

Because we have aggregated quantifications by `RepName` the number of TE
quantified features has been substantially reduced with respect to the original
number of TE features.

```{r}
table(rowData(ttq)$isTE)
```

Let's say we want to select full-length LTRs features, this could be a way of
doing it.

```{r}
temask <- rowData(ttq)$isTE
fullLTRs <- rowData(ttq)$Status == "full-lengthLTR"
fullLTRs <- (sapply(fullLTRs, sum, na.rm=TRUE) == 1) &
            (lengths(rowData(ttq)$Status) == 1)
sum(fullLTRs)
rowData(ttq)[fullLTRs, ]
```

Note also that since we restricted expression quantification to LTRs, we do
have only quantification for that TE class.

```{r}
table(rowData(ttq)$Class[temask])
```

# Session information

```{r session_info, cache=FALSE}
sessionInfo()
```

# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annotations.R
\name{annotaTEs}
\alias{annotaTEs}
\title{Get RepeatMasker UCSC annotations}
\usage{
annotaTEs(
  genome = "hg38",
  parsefun = rmskidentity,
  verbose = TRUE,
  AHid = NULL,
  ...
)
}
\arguments{
\item{genome}{The genome version of the desired RepeatMasker annotations
(e.g. "hg38").}

\item{parsefun}{A \code{function} to parse the annotations:
\itemize{
  \item Function \code{rmskidentity} returns RepeatMasker annotations as
        present in \link[AnnotationHub]{AnnotationHub}, without
        processing them.
  \item Function \code{rmskbasicparser} parses annotations by removing
        low complexity regions, simple repeats, satellites, rRNA, scRNA,
        snRNA, srpRNA and tRNA. Also removes TEs
        with a strand different than "+" or "-". Modifies "repFamily" and
        "repClass" columns when a "?" is present or when they are defined
        as "Unknown" or "Other". Finally, assigns a unique id to each TE
        instance by adding the suffix "_dup" plus a number at the end of
        the "repName".
  \item Function \code{rmskatenaparser} parses RepeatMasker annotations 
        reconstructing fragmented TEs by assembling together fragments from
        the same TE that are close enough. For LTR class TEs, it tries to
        reconstruct full-length and partial TEs following the LTR - internal
        region - LTR structure. Input is a \code{GRanges} object and output
        is a \code{GRangesList} object.
  \item Function \code{OneCodeToFindThemAll} parses annotations following
        the 'One code to find them all' method by 
        \href{https://doi.org/10.1186/1759-8753-5-13}{(Bailly-Bechet et al. 2014)}. 
        Input is a \code{GRanges} object and output is a \code{GRangesList} 
        object.
  \item User-defined function. Input and output should be \code{GRanges}
        objects.
}}

\item{verbose}{(Default \code{TRUE}) Logical value indicating whether to
report progress.}

\item{AHid}{AnnotationHub unique identifier, of the form AH12345, of an
object with TE annotations. This is an optional argument to
specify a concrete AnnotationHub resource, for instance
when more there is more than one RepeatMasker annotation
available for a specific genome version. If \code{AHid} is
not specified, the latest RepeatMasker annotation is be used.}

\item{...}{Arguments passed to \code{parsefun}.}
}
\value{
A [`GRanges`][GenomicRanges::GRanges-class] object with
        transposable element annotations.
}
\description{
The \code{annotaTEs()} function fetches RepeatMasker UCSC transposable
element (TE) annotations using 
\link[AnnotationHub]{AnnotationHub} and parses them.
}
\details{
Given a specific genome version, the \code{annotaTEs()} function fetches
RepeatMasker annotations from UCSC Genome Browser using the 
\link[AnnotationHub]{AnnotationHub} package. Since RepeatMasker not only
provides TE annotations but also low complexity DNA sequences and other
types of repeats, a specific \code{parsefun} can be set to parse these
annotations (e.g. \code{rmskbasicparser} or a user-defined function). If no
parsing is required, \code{parsefun} can be set to \code{rmskidentity}.
}
\examples{
rmskid <- annotaTEs(genome="hg19", parsefun=rmskidentity)
rmskid


}
\seealso{
\code{\link[AnnotationHub]{AnnotationHub}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{class}
\name{QuantifyParam-class}
\alias{QuantifyParam-class}
\alias{path,QuantifyParam-method}
\alias{features,QuantifyParam-method}
\title{QuantifyParam parameter class}
\usage{
\S4method{path}{QuantifyParam}(object)

\S4method{features}{QuantifyParam}(x)
}
\arguments{
\item{object}{A [`QuantifyParam`] object.}

\item{x}{A [`QuantifyParam`] object.}
}
\value{
\code{path()}: Filesystem paths to the BAM files in the input
parameter object.

\code{features()}: The \code{GenomicRanges} or
\code{GenomicRangesList} object with the features in the input parameter
object.
}
\description{
This is a virtual class from which other classes are derived
for storing parameters provided to quantification methods of
transposable elements from RNA-seq data.
}
\section{Slots}{

\describe{
\item{\code{bfl}}{A [`BamFileList`][Rsamtools::BamFileList-class] object.}

\item{\code{features}}{A [`GRanges`][GenomicRanges::GRanges-class] object.}

\item{\code{aggregateby}}{Character vector with column names in the annotation
to be used to aggregate quantifications.}

\item{\code{ovMode}}{Character vector indicating the overlapping mode. Available
options are: "ovUnion" (default) and "ovIntersectionStrict",
which implement the corresponding methods from HTSeq
(\url{https://htseq.readthedocs.io/en/release_0.11.1/count.html}).
In the TEtranscripts, ERVmap and Telescope methods ambiguous
alignments (alignments overlapping > 1 feature) are addressed
differently depending on the method. In the atena method, those
overlaps are not counted.}
}}

\examples{
bamfiles <- list.files(system.file("extdata", package="atena"),
                       pattern="*.bam", full.names=TRUE)
\dontrun{
## use the following two instructions to fetch annotations, they are here
## commented out to enable running this example quickly when building and
## checking the package
rmskat <- annotaTEs(genome="dm6", parsefun=rmskatenaparser,
                    strict=FALSE, insert=500)
rmskLTR <- getLTRs(rmskat, relLength=0.8,
                   fullLength=TRUE,
                   partial=TRUE)
}

## DO NOT TYPE THIS INSTRUCTION, WHICH JUST LOADS A PRE-COMPUTED ANNOTATION
## YOU SHOULD USE THE INSTRUCTIONS ABOVE TO FETCH ANNOTATIONS
rmskLTR <- readRDS(system.file("extdata", "rmskatLTRrlen80flenpartoth.rds",
                               package="atena"))

## build a parameter object for TEtranscripts
ttpar <- TEtranscriptsParam(bamfiles,
                            teFeatures=rmskLTR,
                            singleEnd=TRUE,
                            ignoreStrand=TRUE)
## just check that the parameter object belongs to the expected classes
is(ttpar, "QuantifyParam")
is(ttpar, "TEtranscriptsParam")

}
\seealso{
\code{\link{ERVmapParam-class}}
\code{\link{TelescopeParam-class}}
\code{\link{TEtranscriptsParam-class}}
\code{\link{atenaParam-class}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annotations.R
\name{OneCodeToFindThemAll}
\alias{OneCodeToFindThemAll}
\title{OneCodeToFindThemAll parser of RepeatMasker annotations}
\usage{
OneCodeToFindThemAll(
  gr,
  dictionary = NULL,
  fuzzy = FALSE,
  strict = FALSE,
  insert = -1,
  BPPARAM = SerialParam(progressbar = TRUE)
)
}
\arguments{
\item{gr}{A [`GRanges`][GenomicRanges::GRanges-class] object with
RepeatMasker annotations from \link[AnnotationHub]{AnnotationHub}}

\item{dictionary}{(Default NULL) When NULL, a dictionary is built based 
on names of repeats. If not, a data.frame with equivalences LTR - internal
regions created by the user, where first column should be the name of the
internal region and the second column should be the LTR(s). When more than
one LTR, these should be separated by ":".}

\item{fuzzy}{(Default FALSE) A logical; if TRUE, the search for equivalences
between internal parts and LTRs to reconstruct LTR class transposable 
elements is less stringent, allowing more matches between corresponding
subparts. This option can increase the proportion of false positives
(incorrectly reconstructed LTR class TEs).}

\item{strict}{(Default FALSE) A logical; if TRUE, the 80-80 rule is applied,
i.e. only copies with more than 80% identity to the reference
and more than 80 bp long are reported.}

\item{insert}{(Default -1) An integer. When \code{insert} < 0, two fragments
are assembled if the distance separating their furthest extremities is less 
than twice the reference length of the element. When \code{insert} > 0,
fragments are assembled if the distance between their closest extremities
is equal or less than \code{insert}. When \code{insert} = 0, two fragments
are assembled if they are in contact next to each other.}

\item{BPPARAM}{See \code{?\link[BiocParallel:bplapply]{bplapply}} in the 
BiocParallel package. Can be used to run calculations in parallel.}
}
\value{
A [`GRangesList`][GenomicRanges::GRangesList-class] object.
}
\description{
OneCodeToFindThemAll parser of RepeatMasker annotations
}
\details{
Implementation of One code to find them all 
\href{https://doi.org/10.1186/1759-8753-5-13}{(Bailly-Bechet et al. 2014)}.
Parses RepeatMasker annotations from UCSC by assembling together fragments
from the same transposable elemenet (TE) that are close enough (determined
by the \code{insert} parameter). For TEs from the LTR class, the parser
tries to reconstruct full-length, when possible, or partial TEs following
the LTR - internal region - LTR structure. Equivalences between internal
regions and flanking LTRs can be set by the user with the \code{dictionary}
parameter or can be obtained by the parser. In this last case, the 
\code{fuzzy} parameter determines the level of stringency when searching 
for LTR - internal region equivalences.
}
\examples{
\dontrun{
rmskoc <- annotaTEs(genome="dm6", parsefun=OneCodeToFindThemAll,
                    fuzzy=FALSE, strict=FALSE)
}

}
\references{
Bailly-Bechet et al. "One code to find them all": a perl tool to 
conveniently parse RepeatMasker output files. Mobile DNA. 2014;5(1):1-15.
DOI: \url{https://doi.org/10.1186/1759-8753-5-13}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annotations.R
\name{rmskatenaparser}
\alias{rmskatenaparser}
\title{atena annotation parser of RepeatMasker annotations}
\usage{
rmskatenaparser(gr, strict = FALSE, insert = 1000)
}
\arguments{
\item{gr}{A [`GRanges`][GenomicRanges::GRanges-class] object with
RepeatMasker annotations from \link[AnnotationHub]{AnnotationHub}}

\item{strict}{(Default FALSE) A logical; if TRUE, the 80-80 rule is applied,
i.e. only copies with more than 80% identity to the reference
and more than 80 bp long are reported.}

\item{insert}{(Default 1000L) An integer > 0. Fragments are assembled 
together if the distance between their closest extremities
is equal or less than \code{insert}. When \code{insert} = 0, two fragments
are assembled if they are in contact next to each other.}
}
\value{
A [`GRangesList`][GenomicRanges::GRangesList-class] object.
}
\description{
atena annotation parser of RepeatMasker annotations
}
\details{
atena annotation parser of RepeatMasker annotations.
Parses RepeatMasker annotations from UCSC by assembling together fragments
from the same transposable element (TE) that are close enough (determined
by the \code{insert} parameter). For TEs from the LTR class, the parser
tries to reconstruct full-length, when possible, or partial TEs following
the LTR - internal region - LTR structure. Equivalences between LTR and
internal regions are found by, first, identifying LTR regions (those with
the "LTR" substring in their name) and internal regions (those with a
suffix such as "-int", "-I", etc.). Then, LTR are assigned to internal
regions for which the comparison of the two names are has a higher number
of equal consecutive characters.
}
\examples{
rmskat <- annotaTEs(genome="dm6", parsefun=rmskatenaparser,
                    strict=FALSE)
rmskat

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R, R/TEtranscripts.R
\docType{class}
\name{TEtranscriptsParam-class}
\alias{TEtranscriptsParam-class}
\alias{TEtranscriptsParam}
\alias{show,TEtranscriptsParam-method}
\title{TEtranscripts parameter class}
\usage{
TEtranscriptsParam(
  bfl,
  teFeatures,
  aggregateby = character(0),
  ovMode = "ovUnion",
  geneFeatures = NULL,
  singleEnd = TRUE,
  ignoreStrand = FALSE,
  strandMode = 1L,
  fragments = TRUE,
  tolerance = 1e-04,
  maxIter = 100L,
  verbose = TRUE
)

\S4method{show}{TEtranscriptsParam}(object)
}
\arguments{
\item{bfl}{a character string vector of BAM file names.}

\item{teFeatures}{A \code{GRanges} or \code{GRangesList} object with the
TE annotated features to be quantified. Elements in this object should have
names, which are used as a grouping factor for genomic ranges forming a
common locus, unless other metadata column names are specified in the
\code{aggregateby} parameter.}

\item{aggregateby}{Character vector with column names from the annotation
to be used to aggregate quantifications. By default, this is an empty
vector, which means that the names of the input \code{GRanges} or
\code{GRangesList} object given in the \code{teFeatures} parameter are used
to aggregate quantifications.}

\item{ovMode}{Character vector indicating the overlapping mode. Available
options are: "ovUnion" (default) and "ovIntersectionStrict",
which implement the corresponding methods from HTSeq
(\url{https://htseq.readthedocs.io/en/release_0.11.1/count.html}).
Ambiguous alignments (alignments overlapping > 1 feature) are addressed
as in the original TEtranscripts method.}

\item{geneFeatures}{(Default NULL) A \code{GRanges} or 
\code{GRangesList} object with the
gene annotated features to be quantified. Following the TEtranscripts
algorithm, overlaps with unique reads are first tallied with respect to
these gene features. Elements should have names indicating the gene name/id.
In case that \code{geneFeatures} is a \code{GRanges} and contains a 
metadata column named
\code{type}, only the elements with \code{type} = \code{exon} are considered
for the analysis. Then, exon counts are summarized to the gene level. If
NULL, gene expression is not quantified.}

\item{singleEnd}{(Default TRUE) Logical value indicating if reads are single
(\code{TRUE}) or paired-end (\code{FALSE}).}

\item{ignoreStrand}{(Default FALSE) Logical value that defines if the strand
should be taken into consideration when computing the overlap between reads
and annotated features. When \code{ignoreStrand = FALSE}, an aligned read
is considered to be overlapping an annotated feature as long as they
have a non-empty intersecting genomic range on the same strand, while when
\code{ignoreStrand = TRUE} the strand is not considered.}

\item{strandMode}{(Default 1) Numeric vector which can take values 0, 1 or
2.
The strand mode is a per-object switch on
\code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
objects that controls the behavior of the strand getter. See
\code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
class for further detail. If \code{singleEnd = TRUE}, then \code{strandMode}
is ignored.}

\item{fragments}{(Default TRUE) Logical value applied to paired-end data
only. In both cases (\code{fragments=FALSE} and \code{fragments=TRUE}), the
read-counting method discards not properly paired reads. Moreover,
when \code{fragments=FALSE}, only non-ambiguous properly paired reads are
counted. When \code{fragments=TRUE}, ambiguous reads are also counted 
(see "Pairing criteria" in \code{\link[GenomicAlignments]{readGAlignments}()}). 
\code{fragments=TRUE} is equivalent to
the behavior of the TEtranscripts algorithm. For further details see
\code{\link[GenomicAlignments]{summarizeOverlaps}()}.}

\item{tolerance}{A positive numeric scalar storing the minimum tolerance
above which the SQUAREM algorithm (Du and Varadhan, 2020) keeps iterating.
Default is \code{1e-4} and this value is passed to the \code{tol} parameter
of the \code{\link[SQUAREM]{squarem}()} function.}

\item{maxIter}{A positive integer scalar storing the maximum number of
iterations of the SQUAREM algorithm (Du and Varadhan, 2020). Default
is 100 and this value is passed to the \code{maxiter} parameter of the
\code{\link[SQUAREM]{squarem}()} function.}

\item{verbose}{(Default \code{TRUE}) Logical value indicating whether to
report progress.}

\item{object}{A \linkS4class{TEtranscriptsParam} object.}
}
\value{
A \linkS4class{TEtranscriptsParam} object.
}
\description{
This is a class for storing parameters provided to the TEtranscripts
algorithm. It is a subclass of the [`QuantifyParam`].

Build an object of the class \code{TEtranscriptsParam}
}
\details{
This is the constructor function for objects of the class
\code{TEtranscriptsParam-class}. This type of object is the input to the
function \code{\link{qtex}()} for quantifying expression of transposable
elements using the TEtranscripts method
\href{https://doi.org/10.1093/bioinformatics/btv422}{Jin et al. (2015)}. The
TEtranscripts algorithm quantifies TE expression by using an EM algorithm
to optimally distribute ambiguously mapped reads.
}
\section{Slots}{

\describe{
\item{\code{singleEnd}}{(Default FALSE) Logical value indicating if reads are single
(\code{TRUE}) or paired-end (\code{FALSE}).}

\item{\code{ignoreStrand}}{(Default FALSE) A logical which defines if the strand
should be taken into consideration when computing the overlap between reads
and annotated features. When \code{ignoreStrand = FALSE}, an aligned read
will be considered to be overlapping an annotated feature as long as they
have a non-empty intersecting genomic ranges on the same strand, while when
\code{ignoreStrand = TRUE} the strand will not be considered.}

\item{\code{strandMode}}{(Default 1) Numeric vector which can take values 0, 1 or 2.
The strand mode is a per-object switch on
\code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
objects that controls the behavior of the strand getter. See
\code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
class for further detail. If \code{singleEnd = TRUE}, then use either
\code{strandMode = NULL} or do not specify the \code{strandMode} parameter.}

\item{\code{fragments}}{(Default TRUE) A logical; applied to paired-end data only.
In both cases (\code{fragments=FALSE} and \code{fragments=TRUE}), the
read-counting method discards not properly paired reads. Moreover,
when \code{fragments=FALSE}, only non-ambiguous properly paired reads are
counted. When \code{fragments=TRUE}, ambiguous reads are also counted 
(see "Pairing criteria" in \code{\link[GenomicAlignments]{readGAlignments}()}). 
\code{fragments=TRUE} is equivalent to
the behavior of the TEtranscripts algorithm. For further details see
\code{\link[GenomicAlignments]{summarizeOverlaps}()}.}

\item{\code{tolerance}}{A positive numeric scalar storing the minimum tolerance
above which the SQUAREM algorithm (Du and Varadhan, 2020) keeps iterating.
Default is \code{1e-4} and this value is passed to the \code{tol} parameter
of the \code{\link[SQUAREM]{squarem}()} function.}

\item{\code{maxIter}}{A positive integer scalar storing the maximum number of
iterations of the SQUAREM algorithm (Du and Varadhan, 2020). Default
is 100 and this value is passed to the \code{maxiter} parameter of the
\code{\link[SQUAREM]{squarem}()} function.}
}}

\examples{
bamfiles <- list.files(system.file("extdata", package="atena"),
                       pattern="*.bam", full.names=TRUE)
\dontrun{
## use the following two instructions to fetch annotations, they are here
## commented out to enable running this example quickly when building and
## checking the package
rmskat <- annotaTEs(genome="dm6", parsefun=rmskatenaparser,
                    strict=FALSE, insert=500)
rmskLTR <- getLTRs(rmskat, relLength=0.8,
                   fullLength=TRUE,
                   partial=TRUE,
                   otherLTR=TRUE)
}

## DO NOT TYPE THIS INSTRUCTION, WHICH JUST LOADS A PRE-COMPUTED ANNOTATION
## YOU SHOULD USE THE INSTRUCTIONS ABOVE TO FETCH ANNOTATIONS
rmskLTR <- readRDS(system.file("extdata", "rmskatLTRrlen80flenpartoth.rds",
                               package="atena"))


library(TxDb.Dmelanogaster.UCSC.dm6.ensGene)
txdb <- TxDb.Dmelanogaster.UCSC.dm6.ensGene
txdb_genes <- genes(txdb)

## build a parameter object for TEtranscripts
ttpar <- TEtranscriptsParam(bamfiles,
                            teFeatures=rmskLTR,
                            geneFeatures=txdb_genes,
                            singleEnd=TRUE,
                            ignoreStrand=TRUE,
                            aggregateby="repName")
ttpar

}
\references{
Jin Y et al. TEtranscripts: a package for including transposable elements
in differential expression analysis of RNA-seq datasets.
Bioinformatics. 2015;31(22):3593-3599. DOI:
\url{https://doi.org/10.1093/bioinformatics/btv422}

Jin Y et al. TEtranscripts: a package for including transposable elements
in differential expression analysis of RNA-seq datasets.
Bioinformatics. 2015;31(22):3593-3599. DOI:
\url{https://doi.org/10.1093/bioinformatics/btv422}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R, R/ERVmap.R
\docType{class}
\name{ERVmapParam-class}
\alias{ERVmapParam-class}
\alias{ERVmapParam}
\alias{show,ERVmapParam-method}
\title{ERVmap parameter class}
\usage{
ERVmapParam(
  bfl,
  teFeatures,
  aggregateby = character(0),
  ovMode = "ovUnion",
  geneFeatures = NULL,
  singleEnd = TRUE,
  ignoreStrand = TRUE,
  strandMode = 1L,
  fragments = !singleEnd,
  maxMismatchRate = 0.02,
  suboptimalAlignmentTag = "auto",
  suboptimalAlignmentCutoff = 5,
  geneCountMode = "all",
  verbose = TRUE
)

\S4method{show}{ERVmapParam}(object)
}
\arguments{
\item{bfl}{A \code{BamFile} or \code{BamFileList} object, or a character
string vector of BAM filenames.}

\item{teFeatures}{A \code{GRanges} or \code{GRangesList} object with the
transposable element (TE) annotated features to be quantified. Elements in 
this object should have names, which are used as a grouping factor for 
genomic ranges forming a common locus, unless other metadata column names 
are specified in the \code{aggregateby} parameter.}

\item{aggregateby}{Character vector with column names in the annotation
to be used to aggregate quantifications. By default, this is an empty
vector, which means that the names of the input \code{GRanges} or 
\code{GRangesList} object given in the \code{teFeatures} parameter are used
to aggregate quantifications.}

\item{ovMode}{Character vector indicating the overlapping mode. Available
options are: "ovUnion" (default) and "ovIntersectionStrict",
which implement the corresponding methods from HTSeq
(\url{https://htseq.readthedocs.io/en/release_0.11.1/count.html}).
Ambiguous alignments (alignments overlapping > 1 feature) are addressed
as in the original ERVmap algorithm.}

\item{geneFeatures}{(Default NULL) A \code{GRanges} or 
\code{GRangesList} object with the
gene annotated features to be quantified. Overlaps with unique reads are 
first tallied with respect to these gene features. Elements should have 
names indicating the gene name/id. In case that \code{geneFeatures} 
is a \code{GRanges} and contains
a metadata column named \code{type}, only the elements with 
\code{type} = \code{exon} are considered for the analysis. Then, exon
counts are summarized to the gene level. If NULL, gene expression is
not quantified.}

\item{singleEnd}{(Default TRUE) Logical value indicating if reads are single
(\code{TRUE}) or paired-end (\code{FALSE}).}

\item{ignoreStrand}{(Default TRUE) A logical which defines if the strand
should be taken into consideration when computing the overlap between reads
and annotated features. When \code{ignore_strand = FALSE}, an aligned read
is considered to be overlapping an annotated feature as long as they
have a non-empty intersecting genomic range on the same strand, while when
\code{ignoreStrand = TRUE} the strand is not considered.}

\item{strandMode}{(Default 1) Numeric vector which can take values 0, 1 or 
2.
  The strand mode is a per-object switch on
  \code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
  objects that controls the behavior of the strand getter. See
  \code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
  class for further detail. If \code{singleEnd = TRUE}, then
  \code{strandMode} is ignored.}

\item{fragments}{(Default not \code{singleEnd}) A logical; applied to
paired-end data only. When \code{fragments=TRUE}, the read-counting
method in the original ERVmap algorithm is applied: each mate of a
paired-end read is counted (including ambiguous and not properly paired 
reads). When
\code{fragments=FALSE}, if the two mates of a paired-end read map to the
same element, they are counted as a single hit and singletons, reads with
unmapped pairs and other ambiguous or not properly paired fragments are 
not counted (see "Pairing criteria" in 
\code{\link[GenomicAlignments]{readGAlignments}()}).}

\item{maxMismatchRate}{(Default 0.02) Numeric value storing the maximum
mismatch rate employed by the ERVmap algorithm to discard aligned reads
whose rate of sum of hard and soft clipping or whose rate of the edit
distance over the genome reference to the length of the read is above this
threshold.}

\item{suboptimalAlignmentTag}{(Default "auto") Character string storing the
tag name in the BAM files that stores the suboptimal alignment score used in
the third filter of ERVmap; see
\href{https://doi.org/10.1073/pnas.1814589115}{Tokuyama et al. (2018)}.
The default, \code{suboptimalAlignmentTag="auto"}, first extracts the name
of the read mapper software from one or more BAM files. If BAM files were
generated by BWA, the suboptimal alignment scores are obtained from a tag
called \code{XS}. For other read mappers, the suboptimal alignment score
is considered to be missing since, except from BWA, no other aligner
provides a tag with suboptimal alignment scores. In this case, the available
secondary alignments are used to implement an analogous approach to that
of the third ERVmap filter. When \code{suboptimalAlignmentTag="none"}, it
also performs the latter approach even when the tag \code{XS} is available.
When this parameter is different from \code{"auto"} and \code{"none"}, a tag
with the given name is used to extract the suboptimal alignment score.}

\item{suboptimalAlignmentCutoff}{(Default 5) Numeric value storing the
cutoff above which the difference between the alignment score and the
suboptimal alignment score is considered sufficiently large to retain the
alignment. When this value is set to \code{NA}, the filtering step based on
suboptimal alignment scores is skipped.}

\item{geneCountMode}{(Default \code{"all"}) Character string indicating if 
the ERVmap read filters applied to quantify TEs expression should also be 
applied when quantifying gene expression (\code{"ervmap"}) or not 
(\code{"all"}), in which case all primary alignments mapping to genes are 
counted.}

\item{verbose}{(Default \code{TRUE}) Logical value indicating whether to
report progress.}

\item{object}{A \linkS4class{ERVmapParam} object.}
}
\value{
A \linkS4class{ERVmapParam} object.
}
\description{
This is a class for storing parameters provided to the ERVmap algorithm.
It is a subclass of the [`QuantifyParam`] class.

Build an object of the class \code{ERVmapParam}
}
\details{
This is the constructor function for objects of the class
\code{ERVmapParam-class}. This type of object is the input to the
function \code{\link{qtex}()} for quantifying expression of transposable
elements using the ERVmap method
\href{https://doi.org/10.1073/pnas.1814589115}{Tokuyama et al. (2018)}. The
ERVmap algorithm processes reads following conservative filtering criteria
to provide reliable raw count data for each TE.
}
\section{Slots}{

\describe{
\item{\code{readMapper}}{The name of the software used to align reads, obtained from
the BAM file header.}

\item{\code{singleEnd}}{(Default FALSE) Logical value indicating if reads are single
(\code{TRUE}) or paired-end (\code{FALSE}).}

\item{\code{strandMode}}{(Default 1) Numeric vector which can take values 0, 1 or 2.
The strand mode is a per-object switch on
\code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
objects that controls the behavior of the strand getter. See
\code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
class for further detail. If \code{singleEnd = TRUE}, then
\code{strandMode} #'   is ignored.}

\item{\code{ignoreStrand}}{(Default TRUE) A logical which defines if the strand
should be taken into consideration when computing the overlap between reads
and TEs in the annotations. When \code{ignore_strand = FALSE}, only those
reads which overlap the TE and are on the same strand are counted. On the
contrary, when \code{ignore_strand = TRUE}, any read overlapping an element
in \code{teFeatures} is counted regardless of the strand.}

\item{\code{fragments}}{(Default not \code{singleEnd}) A logical; applied to
paired-end data only. When \code{fragments=TRUE}, the read-counting
method in the original ERVmap algorithm is applied: each mate of a
paired-end read is counted (including ambiguous and not properly paired 
reads). When
\code{fragments=FALSE}, if the two mates of a paired-end read map to the
same element, they are counted as a single hit and singletons, reads with
unmapped pairs and other ambiguous or not properly paired fragments are 
not counted (see "Pairing criteria" in 
\code{\link[GenomicAlignments]{readGAlignments}()}).}

\item{\code{maxMismatchRate}}{(Default 0.02) Numeric value storing the maximum
mismatch rate employed by the ERVmap algorithm to discard aligned reads
whose rate of sum of hard and soft clipping, or of the edit distance over
the genome reference, to the length of the read is above this threshold.}

\item{\code{suboptimalAlignmentTag}}{(Default "auto") Character string storing the
tag name in the BAM files that stores the suboptimal alignment score used in
the third filter of ERVmap; see Tokuyama et al. (2018). The default,
\code{suboptimalAlignmentTag="auto"}, assumes that either the BAM files were
generated by BWA and include a tag called \code{XS} that stores the
suboptimal alignment score or, if the \code{XS} tag is not available, then
it uses the available secondary alignments to implement an analogous
approach to that of the third ERVmap filter. When
\code{suboptimalAlignmentTag="none"}, it also performs the latter approach
even when the tag \code{XS} is available.
When this parameter is different from \code{"auto"} and \code{"none"}, a tag
with the given name is used to extract the suboptimal alignment score.
The absence of that tag will prompt an error.}

\item{\code{suboptimalAlignmentCutoff}}{(Default 5) Numeric value storing the cutoff
above which the difference between the alignment score and the suboptimal
alignment score is considered sufficiently large to retain the alignment.
When this value is set to \code{NA}, then the filtering step based on
suboptimal alignment scores is skipped.}

\item{\code{geneCountMode}}{(Default "all") Character string indicating if the
ERVmap read filters applied to quantify TEs expression should also be
applied when quantifying gene expression ("ervmap") or not ("all"), in which
case all primary alignments mapping to genes are counted.}
}}

\examples{
bamfiles <- list.files(system.file("extdata", package="atena"),
                       pattern="*.bam", full.names=TRUE)
\dontrun{
rmskat <- annotaTEs(genome="dm6", parsefun=rmskatenaparser,
                    strict=FALSE, insert=500)
rmskLTR <- getLTRs(rmskat, relLength=0.8,
                   fullLength=TRUE,
                   partial=TRUE,
                   otherLTR=TRUE)
}

## DO NOT TYPE THIS INSTRUCTION, WHICH JUST LOADS A PRE-COMPUTED ANNOTATION
## YOU SHOULD USE THE INSTRUCTIONS ABOVE TO FETCH ANNOTATIONS
rmskLTR <- readRDS(system.file("extdata", "rmskatLTRrlen80flenpartoth.rds",
                               package="atena"))

## build a parameter object for ERVmap
empar <- ERVmapParam(bamfiles,
                     teFeatures=rmskLTR,
                     singleEnd=TRUE,
                     ignoreStrand=TRUE,
                     suboptimalAlignmentCutoff=NA)
empar

}
\references{
Tokuyama M et al. ERVmap analysis reveals genome-wide transcription of human
endogenous retroviruses. PNAS. 2018;115(50):12565-12572. DOI:
\url{https://doi.org/10.1073/pnas.1814589115}

Tokuyama M et al. ERVmap analysis reveals genome-wide transcription of human
endogenous retroviruses. PNAS. 2018;115(50):12565-12572. DOI:
\url{https://doi.org/10.1073/pnas.1814589115}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ERVmap.R, R/TEtranscripts.R, R/Telescope.R,
%   R/atenaMethod.R, R/qtex.R
\name{qtex,ERVmapParam-method}
\alias{qtex,ERVmapParam-method}
\alias{qtex}
\alias{qtex,TEtranscriptsParam-method}
\alias{qtex,TelescopeParam-method}
\alias{qtex,atenaParam-method}
\alias{qtex,AtenaParam-method}
\title{Quantify transposable element expression}
\usage{
\S4method{qtex}{ERVmapParam}(
  x,
  phenodata = NULL,
  mode = ovUnion,
  yieldSize = 1000000L,
  verbose = 1,
  BPPARAM = SerialParam(progressbar = ifelse(verbose == 1, TRUE, FALSE))
)

\S4method{qtex}{TEtranscriptsParam}(
  x,
  phenodata = NULL,
  mode = ovUnion,
  yieldSize = 1000000L,
  BPPARAM = SerialParam(progressbar = TRUE)
)

\S4method{qtex}{TelescopeParam}(
  x,
  phenodata = NULL,
  mode = ovUnion,
  yieldSize = 1000000L,
  auxiliaryFeatures = FALSE,
  BPPARAM = SerialParam(progressbar = TRUE)
)

\S4method{qtex}{atenaParam}(
  x,
  phenodata = NULL,
  mode = ovUnion,
  yieldSize = 1000000L,
  auxiliaryFeatures = FALSE,
  BPPARAM = SerialParam(progressbar = TRUE)
)
}
\arguments{
\item{x}{An \code{QuantifyParam} object of one of the following
subclasses:
\itemize{
  \item A \code{TEtranscriptsParam} object built using the constructor
        function \code{\link{TEtranscriptsParam}()}. This object will
        trigger \code{qtex()} to use the quantification algorithm by
        Jin et al. (2015).
  \item A \code{ERVmapParam} object built using the constructor
        function \code{\link{ERVmapParam}()}. This object will
        trigger \code{qtex()} to use the quantification algorithm by
        Tokuyama et al. (2018).
  \item A \code{TelescopeParam} object built using the constructor
        function \code{\link{TelescopeParam}()}. This object will
        trigger \code{qtex()} to use the quantification algorithm by
        Bendall et al. (2019).
  \item An \code{atenaParam} object built using the constructor
        function \code{\link{atenaParam}()}. This object will
        trigger \code{qtex()} to use a quantification algorithm
        specifically developed in this package.
}}

\item{phenodata}{A \code{data.frame} or \code{DataFrame} object storing
phenotypic data to include in the resulting
[`SummarizedExperiment`][SummarizedExperiment::SummarizedExperiment-class]
object. If \code{phenodata} is set,
its row names will become the column names of the resulting
[`SummarizedExperiment`][SummarizedExperiment::SummarizedExperiment-class]
object.}

\item{mode}{One of the pre-defined overlapping methods such as
\code{ovUnion()}, \code{ovIntersectionStrict} or a user-supplied
overlapping function. For a user-supplied overlapping function, the
input parameters must match those of the pre-defined methods and
the function must return a \code{\link[S4Vectors]{Hits}} object with
subject hits matching the annotated features. This parameter is
analogous to the \code{mode} parameter of the
\code{\link[GenomicAlignments]{summarizeOverlaps}()} function from
the \code{GenomicAlignments} package.}

\item{yieldSize}{Field inherited from \code{\link[Rsamtools]{BamFile}}.
The method for signature \code{\link{ERVmapParam}()} reads the BAM
file by chunks. \code{yieldSize} represents the number of records
(chunk size) to yield each time the file is read.}

\item{verbose}{(Default 1). When \code{verbose} > 1, detailed information on
the quantification steps is provided. Warnings are always present
regardless of the value of \code{verbose}.}

\item{BPPARAM}{An object of a
[`BiocParallelParam`][BiocParallel::BiocParallelParam-class] subclass
to configure the parallel execution of the code. By default,
[`BiocParallelParam`][BiocParallel::SerialParam-class] object is used,
which does not use any parallelization, with the flag
\code{progress=TRUE} to show progress through the calculations.}

\item{auxiliaryFeatures}{(Default \code{FALSE}). It only applies when `x` is
a [`TelescopeParam`] or an [`atenaParam`] object. When \code{TRUE},
auxiliary features created during expression quantification are also
returned in the [`SummarizedExperiment`]
[`SummarizedExperiment`][SummarizedExperiment::SummarizedExperiment-class]
object.}
}
\value{
A
       [`SummarizedExperiment`][SummarizedExperiment::SummarizedExperiment-class]
       object.
}
\description{
The \code{qtex()} method quantifies transposable element expression.
}
\details{
Giving some \code{AtenaParam} object sub-class as input, the
\code{qtex()} method quantifies the expression of transposable
elements (TEs). The particular algorithm to perform the
quantification will be selected depending on the specific
sub-class of input \code{AtenaParam} object, see argument
\code{x} above.
}
\examples{
bamfiles <- list.files(system.file("extdata", package="atena"),
                       pattern="*.bam", full.names=TRUE)
\dontrun{
## use the following two instructions to fetch annotations, they are here
## commented out to enable running this example quickly when building and
## checking the package
rmskat <- annotaTEs(genome="dm6", parsefun=rmskatenaparser,
                    strict=FALSE, insert=500)
rmskLTR <- getLTRs(rmskat, relLength=0.8,
                   fullLength=TRUE,
                   partial=TRUE,
                   otherLTR=TRUE)
}

## DO NOT TYPE THIS INSTRUCTION, WHICH JUST LOADS A PRE-COMPUTED ANNOTATION
## YOU SHOULD USE THE INSTRUCTIONS ABOVE TO FETCH ANNOTATIONS
rmskLTR <- readRDS(system.file("extdata", "rmskatLTRrlen80flenpartoth.rds",
                               package="atena"))

## build a parameter object for Telescope
tspar <- TelescopeParam(bfl=bamfiles,
                        teFeatures=rmskLTR,
                        singleEnd=TRUE,
                        ignoreStrand=TRUE)
## quantify expression
qts <- qtex(tspar)
}
\references{
Jin Y et al. TEtranscripts: a package for including transposable elements
in differential expression analysis of RNA-seq datasets.
Bioinformatics. 2015;31(22):3593-3599. DOI:
\url{https://doi.org/10.1093/bioinformatics/btv422}

Tokuyama M et al. ERVmap analysis reveals genome-wide transcription of human
endogenous retroviruses. PNAS, 115(50):12565-12572, 2018.
\url{https://doi.org/10.1073/pnas.1814589115}

Bendall ML et al. Telescope: characterization of the retrotranscriptome by
accurate estimation of transposable element expression.
PLOS Computational Biology, 15:e1006453, 2019.
\url{https://doi.org/10.1371/journal.pcbi.1006453}
}
\seealso{
\code{\link{TEtranscriptsParam}}
\code{\link{ERVmapParam}}
\code{\link{TelescopeParam}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/overlappingModes.R
\name{ovUnion}
\alias{ovUnion}
\alias{ovIntersectionStrict}
\title{Pre-defined overlapping mode functions}
\usage{
ovUnion(reads, features, ignoreStrand, inter.feature = TRUE)

ovIntersectionStrict(reads, features, ignoreStrand, inter.feature = TRUE)
}
\arguments{
\item{reads}{A \code{GAlignments}, \code{GAlignmentList} or a
\code{GAlignmentPairs} object.}

\item{features}{A \code{GRanges} object with annotated features.}

\item{ignoreStrand}{(Default FALSE) A logical which defines if the strand
should be taken into consideration when computing the overlap between reads
and annotated features. When \code{ignoreStrand = FALSE}, an aligned read
will be considered to be overlapping an annotated feature as long as they
have a non-empty intersecting genomic ranges on the same strand, while when
\code{ignoreStrand = TRUE} the strand will not be considered.}

\item{inter.feature}{When TRUE, ambiguous alignments (alignments
overlapping > 1 features) are removed and not counted. When 
\code{inter.feature} is set to FALSE, these ambiguous overlaps are taken
into account and addressed differently depending on the TE quantification.}
}
\value{
A \code{Hits} object; see the \code{\link[S4Vectors]{Hits-class}}
manual page.
}
\description{
The following functions control the way in which overlaps between
aligned reads and annotated features are resolved when an aligned
read overlaps more than one feature on the same locus:
}
\details{
\itemize{
  \item \code{ovUnion()}: (default)
  \item \code{ovIntersectionStrict()}:
  \item User supplied: a function taking the same parameters as the
        previous three functions and returning a
        \code{\link[S4Vectors]{Hits}} object.
}

They take the following parameters:

These functions are given to the \code{mode} parameter of the
\code{\link{qtex}()} function and are similar to the functions
\code{\link[GenomicAlignments]{Union}()} and
\code{\link[GenomicAlignments]{IntersectionStrict}()} from the
\code{GenomicAlignments} package, with the difference that instead
of returning counts of reads overlapping annotated features, they
return the actual overlaps, because the counting is deferred to other
algorithms that follow some specific strategy when a read maps to
more than one feature. For this same reason, these functions lack
the \code{inter.feature} argument found in the corresponding functions
from the \code{GenomicAlignments} package.
}
\examples{
bamfiles <- list.files(system.file("extdata", package="atena"),
                       pattern="*.bam", full.names=TRUE)
\dontrun{
## use the following two instructions to fetch annotations, they are here
## commented out to enable running this example quickly when building and
## checking the package
rmskat <- annotaTEs(genome="dm6", parsefun=rmskatenaparser,
                    strict=FALSE, insert=500)
rmskLTR <- getLTRs(rmskat, relLength=0.8, fullLength=TRUE, partial=TRUE,
                   otherLTR=TRUE)
}

## DO NOT TYPE THIS INSTRUCTION, WHICH JUST LOADS A PRE-COMPUTED ANNOTATION
## YOU SHOULD USE THE INSTRUCTIONS ABOVE TO FETCH ANNOTATIONS
rmskLTR <- readRDS(system.file("extdata", "rmskatLTRrlen80flenpartoth.rds",
                               package="atena"))

## build a parameter object for Telescope
tspar <- TelescopeParam(bfl=bamfiles,
                        teFeatures=rmskLTR,
                        singleEnd=TRUE,
                        ignoreStrand=TRUE)

## quantify expression using the 'ovIntersectionStrict()' mode function
tsquant <- qtex(tspar, mode=ovIntersectionStrict)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annotations.R
\name{annotateTEsGetters}
\alias{annotateTEsGetters}
\alias{getLTRs}
\alias{getLINEs}
\alias{getSINEs}
\alias{getDNAtransposons}
\title{Getter functions of TE classes from parsed RepeatMasker annotations.}
\usage{
getLTRs(
  annot,
  relLength = 0.9,
  fullLength = TRUE,
  partial = FALSE,
  soloLTR = FALSE,
  otherLTR = FALSE,
  returnMask = FALSE
)

getLINEs(annot, relLength = 0.9, returnMask = FALSE)

getSINEs(annot, relLength = 0.9, returnMask = FALSE)

getDNAtransposons(annot, relLength = 0.9, returnMask = FALSE)
}
\arguments{
\item{annot}{A [`GRanges`][GenomicRanges::GRanges-class] or
[`GRangesList`][GenomicRanges::GRangesList-class] object obtained
with the function `annotaTES()`, using either [`OneCodeToFindThemAll`]
or [`rmskatenaparser`] as RepeatMasker parser functions.
Alternatively, if `annot` is a [`QuantifyParam`] or a
[`SummarizedExperiment`][SummarizedExperiment::SummarizedExperiment-class]
object produced by the `qtex()` function, this function will
attempt to extract the corresponding annotations from inside
those objects.}

\item{relLength}{(Default 0.9) Numeric value that can take values between 0
to 1. Sets the minimum relative length required for
features. Elements with a lower relative length than
\code{relLength} will be filtered. The relative length
used is the one obtained by \code{OneCodeToFindThemAll()}
or \code{rmskatenaparser()}.
(length of the reconstructed TE / length of the reference).}

\item{fullLength}{(Default TRUE) Logical value on whether reconstructed
full-length LTR TEs (elements with structure 
LTR - internal region - LTR) should be selected.}

\item{partial}{(Default FALSE) Logical value on whether partially
reconstructed LTR TEs should be selected (structure LTR -
internal region or internal region - LTR).}

\item{soloLTR}{(Default FALSE) Logical value on whether solo LTRs should be
selected. Note that only fragments unambiguously identified
as LTRs thanks to the identification of their equivalent
internal region are considered as LTRs.}

\item{otherLTR}{(Default FALSE) Logical value on whether other TEs from the
LTR class, not included in any of the previous three
categories, should be selected. These include TEs from LTR
class that cannot be unambiguously identified as LTR o
internal region, and thus cannot be reconstructed into
partial or full-length elements; as well as solo internal
regions.}

\item{returnMask}{(Default FALSE) Logical value indicating whether a subset
of the input annotations should be returned (default) or
a logical mask of the same length as the input annotations
where \code{TRUE} values indicate what annotations belong
to the TE class we want to obtain with the getter function.}
}
\value{
A [`GRangesList`][GenomicRanges::GRangesList-class] object with
        annotations from class corresponding to the getter function (LTRs,
        LINEs, SINEs or DNA transposons).
}
\description{
Getter functions of TE classes from parsed RepeatMasker annotations.
}
\details{
Retrieves annotations from the TE class corresponding to the getter function,
using RepeatMasker annotations after parsing them with the
\code{OneCodeToFindThemAll()} or \code{rmskatenaparser()} function. The
\code{relLength} parameter can be used to filter out elements with a lower
relative length. Further parameters can be used to fine-tune the type of
elements to be reported.
}
\examples{
rmskat <- annotaTEs(genome="dm6", parsefun=rmskatenaparser,
                    strict=FALSE)

rmskatLTR <- getLTRs(rmskat, relLength=0.95, fullLength=TRUE,
                     partial=TRUE)
rmskatLTR
 
rmskat_line <- getLINEs(rmskat, relLength=0.95)
 
rmskat_sine <- getSINEs(rmskat, relLength=0.95)
 
rmskat_DNAtrans <- getDNAtransposons(rmskat, relLength=0.95)
 
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annotations.R
\name{rmskidentity}
\alias{rmskidentity}
\title{Identity function for parsefun}
\usage{
rmskidentity(gr)
}
\arguments{
\item{gr}{A [`GRanges`][GenomicRanges::GRanges-class] object.}
}
\value{
A [`GRanges`][GenomicRanges::GRanges-class] object.
}
\description{
Identity function for parsefun
}
\details{
Identity function: returns the [`GRanges`][GenomicRanges::GRanges-class]
object without any modification.
}
\examples{
rmskid <- annotaTEs(genome="dm6", parsefun=rmskidentity)
rmskid
        
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R, R/Telescope.R
\docType{class}
\name{TelescopeParam-class}
\alias{TelescopeParam-class}
\alias{TelescopeParam}
\alias{show,TelescopeParam-method}
\title{Telescope parameter class}
\usage{
TelescopeParam(
  bfl,
  teFeatures,
  aggregateby = character(0),
  ovMode = "ovUnion",
  geneFeatures = NULL,
  singleEnd = TRUE,
  strandMode = 1L,
  ignoreStrand = FALSE,
  fragments = FALSE,
  minOverlFract = 0.2,
  pi_prior = 0L,
  theta_prior = 0L,
  em_epsilon = 1e-07,
  maxIter = 100L,
  reassign_mode = "exclude",
  conf_prob = 0.9,
  verbose = TRUE
)

\S4method{show}{TelescopeParam}(object)
}
\arguments{
\item{bfl}{A \code{BamFile} or \code{BamFileList} object, or a character
string vector of BAM filenames.}

\item{teFeatures}{A \code{GRanges} or \code{GRangesList} object. Elements
in this object should have names, which are used as a grouping factor for
genomic ranges forming a common locus (equivalent to "locus" column in
Telescope). This grouping is performed previous to TE expression
quantification, unlike the aggregation of quantifications performed when
the \code{aggregateby} parameter is specified, which is performed after
individual TE instances are quantified.}

\item{aggregateby}{Character vector with column names from the annotation
to be used to aggregate quantifications. By default, this is an empty
vector, which means that the names of the input \code{GRanges} or
\code{GRangesList} object given in the \code{teFeatures} parameter are used
to aggregate quantifications.}

\item{ovMode}{Character vector indicating the overlapping mode. Available
options are: "ovUnion" (default) and "ovIntersectionStrict",
which implement the corresponding methods from HTSeq
(\url{https://htseq.readthedocs.io/en/release_0.11.1/count.html}).
Ambiguous alignments (alignments overlapping > 1 feature) are addressed
as in the original Telescope method: the overlap with the longest
overlapping length is kept.}

\item{geneFeatures}{(Default NULL) A \code{GRanges} or 
\code{GRangesList} object with the
gene annotated features to be quantified. The TEtranscripts approach for
gene expression quantification is used, in which overlaps with multi-mapping
reads are preferentially assigned to TEs. Elements should have names
indicating the gene name/id. In case that \code{geneFeatures} 
is a \code{GRanges} and contains a
metadata column named \code{type}, only the elements with
\code{type} = \code{exon} are considered for the analysis. Then, exon counts
are summarized to the gene level. If NULL, gene expression is not
quantified.}

\item{singleEnd}{(Default TRUE) Logical value indicating if reads are single
(\code{TRUE}) or paired-end (\code{FALSE}).}

\item{strandMode}{(Default 1) Numeric vector which can take values 0, 1 or
2.
The strand mode is a per-object switch on
\code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
objects that controls the behavior of the strand getter. See
\code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
class for further detail. If \code{singleEnd = TRUE}, then \code{strandMode}
is ignored.}

\item{ignoreStrand}{(Default FALSE) A logical which defines if the strand
should be taken into consideration when computing the overlap between reads
and annotated features. When \code{ignoreStrand = FALSE}, an aligned read
is considered to be overlapping an annotated feature as long as they
have a non-empty intersecting genomic range on the same strand, while when
\code{ignoreStrand = TRUE} the strand is not considered.}

\item{fragments}{(Default FALSE) A logical; applied to paired-end data only.
When \code{fragments=FALSE}, the read-counting method only counts
‘mated pairs’ from opposite strands (non-ambiguous properly paired reads), 
while when \code{fragments=TRUE} same-strand pairs, singletons, reads with 
unmapped pairs and other ambiguous or not properly paired fragments
are also counted (see "Pairing criteria" in 
\code{\link[GenomicAlignments]{readGAlignments}()}). \code{fragments=TRUE} 
is equivalent to the original Telescope algorithm. For further details see
\code{\link[GenomicAlignments]{summarizeOverlaps}()}.}

\item{minOverlFract}{(Default 0.2) A numeric scalar. \code{minOverlFract}
is multiplied by the read length and the resulting value is used to
discard alignments for which the overlapping length (number of base
pairs the alignment and the feature overlap) is lower. When no minimum 
overlap is required, set \code{minOverlFract = 0}.}

\item{pi_prior}{(Default 0) A positive integer scalar indicating the prior
on pi. This is equivalent to adding n unique reads.}

\item{theta_prior}{(Default 0) A positive integer scalar storing the prior
on Q. Equivalent to adding n non-unique reads.}

\item{em_epsilon}{(Default 1e-7) A numeric scalar indicating the EM
Algorithm Epsilon cutoff.}

\item{maxIter}{A positive integer scalar storing the maximum number of
iterations of the EM SQUAREM algorithm (Du and Varadhan, 2020). Default
is 100 and this value is passed to the \code{maxiter} parameter of the
\code{\link[SQUAREM]{squarem}()} function.}

\item{reassign_mode}{(Default 'exclude') Character vector indicating
reassignment mode after EM step. 
Available methods are 'exclude' (reads with more than one best
assignment are excluded from the final counts), 'choose' (when reads have
more than one best assignment, one of them is randomly chosen), 'average'
(the read count is divided evenly among the best assignments) and 'conf'
(only assignments that exceed a certain threshold -defined by 
\code{conf_prob} parameter- are accepted, then the read count is
proportionally divided among the assignments above \code{conf_prob}).}

\item{conf_prob}{(Default 0.9) Minimum probability for high confidence
assignment.}

\item{verbose}{(Default \code{TRUE}) Logical value indicating whether to
report progress.}

\item{object}{A \linkS4class{TelescopeParam} object.}
}
\value{
A \linkS4class{TelescopeParam} object.
}
\description{
This is a class for storing parameters
provided to the Telescope algorithm.

Build an object of the class \code{TelescopeParam}.
}
\details{
This is the constructor function for objects of the class
\code{TelescopeParam-class}. This type of object is the input to the
function \code{\link{qtex}()} for quantifying expression of transposable
elements, which will call the Telescope algorithm
\href{https://doi.org/10.1371/journal.pcbi.1006453}{Bendall et al. (2019)}
with this type of object.
}
\section{Slots}{

\describe{
\item{\code{singleEnd}}{(Default TRUE) Logical value indicating if reads are single
(\code{TRUE}) or paired-end (\code{FALSE}).}

\item{\code{strandMode}}{(Default 1) Numeric vector which can take values 0, 1 or 2.
The strand mode is a per-object switch on
\code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
objects that controls the behavior of the strand getter. See
\code{\link[GenomicAlignments:GAlignmentPairs-class]{GAlignmentPairs}}
class for further detail. If \code{singleEnd = TRUE}, then \code{strandMode}
is ignored.}

\item{\code{ignoreStrand}}{(Default FALSE) A logical which defines if the strand
should be taken into consideration when computing the overlap between reads
and annotated features. When \code{ignoreStrand = FALSE}, an aligned read
is considered to be overlapping an annotated feature as long as they
have a non-empty intersecting genomic range on the same strand, while when
\code{ignoreStrand = TRUE} the strand is not considered.}

\item{\code{fragments}}{(Default FALSE) A logical; applied to paired-end data only.
When \code{fragments=FALSE}, the read-counting method only counts
‘mated pairs’ from opposite strands (non-ambiguous properly paired reads), 
while when \code{fragments=TRUE} same-strand pairs, singletons, reads with 
unmapped pairs and other ambiguous or not properly paired fragments
are also counted (see "Pairing criteria" in 
\code{\link[GenomicAlignments]{readGAlignments}()}). \code{fragments=TRUE} 
is equivalent to the original Telescope algorithm. For further details see
\code{\link[GenomicAlignments]{summarizeOverlaps}()}.}

\item{\code{minOverlFract}}{(Default 0.2) A numeric scalar. \code{minOverlFract}
is multiplied by the read length and the resulting value is used to
discard alignments for which the overlapping length (number of base
pairs the alignment and the feature overlap) is lower. When no minimum 
overlap is required, set \code{minOverlFract = 0}.}

\item{\code{pi_prior}}{(Default 0) A positive integer scalar indicating the prior
on pi. This is equivalent to adding n unique reads.}

\item{\code{theta_prior}}{(Default 0) A positive integer scalar storing the prior
on Q. Equivalent to adding n non-unique reads.}

\item{\code{em_epsilon}}{(Default 1e-7) A numeric scalar indicating the EM
Algorithm Epsilon cutoff.}

\item{\code{maxIter}}{A positive integer scalar storing the maximum number of
iterations of the EM SQUAREM algorithm (Du and Varadhan, 2020). Default
is 100 and this value is passed to the \code{maxiter} parameter of the
\code{\link[SQUAREM]{squarem}()} function.}

\item{\code{reassign_mode}}{(Default 'exclude') Character vector indicating
reassignment mode after EM step. 
Available methods are 'exclude' (reads with more than one best
assignment are excluded from the final counts), 'choose' (when reads have
more than one best assignment, one of them is randomly chosen), 'average'
(the read count is divided evenly among the best assignments) and 'conf'
(only assignments that exceed a certain threshold -defined by 
\code{conf_prob} parameter- are accepted, then the read count is
proportionally divided among the assignments above \code{conf_prob}).}

\item{\code{conf_prob}}{(Default 0.9) Minimum probability for high confidence
assignment.}
}}

\examples{
bamfiles <- list.files(system.file("extdata", package="atena"),
                       pattern="*.bam", full.names=TRUE)
\dontrun{
## use the following two instructions to fetch annotations, they are here
## commented out to enable running this example quickly when building and
## checking the package
rmskat <- annotaTEs(genome="dm6", parsefun=rmskatenaparser,
                    strict=FALSE, insert=500)
rmskLTR <- getLTRs(rmskat, relLength=0.8,
                   fullLength=TRUE,
                   partial=TRUE,
                   otherLTR=TRUE)
}

## DO NOT TYPE THIS INSTRUCTION, WHICH JUST LOADS A PRE-COMPUTED ANNOTATION
## YOU SHOULD USE THE INSTRUCTIONS ABOVE TO FETCH ANNOTATIONS
rmskLTR <- readRDS(system.file("extdata", "rmskatLTRrlen80flenpartoth.rds",
                               package="atena"))

## build a parameter object for Telescope
tspar <- TelescopeParam(bfl=bamfiles,
                        teFeatures=rmskLTR,
                        singleEnd=TRUE,
                        ignoreStrand=TRUE)
tspar

}
\references{
Bendall et al. Telescope: characterization of the retrotranscriptome by
accurate estimation of transposable element expression.
PLOS Comp. Biol. 2019;15(9):e1006453. DOI:
\url{https://doi.org/10.1371/journal.pcbi.1006453}

Bendall et al. Telescope: characterization of the retrotranscriptome by
accurate estimation of transposable element expression.
PLOS Comp. Biol. 2019;15(9):e1006453. DOI:
\url{https://doi.org/10.1371/journal.pcbi.1006453}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annotations.R
\name{rmskbasicparser}
\alias{rmskbasicparser}
\title{Parser of RepeatMasker annotations}
\usage{
rmskbasicparser(gr)
}
\arguments{
\item{gr}{A [`GRanges`][GenomicRanges::GRanges-class] object with
RepeatMasker annotations from \link[AnnotationHub]{AnnotationHub}}
}
\value{
A [`GRanges`][GenomicRanges::GRanges-class] object.
}
\description{
Parser of RepeatMasker annotations
}
\details{
Parses annotations by removing low complexity regions, simple repeats,
satellites, rRNA, scRNA, snRNA, srpRNA and tRNA. Also removes TEs with a
strand different than "+" or "-". Modifies "repFamily" and
"repClass" columns when a "?" is present or when they are defined
as "Unknown" or "Other". Finally, assigns a unique id to each TE instance
by adding the suffix "_dup" plus a number at the end of the "repName".
}
\examples{
rmskba <- annotaTEs(genome="dm6", parsefun=rmskbasicparser)
rmskba

}
